//
//  TComStatistics.cpp
//  3DV-HTM
//
//  Created by Fabian Jäger on 15.03.12.
//  Copyright (c) 2012 Institut für Nachrichtentechnik. All rights reserved.
//

/** \file     TComStatistics.h
 \brief    statistics class to analyze mode decisions for CUs
 */


#if RWTH_YUV_STATISTICS

typedef struct _StatColor {
  Pel cbColor;
  Pel crColor;
} StatColor;

static StatColor GrayColor    = { 128, 128 };
static StatColor RedColor     = { 84, 255 };
static StatColor GreenColor   = { 85, 74 };
static StatColor YellowColor  = { 0, 148 };
//static StatColor OrangeColor  = { 36, 190 };
static StatColor BlueColor    = { 255, 63 };
static StatColor PinkColor    = { 212, 234 };
//static StatColor ViolettColor = { 156, 148 };

// ====================================================================================================================
// Constants
// ====================================================================================================================

// ====================================================================================================================
// Tables
// ====================================================================================================================


// ====================================================================================================================
// Constructor / destructor / create / destroy
// ====================================================================================================================

TComYuvStatistics::TComYuvStatistics()
{
  m_pcCurSourcePic      = NULL;
  m_pcCurTargetPicYuv   = NULL;
}

TComYuvStatistics::~TComYuvStatistics()
{
}

// ====================================================================================================================
// Public member functions
// ====================================================================================================================

Void TComYuvStatistics::create()
{
}

Void TComYuvStatistics::init()
{  
}

Void TComYuvStatistics::destroy()
{ 
}

Void TComYuvStatistics::updateStatistics( TComPic* pcSourcePic, TComPicYuv* pcTargetPicYuv, Bool* m_abMyCodingArgument )
{
  // save source and target
  m_pcCurSourcePic = pcSourcePic;
  m_pcCurTargetPicYuv = pcTargetPicYuv;
  
  // initialize statistics YUV with reconstruction
  m_pcCurSourcePic->getPicYuvRec()->copyToPic(m_pcCurTargetPicYuv);
  
  // first fill each CU with color for coding mode
  xFillModes( m_abMyCodingArgument );
  
  // draw CU grid on top
  xDrawGrid();
}

Void TComYuvStatistics::xDrawGrid()
{
  // for every CU in pic
  for( UInt uiCUAddr = 0; uiCUAddr < m_pcCurSourcePic->getNumCUsInFrame(); uiCUAddr++ )
  {
    TComDataCU* pcCU = m_pcCurSourcePic->getCU( uiCUAddr );
    
    xDrawGridPerCU(pcCU, 0, 0);
  }
}

Void TComYuvStatistics::xFillModes( Bool* m_abMyCodingArgument )
{
  // for every CU in pic
  for( UInt uiCUAddr = 0; uiCUAddr < m_pcCurSourcePic->getNumCUsInFrame(); uiCUAddr++ )
  {
    TComDataCU* pcCU = m_pcCurSourcePic->getCU( uiCUAddr );
    
    xFillModesPerCU(pcCU, 0, 0, m_abMyCodingArgument);
  }
}

// ====================================================================================================================
// Private member functions
// ====================================================================================================================

Void TComYuvStatistics::xDrawGridPerCU( TComDataCU *pcCU, UInt uiAbsPartIdx, UInt uiDepth )
{
  UInt uiQparts = pcCU->getPic()->getNumPartInCU() >> ((uiDepth+1) << 1);
  
  // Check if CU is inside of the picture
  UInt uiLPelX   = pcCU->getCUPelX() + g_auiRasterToPelX[ g_auiZscanToRaster[uiAbsPartIdx] ];
  UInt uiTPelY   = pcCU->getCUPelY() + g_auiRasterToPelY[ g_auiZscanToRaster[uiAbsPartIdx] ];
  if ( ( uiLPelX >= pcCU->getSlice()->getSPS()->getPicWidthInLumaSamples() ) || ( uiTPelY >= pcCU->getSlice()->getSPS()->getPicHeightInLumaSamples() ) )
  {
    // The CU lies outside of the picture
    return;
  }

  if (pcCU->getDepth(uiAbsPartIdx) != uiDepth) 
  {
    for (Int i=0; i<4; i++)
    {
      xDrawGridPerCU(pcCU, uiAbsPartIdx + i*uiQparts, uiDepth+1);
    }
  }
  else
  {
    UInt    uiSize            = g_uiMaxCUWidth >> uiDepth;
    
    UInt    uiLCUAddr         = pcCU->getAddr();
    UInt    uiZOrderIdxInCU   = pcCU->getZorderIdxInCU();
    
    Pel*    pDstLuma          = m_pcCurTargetPicYuv->getLumaAddr(uiLCUAddr, uiZOrderIdxInCU + uiAbsPartIdx);
    UInt    uiStride          = m_pcCurTargetPicYuv->getStride();
    Pel*    pDstCb            = m_pcCurTargetPicYuv->getCbAddr(uiLCUAddr, uiZOrderIdxInCU + uiAbsPartIdx);
    Pel*    pDstCr            = m_pcCurTargetPicYuv->getCrAddr(uiLCUAddr, uiZOrderIdxInCU + uiAbsPartIdx);
    UInt    uiStrideC         = m_pcCurTargetPicYuv->getCStride();
    
    for (Int x = 0; x<uiSize; x++)
    {
      pDstLuma[x] = 255 << g_uiBitIncrement;
    }
    for (int x = 0; x<uiSize/2; x++)
    {
      pDstCb[x] = GrayColor.cbColor << g_uiBitIncrement;
      pDstCr[x] = GrayColor.crColor << g_uiBitIncrement;
    }
    
    for (Int y = 0; y<uiSize; y++) 
    {
      pDstLuma[0] = 255 << g_uiBitIncrement;
      pDstLuma += uiStride;
    }
    for (int y = 0; y<uiSize/2; y++) 
    {
      pDstCb[0] = GrayColor.cbColor << g_uiBitIncrement;
      pDstCr[0] = GrayColor.crColor << g_uiBitIncrement;
      pDstCb += uiStrideC;
      pDstCr += uiStrideC;
    }
    
    if ( true )
    {
      // Don't draw PU info
      return;
    }

    // get info about PUs
    PartSize ePartSize = pcCU->getPartitionSize(uiAbsPartIdx);
    
    // draw vertical line
    if( ePartSize == SIZE_Nx2N || ePartSize == SIZE_NxN )
    {
      pDstLuma = m_pcCurTargetPicYuv->getLumaAddr(uiLCUAddr, uiZOrderIdxInCU + uiAbsPartIdx);
      
      for (Int y = 0; y<uiSize; y++) 
      {
        pDstLuma[uiSize>>1] = 255 << g_uiBitIncrement;
        pDstLuma += uiStride;
      }
    }
    
    // draw horizontal line
    if( ePartSize == SIZE_2NxN || ePartSize == SIZE_NxN )
    {
      pDstLuma = m_pcCurTargetPicYuv->getLumaAddr(uiLCUAddr, uiZOrderIdxInCU + uiAbsPartIdx);
      pDstLuma += (uiSize>>1)*uiStride;
      
      for (Int x = 0; x<uiSize; x++) 
      {
        pDstLuma[x] = 255 << g_uiBitIncrement;
      }
    }
    
  }
}

/**
  * Legend:
  * 
  * In case of normal HEVC:
  * - Red:   Intra
  * - Green: Skip
  * - Blue:  Inter
  *
  * In case of normal HEVC SVC:
  * - Red:    Intra
  * - Green:  Skip
  * - Blue:   Inter
  * - Yellow: Inter with reference from BL
  *
  * In case of quantization refinement (for Layer 0 the normal HEVC sceme applies) for Layer 1:
  * - Pink:   IL resi ferinement
  * - Yellow: BL has no coeffs. Normal refined coefficient coding.
  * - Green:  IL resi refinement skip refinement
  * - Red:    Normal intra prediction from enahncement layer pixels
  */
Void TComYuvStatistics::xFillModesPerCU( TComDataCU *pcCU, UInt uiAbsPartIdx, UInt uiDepth, Bool *m_abMyCodingArgument )
{
  for( uiAbsPartIdx = 0; uiAbsPartIdx < pcCU->getPic()->getNumPartInCU(); uiAbsPartIdx++ )
  {
    // Check if CU is inside of the picture
    UInt uiLPelX   = pcCU->getCUPelX() + g_auiRasterToPelX[ g_auiZscanToRaster[uiAbsPartIdx] ];
    UInt uiTPelY   = pcCU->getCUPelY() + g_auiRasterToPelY[ g_auiZscanToRaster[uiAbsPartIdx] ];
    if ( ( uiLPelX >= pcCU->getSlice()->getSPS()->getPicWidthInLumaSamples() ) || ( uiTPelY >= pcCU->getSlice()->getSPS()->getPicHeightInLumaSamples() ) )
    {
      // The CU lies outside of the picture
      continue;
    }

    StatColor targetColor  = GrayColor;
    
#if RWTH_WBP_REFLIST    
    TComCUMvField*  aiCurrMvField[2]  = { pcCU->getCUMvField( REF_PIC_LIST_0 ),        pcCU->getCUMvField( REF_PIC_LIST_1 )        };
    Int             aiCurrRefIdx [2]  = { aiCurrMvField[0]->getRefIdx( uiAbsPartIdx ), aiCurrMvField[1]->getRefIdx( uiAbsPartIdx ) };
#endif
    
    switch( pcCU->getPredictionMode( uiAbsPartIdx ) )
    {
      case MODE_INTRA:
#if RWTH_WBP_SKIP
        if( pcCU->getWarpSkipFlag( uiAbsPartIdx ) )
          targetColor = OrangeColor;
        else
#endif
        targetColor = RedColor;
        break;
      case MODE_INTER:
        targetColor = BlueColor;
        if ( pcCU->isSkipped( uiAbsPartIdx ) )
        {
          targetColor = GreenColor;
        }
        {
          //TComPicMotionInfo* pcMv = pcCU->getPic()->getPicMotionInfo();
          //if (pcMv)
          //{
          //  UInt uiRef0, uiRef1;
          //  pcMv->getRefIndices( pcCU, uiAbsPartIdx, uiRef0, uiRef1 );
          //  if (uiRef0 == pcCU->getPic()->getPOC() || uiRef1 == pcCU->getPic()->getPOC())
          //  {
          //    // We are predicting from the base layer picture
          //    targetColor = YellowColor;
          //  }
          //}
        }
        break;
      case MODE_IL_RWTH_RESI_REFINEMENT:
        targetColor = PinkColor;
        {
          if (m_abMyCodingArgument[RESI_NO_REFINEMENT_IF_BL_NO_RESI])
          {
            // Get base layer CU
            assert ( pcCU->getLayerId() > 0 );
            UInt uiCUAddrBase, uiAbsPartIdxBase;
            TComDataCU* pcColCU = pcCU->getBaseColCU( 0, uiCUAddrBase, uiAbsPartIdxBase );
            if ( !pcColCU->getCbf( uiAbsPartIdx, TEXT_LUMA, uiDepth ) &&
                 !pcColCU->getCbf( uiAbsPartIdx, TEXT_CHROMA_U, uiDepth ) &&
                 !pcColCU->getCbf( uiAbsPartIdx, TEXT_CHROMA_V, uiDepth ) )
            {
              // Encode coefficitnets as if CU was MODE_INTER
              targetColor = YellowColor;
            }
          }
        }
        break;
      case MODE_IL_RESI_NO_REFINEMENT:
        targetColor = GreenColor;
      default:
        // ???
        break;
    }
    
    // fill with selected color
    UInt  uiWidthC        = 2;
    UInt  uiHeightC       = 2;
    
    UInt  uiStrideC       = m_pcCurTargetPicYuv->getCStride  ();
    Pel*  pDestCb         = m_pcCurTargetPicYuv->getCbAddr( pcCU->getAddr(), uiAbsPartIdx );
    Pel*  pDestCr         = m_pcCurTargetPicYuv->getCrAddr( pcCU->getAddr(), uiAbsPartIdx );
    
    for( UInt uiY = 0; uiY < uiHeightC; uiY++ )
    {
      for( UInt uiX = 0; uiX < uiWidthC; uiX++ )
      {
        pDestCb[ uiX ] = targetColor.cbColor << g_uiBitIncrement;
        pDestCr[ uiX ] = targetColor.crColor << g_uiBitIncrement;
      }
      
      pDestCb += uiStrideC;
      pDestCr += uiStrideC;
    }
  }
  
}

#endif
