/* The copyright in this software is being made available under the BSD
 * License, included below. This software may be subject to other third party
 * and contributor rights, including patent rights, and no such rights are
 * granted under this license.  
 *
 * Copyright (c) 2010-2012, ITU/ISO/IEC
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *  * Neither the name of the ITU/ISO/IEC nor the names of its contributors may
 *    be used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

/** \file     TDecSbac.cpp
    \brief    Context-adaptive entropy decoder class
*/

#include "TDecSbac.h"
#include "TDecTop.h"

//! \ingroup TLibDecoder
//! \{

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

TDecSbac::TDecSbac() 
// new structure here
: m_pcBitstream               ( 0 )
, m_pcTDecBinIf               ( NULL )
, m_numContextModels          ( 0 )
, m_cCUSplitFlagSCModel       ( 1,             1,               NUM_SPLIT_FLAG_CTX            , m_contextModels + m_numContextModels, m_numContextModels )
, m_cCUSkipFlagSCModel        ( 1,             1,               NUM_SKIP_FLAG_CTX             , m_contextModels + m_numContextModels, m_numContextModels)
, m_cCUMergeFlagExtSCModel    ( 1,             1,               NUM_MERGE_FLAG_EXT_CTX        , m_contextModels + m_numContextModels, m_numContextModels)
, m_cCUMergeIdxExtSCModel     ( 1,             1,               NUM_MERGE_IDX_EXT_CTX         , m_contextModels + m_numContextModels, m_numContextModels)
, m_cCUPartSizeSCModel        ( 1,             1,               NUM_PART_SIZE_CTX             , m_contextModels + m_numContextModels, m_numContextModels)
, m_cCUPredModeSCModel        ( 1,             1,               NUM_PRED_MODE_CTX             , m_contextModels + m_numContextModels, m_numContextModels)
, m_cCUAlfCtrlFlagSCModel     ( 1,             1,               NUM_ALF_CTRL_FLAG_CTX         , m_contextModels + m_numContextModels, m_numContextModels)
, m_cCUIntraPredSCModel       ( 1,             1,               NUM_ADI_CTX                   , m_contextModels + m_numContextModels, m_numContextModels)
, m_cCUChromaPredSCModel      ( 1,             1,               NUM_CHROMA_PRED_CTX           , m_contextModels + m_numContextModels, m_numContextModels)
, m_cCUDeltaQpSCModel         ( 1,             1,               NUM_DELTA_QP_CTX              , m_contextModels + m_numContextModels, m_numContextModels)
, m_cCUInterDirSCModel        ( 1,             1,               NUM_INTER_DIR_CTX             , m_contextModels + m_numContextModels, m_numContextModels)
, m_cCURefPicSCModel          ( 1,             1,               NUM_REF_NO_CTX                , m_contextModels + m_numContextModels, m_numContextModels)
, m_cCUMvdSCModel             ( 1,             1,               NUM_MV_RES_CTX                , m_contextModels + m_numContextModels, m_numContextModels)
, m_cCUQtCbfSCModel           ( 1,             2,               NUM_QT_CBF_CTX                , m_contextModels + m_numContextModels, m_numContextModels)
, m_cCUTransSubdivFlagSCModel ( 1,             1,               NUM_TRANS_SUBDIV_FLAG_CTX     , m_contextModels + m_numContextModels, m_numContextModels)
, m_cCUQtRootCbfSCModel       ( 1,             1,               NUM_QT_ROOT_CBF_CTX           , m_contextModels + m_numContextModels, m_numContextModels)
, m_cCUSigCoeffGroupSCModel   ( 1,             2,               NUM_SIG_CG_FLAG_CTX           , m_contextModels + m_numContextModels, m_numContextModels)
, m_cCUSigSCModel             ( 1,             1,               NUM_SIG_FLAG_CTX              , m_contextModels + m_numContextModels, m_numContextModels)
, m_cCuCtxLastX               ( 1,             2,               NUM_CTX_LAST_FLAG_XY          , m_contextModels + m_numContextModels, m_numContextModels)
, m_cCuCtxLastY               ( 1,             2,               NUM_CTX_LAST_FLAG_XY          , m_contextModels + m_numContextModels, m_numContextModels)
, m_cCUOneSCModel             ( 1,             1,               NUM_ONE_FLAG_CTX              , m_contextModels + m_numContextModels, m_numContextModels)
, m_cCUAbsSCModel             ( 1,             1,               NUM_ABS_FLAG_CTX              , m_contextModels + m_numContextModels, m_numContextModels)
, m_cMVPIdxSCModel            ( 1,             1,               NUM_MVP_IDX_CTX               , m_contextModels + m_numContextModels, m_numContextModels)
, m_cALFFlagSCModel           ( 1,             1,               NUM_ALF_FLAG_CTX              , m_contextModels + m_numContextModels, m_numContextModels)
, m_cALFUvlcSCModel           ( 1,             1,               NUM_ALF_UVLC_CTX              , m_contextModels + m_numContextModels, m_numContextModels)
, m_cALFSvlcSCModel           ( 1,             1,               NUM_ALF_SVLC_CTX              , m_contextModels + m_numContextModels, m_numContextModels)
, m_cCUAMPSCModel             ( 1,             1,               NUM_CU_AMP_CTX                , m_contextModels + m_numContextModels, m_numContextModels)
#if !SAO_ABS_BY_PASS
, m_cSaoUvlcSCModel           ( 1,             1,               NUM_SAO_UVLC_CTX              , m_contextModels + m_numContextModels, m_numContextModels)
#endif
#if SAO_MERGE_ONE_CTX
, m_cSaoMergeSCModel          ( 1,             1,               NUM_SAO_MERGE_FLAG_CTX   , m_contextModels + m_numContextModels, m_numContextModels)
#else
, m_cSaoMergeLeftSCModel      ( 1,             1,               NUM_SAO_MERGE_LEFT_FLAG_CTX   , m_contextModels + m_numContextModels, m_numContextModels)
, m_cSaoMergeUpSCModel        ( 1,             1,               NUM_SAO_MERGE_UP_FLAG_CTX     , m_contextModels + m_numContextModels, m_numContextModels)
#endif
#if INTRA_BL
, m_cIntraBLPredFlagSCModel   (1,              1,               NUM_INTRA_BL_PRED_CTX         , m_contextModels + m_numContextModels, m_numContextModels)
#endif
, m_cSaoTypeIdxSCModel        ( 1,             1,               NUM_SAO_TYPE_IDX_CTX          , m_contextModels + m_numContextModels, m_numContextModels)
, m_cTransformSkipSCModel     ( 1,             2,               NUM_TRANSFORMSKIP_FLAG_CTX    , m_contextModels + m_numContextModels, m_numContextModels)
, m_CUTransquantBypassFlagSCModel( 1,          1,               NUM_CU_TRANSQUANT_BYPASS_FLAG_CTX, m_contextModels + m_numContextModels, m_numContextModels)
#if RWTH_CTX_FOR_REFINEMENT
, m_cCURefinementModel        ( 1,             2,               NUM_REFINEMNT_CTX             , m_contextModels + m_numContextModels, m_numContextModels)
#endif
#if RWTH_CODE_LASTXY_DIFF
, m_cCuCtxLastDiffSignalling  ( 1,             2,               NUM_CTX_LAST_FLAG_XY_DIFF_SIG , m_contextModels + m_numContextModels, m_numContextModels)
, m_cCuCtxLastDiffX           ( 1,             2,               NUM_CTX_LAST_FLAG_XY          , m_contextModels + m_numContextModels, m_numContextModels)
, m_cCuCtxLastDiffY           ( 1,             2,               NUM_CTX_LAST_FLAG_XY          , m_contextModels + m_numContextModels, m_numContextModels)
#endif
{
  assert( m_numContextModels <= MAX_NUM_CTX_MOD );
#if !REMOVE_FGS
  m_iSliceGranularity = 0;
#endif
}

TDecSbac::~TDecSbac()
{
}

// ====================================================================================================================
// Public member functions
// ====================================================================================================================

Void TDecSbac::resetEntropy(TComSlice* pSlice)
{
  SliceType sliceType  = pSlice->getSliceType();
  Int       qp         = pSlice->getSliceQp();

  if (pSlice->getPPS()->getCabacInitPresentFlag() && pSlice->getCabacInitFlag())
  {
    switch (sliceType)
    {
    case P_SLICE:           // change initialization table to B_SLICE initialization
      sliceType = B_SLICE; 
      break;
    case B_SLICE:           // change initialization table to P_SLICE initialization
      sliceType = P_SLICE; 
      break;
    default     :           // should not occur
      assert(0);
    }
  }

  m_cCUSplitFlagSCModel.initBuffer       ( sliceType, qp, (UChar*)INIT_SPLIT_FLAG );
  m_cCUSkipFlagSCModel.initBuffer        ( sliceType, qp, (UChar*)INIT_SKIP_FLAG );
  m_cCUMergeFlagExtSCModel.initBuffer    ( sliceType, qp, (UChar*)INIT_MERGE_FLAG_EXT );
  m_cCUMergeIdxExtSCModel.initBuffer     ( sliceType, qp, (UChar*)INIT_MERGE_IDX_EXT );
  m_cCUAlfCtrlFlagSCModel.initBuffer     ( sliceType, qp, (UChar*)INIT_ALF_CTRL_FLAG );
  m_cCUPartSizeSCModel.initBuffer        ( sliceType, qp, (UChar*)INIT_PART_SIZE );
  m_cCUAMPSCModel.initBuffer             ( sliceType, qp, (UChar*)INIT_CU_AMP_POS );
  m_cCUPredModeSCModel.initBuffer        ( sliceType, qp, (UChar*)INIT_PRED_MODE );
  m_cCUIntraPredSCModel.initBuffer       ( sliceType, qp, (UChar*)INIT_INTRA_PRED_MODE );
  m_cCUChromaPredSCModel.initBuffer      ( sliceType, qp, (UChar*)INIT_CHROMA_PRED_MODE );
  m_cCUInterDirSCModel.initBuffer        ( sliceType, qp, (UChar*)INIT_INTER_DIR );
  m_cCUMvdSCModel.initBuffer             ( sliceType, qp, (UChar*)INIT_MVD );
  m_cCURefPicSCModel.initBuffer          ( sliceType, qp, (UChar*)INIT_REF_PIC );
  m_cCUDeltaQpSCModel.initBuffer         ( sliceType, qp, (UChar*)INIT_DQP );
  m_cCUQtCbfSCModel.initBuffer           ( sliceType, qp, (UChar*)INIT_QT_CBF );
  m_cCUQtRootCbfSCModel.initBuffer       ( sliceType, qp, (UChar*)INIT_QT_ROOT_CBF );
  m_cCUSigCoeffGroupSCModel.initBuffer   ( sliceType, qp, (UChar*)INIT_SIG_CG_FLAG );
  m_cCUSigSCModel.initBuffer             ( sliceType, qp, (UChar*)INIT_SIG_FLAG );
  m_cCuCtxLastX.initBuffer               ( sliceType, qp, (UChar*)INIT_LAST );
  m_cCuCtxLastY.initBuffer               ( sliceType, qp, (UChar*)INIT_LAST );
  m_cCUOneSCModel.initBuffer             ( sliceType, qp, (UChar*)INIT_ONE_FLAG );
  m_cCUAbsSCModel.initBuffer             ( sliceType, qp, (UChar*)INIT_ABS_FLAG );
  m_cMVPIdxSCModel.initBuffer            ( sliceType, qp, (UChar*)INIT_MVP_IDX );
  m_cALFFlagSCModel.initBuffer           ( sliceType, qp, (UChar*)INIT_ALF_FLAG );
  m_cALFUvlcSCModel.initBuffer           ( sliceType, qp, (UChar*)INIT_ALF_UVLC );
  m_cALFSvlcSCModel.initBuffer           ( sliceType, qp, (UChar*)INIT_ALF_SVLC );
#if !SAO_ABS_BY_PASS
  m_cSaoUvlcSCModel.initBuffer           ( sliceType, qp, (UChar*)INIT_SAO_UVLC );
#endif
#if SAO_MERGE_ONE_CTX
  m_cSaoMergeSCModel.initBuffer          ( sliceType, qp, (UChar*)INIT_SAO_MERGE_FLAG );
#else
  m_cSaoMergeLeftSCModel.initBuffer      ( sliceType, qp, (UChar*)INIT_SAO_MERGE_LEFT_FLAG );
  m_cSaoMergeUpSCModel.initBuffer        ( sliceType, qp, (UChar*)INIT_SAO_MERGE_UP_FLAG );
#endif
  m_cSaoTypeIdxSCModel.initBuffer        ( sliceType, qp, (UChar*)INIT_SAO_TYPE_IDX );

  m_cCUTransSubdivFlagSCModel.initBuffer ( sliceType, qp, (UChar*)INIT_TRANS_SUBDIV_FLAG );
#if INTRA_BL
  m_cIntraBLPredFlagSCModel.initBuffer   ( sliceType, qp, (UChar*)INIT_INTRA_BL_PRED_FLAG );
#endif
  m_cTransformSkipSCModel.initBuffer     ( sliceType, qp, (UChar*)INIT_TRANSFORMSKIP_FLAG );
  m_CUTransquantBypassFlagSCModel.initBuffer( sliceType, qp, (UChar*)INIT_CU_TRANSQUANT_BYPASS_FLAG );
#if RWTH_CTX_FOR_REFINEMENT
    m_cCURefinementModel.initBuffer      ( sliceType, qp, (UChar*)INIT_REFINEMENT_FLAG );
#endif
#if RWTH_CODE_LASTXY_DIFF
  m_cCuCtxLastDiffSignalling.initBuffer  ( sliceType, qp, (UChar*)INIT_LAST_XY_DIFF_SIGNALLING );
  m_cCuCtxLastDiffX.initBuffer           ( sliceType, qp, (UChar*)INIT_LAST );
  m_cCuCtxLastDiffY.initBuffer           ( sliceType, qp, (UChar*)INIT_LAST );
#endif

  m_uiLastDQpNonZero  = 0;
  
  // new structure
  m_uiLastQp          = qp;
  
  m_pcTDecBinIf->start();
}

/** The function does the following: Read out terminate bit. Flush CABAC. Byte-align for next tile.
 *  Intialize CABAC states. Start CABAC.
 */
Void TDecSbac::updateContextTables( SliceType eSliceType, Int iQp )
{
  UInt uiBit;
  parseTerminatingBit(uiBit);
  m_pcTDecBinIf->finish();  
  m_pcBitstream->readOutTrailingBits();
  m_cCUSplitFlagSCModel.initBuffer       ( eSliceType, iQp, (UChar*)INIT_SPLIT_FLAG );
  m_cCUSkipFlagSCModel.initBuffer        ( eSliceType, iQp, (UChar*)INIT_SKIP_FLAG );
  m_cCUMergeFlagExtSCModel.initBuffer    ( eSliceType, iQp, (UChar*)INIT_MERGE_FLAG_EXT );
  m_cCUMergeIdxExtSCModel.initBuffer     ( eSliceType, iQp, (UChar*)INIT_MERGE_IDX_EXT );
  m_cCUAlfCtrlFlagSCModel.initBuffer     ( eSliceType, iQp, (UChar*)INIT_ALF_CTRL_FLAG );
  m_cCUPartSizeSCModel.initBuffer        ( eSliceType, iQp, (UChar*)INIT_PART_SIZE );
  m_cCUAMPSCModel.initBuffer             ( eSliceType, iQp, (UChar*)INIT_CU_AMP_POS );
  m_cCUPredModeSCModel.initBuffer        ( eSliceType, iQp, (UChar*)INIT_PRED_MODE );
  m_cCUIntraPredSCModel.initBuffer       ( eSliceType, iQp, (UChar*)INIT_INTRA_PRED_MODE );
  m_cCUChromaPredSCModel.initBuffer      ( eSliceType, iQp, (UChar*)INIT_CHROMA_PRED_MODE );
  m_cCUInterDirSCModel.initBuffer        ( eSliceType, iQp, (UChar*)INIT_INTER_DIR );
  m_cCUMvdSCModel.initBuffer             ( eSliceType, iQp, (UChar*)INIT_MVD );
  m_cCURefPicSCModel.initBuffer          ( eSliceType, iQp, (UChar*)INIT_REF_PIC );
  m_cCUDeltaQpSCModel.initBuffer         ( eSliceType, iQp, (UChar*)INIT_DQP );
  m_cCUQtCbfSCModel.initBuffer           ( eSliceType, iQp, (UChar*)INIT_QT_CBF );
  m_cCUQtRootCbfSCModel.initBuffer       ( eSliceType, iQp, (UChar*)INIT_QT_ROOT_CBF );
  m_cCUSigCoeffGroupSCModel.initBuffer   ( eSliceType, iQp, (UChar*)INIT_SIG_CG_FLAG );
  m_cCUSigSCModel.initBuffer             ( eSliceType, iQp, (UChar*)INIT_SIG_FLAG );
  m_cCuCtxLastX.initBuffer               ( eSliceType, iQp, (UChar*)INIT_LAST );
  m_cCuCtxLastY.initBuffer               ( eSliceType, iQp, (UChar*)INIT_LAST );
  m_cCUOneSCModel.initBuffer             ( eSliceType, iQp, (UChar*)INIT_ONE_FLAG );
  m_cCUAbsSCModel.initBuffer             ( eSliceType, iQp, (UChar*)INIT_ABS_FLAG );
  m_cMVPIdxSCModel.initBuffer            ( eSliceType, iQp, (UChar*)INIT_MVP_IDX );
  m_cALFFlagSCModel.initBuffer           ( eSliceType, iQp, (UChar*)INIT_ALF_FLAG );
  m_cALFUvlcSCModel.initBuffer           ( eSliceType, iQp, (UChar*)INIT_ALF_UVLC );
  m_cALFSvlcSCModel.initBuffer           ( eSliceType, iQp, (UChar*)INIT_ALF_SVLC );
#if !SAO_ABS_BY_PASS
  m_cSaoUvlcSCModel.initBuffer           ( eSliceType, iQp, (UChar*)INIT_SAO_UVLC );
#endif
#if SAO_MERGE_ONE_CTX
  m_cSaoMergeSCModel.initBuffer          ( eSliceType, iQp, (UChar*)INIT_SAO_MERGE_FLAG );
#else
  m_cSaoMergeLeftSCModel.initBuffer      ( eSliceType, iQp, (UChar*)INIT_SAO_MERGE_LEFT_FLAG );
  m_cSaoMergeUpSCModel.initBuffer        ( eSliceType, iQp, (UChar*)INIT_SAO_MERGE_UP_FLAG );
#endif
  m_cSaoTypeIdxSCModel.initBuffer        ( eSliceType, iQp, (UChar*)INIT_SAO_TYPE_IDX );
  m_cCUTransSubdivFlagSCModel.initBuffer ( eSliceType, iQp, (UChar*)INIT_TRANS_SUBDIV_FLAG );
#if INTRA_BL
  m_cIntraBLPredFlagSCModel.initBuffer   ( eSliceType, iQp, (UChar*)INIT_INTRA_BL_PRED_FLAG );
#endif
  m_cTransformSkipSCModel.initBuffer     ( eSliceType, iQp, (UChar*)INIT_TRANSFORMSKIP_FLAG );
  m_CUTransquantBypassFlagSCModel.initBuffer( eSliceType, iQp, (UChar*)INIT_CU_TRANSQUANT_BYPASS_FLAG );
#if RWTH_CTX_FOR_REFINEMENT
  m_cCURefinementModel.initBuffer        ( eSliceType, iQp, (UChar*)INIT_REFINEMENT_FLAG );
#endif
#if RWTH_CODE_LASTXY_DIFF
  m_cCuCtxLastDiffSignalling.initBuffer  ( eSliceType, iQp, (UChar*)INIT_LAST_XY_DIFF_SIGNALLING );
  m_cCuCtxLastDiffX.initBuffer           ( eSliceType, iQp, (UChar*)INIT_LAST );
  m_cCuCtxLastDiffY.initBuffer           ( eSliceType, iQp, (UChar*)INIT_LAST );
#endif

  m_pcTDecBinIf->start();
}

Void TDecSbac::parseTerminatingBit( UInt& ruiBit )
{
  m_pcTDecBinIf->decodeBinTrm( ruiBit );
  DTRACE_CABAC_N
}


Void TDecSbac::xReadUnaryMaxSymbol( UInt& ruiSymbol, ContextModel* pcSCModel, Int iOffset, UInt uiMaxSymbol )
{
  DTRACE_CABAC_T( "\tStart Coding UnaryMaxSymbol - iOffset: " )
  DTRACE_CABAC_V( iOffset )
  DTRACE_CABAC_T( " - uiMaxSymbol: " )
  DTRACE_CABAC_V( uiMaxSymbol )
  DTRACE_CABAC_N

  if (uiMaxSymbol == 0)
  {
    ruiSymbol = 0;
    DTRACE_CABAC_T( "\tEnd Coding UnaryMaxSymbol - ruiSymbol: 0\n")
    return;
  }
  
  m_pcTDecBinIf->decodeBin( ruiSymbol, pcSCModel[0] );
  DTRACE_CABAC_T( "\tctx:0\n" )
  
  if( ruiSymbol == 0 || uiMaxSymbol == 1 )
  {
    DTRACE_CABAC_T( "\tEnd Coding UnaryMaxSymbol - ruiSymbol: " )
    DTRACE_CABAC_V( ruiSymbol )
    DTRACE_CABAC_N
    return;
  }
  
  UInt uiSymbol = 0;
  UInt uiCont;
  
  do
  {
    m_pcTDecBinIf->decodeBin( uiCont, pcSCModel[ iOffset ] );
    DTRACE_CABAC_T( "\tctx:" )
    DTRACE_CABAC_V( iOffset )
    DTRACE_CABAC_N
    uiSymbol++;
  }
  while( uiCont && ( uiSymbol < uiMaxSymbol - 1 ) );
  
  if( uiCont && ( uiSymbol == uiMaxSymbol - 1 ) )
  {
    uiSymbol++;
  }
  
  ruiSymbol = uiSymbol;
  DTRACE_CABAC_T( "\tEnd Coding UnaryMaxSymbol - ruiSymbol: ")
  DTRACE_CABAC_V( ruiSymbol )
  DTRACE_CABAC_N
}

Void TDecSbac::xReadEpExGolomb( UInt& ruiSymbol, UInt uiCount )
{
  DTRACE_CABAC_T( "\tStart Coding EpExGolomb - uiCount: " )
  DTRACE_CABAC_V( uiCount )
  DTRACE_CABAC_N

  UInt uiSymbol = 0;
  UInt uiBit = 1;
  
  while( uiBit )
  {
    m_pcTDecBinIf->decodeBinEP( uiBit );
    DTRACE_CABAC_N
    uiSymbol += uiBit << uiCount++;
  }
  
  if ( --uiCount )
  {
    UInt bins;
    m_pcTDecBinIf->decodeBinsEP( bins, uiCount );
    uiSymbol += bins;
  }
  
  ruiSymbol = uiSymbol;
  DTRACE_CABAC_T( "\tEnd Coding EpExGolomb - ruiSymbol: ")
  DTRACE_CABAC_V( ruiSymbol )
  DTRACE_CABAC_N
}

Void TDecSbac::xReadUnarySymbol( UInt& ruiSymbol, ContextModel* pcSCModel, Int iOffset )
{
  DTRACE_CABAC_T( "\tStart Coding UnarySymbol - iOffset: " )
  DTRACE_CABAC_V( iOffset )
  DTRACE_CABAC_N

  m_pcTDecBinIf->decodeBin( ruiSymbol, pcSCModel[0] );
  DTRACE_CABAC_T( "\tctx:0\n" )
  
  if( !ruiSymbol )
  {
    DTRACE_CABAC_T( "\tEnd Coding UnarySymbol - ruiSymbol: 0\n")
    return;
  }
  
  UInt uiSymbol = 0;
  UInt uiCont;
  
  do
  {
    m_pcTDecBinIf->decodeBin( uiCont, pcSCModel[ iOffset ] );
    DTRACE_CABAC_T( "\tctx:" )
    DTRACE_CABAC_V( iOffset )
    DTRACE_CABAC_N
    uiSymbol++;
  }
  while( uiCont );
  
  ruiSymbol = uiSymbol;
  DTRACE_CABAC_T( "\tEnd Coding UnarySymbol - ruiSymbol: ")
  DTRACE_CABAC_V( ruiSymbol )
  DTRACE_CABAC_N
}


/** Parsing of coeff_abs_level_remaing
 * \param ruiSymbol reference to coeff_abs_level_remaing
 * \param ruiParam reference to parameter
 * \returns Void
 */
Void TDecSbac::xReadCoefRemainExGolomb ( UInt &rSymbol, UInt &rParam )
{
  DTRACE_CABAC_T( "\tStart Coding GoRiceExGolomb - ruiGoRiceParam: " )
  DTRACE_CABAC_V( rParam )
  DTRACE_CABAC_N

  UInt prefix   = 0;
  UInt codeWord = 0;
  do
  {
    prefix++;
    m_pcTDecBinIf->decodeBinEP( codeWord );
    DTRACE_CABAC_N
  }
  while( codeWord);
  codeWord  = 1 - codeWord;
  prefix -= codeWord;
  codeWord=0;
#if COEF_REMAIN_BIN_REDUCTION
  if (prefix < COEF_REMAIN_BIN_REDUCTION )
#else
  if (prefix < 8 )
#endif
  {
    m_pcTDecBinIf->decodeBinsEP(codeWord,rParam);
    rSymbol = (prefix<<rParam) + codeWord;
  }
  else
  {
#if COEF_REMAIN_BIN_REDUCTION
    m_pcTDecBinIf->decodeBinsEP(codeWord,prefix-COEF_REMAIN_BIN_REDUCTION+rParam);
    rSymbol = (((1<<(prefix-COEF_REMAIN_BIN_REDUCTION))+COEF_REMAIN_BIN_REDUCTION-1)<<rParam)+codeWord;
#else
    m_pcTDecBinIf->decodeBinsEP(codeWord,prefix-8+rParam);
    rSymbol = (((1<<(prefix-8))+8-1)<<rParam)+codeWord;
#endif
  }
  DTRACE_CABAC_T( "\tEnd Coding GoRiceExGolomb - ruiGoRiceParam: " )
  DTRACE_CABAC_V( rParam )
  DTRACE_CABAC_T( "\truiSymbol: ")
  DTRACE_CABAC_V( rSymbol )
  DTRACE_CABAC_N
}

/** Parse I_PCM information. 
 * \param pcCU
 * \param uiAbsPartIdx 
 * \param uiDepth
 * \returns Void
 *
 * If I_PCM flag indicates that the CU is I_PCM, parse its PCM alignment bits and codes. 
 */
Void TDecSbac::parseIPCMInfo ( TComDataCU* pcCU, UInt uiAbsPartIdx, UInt uiDepth )
{
  UInt uiSymbol;
  Int numSubseqIPCM = 0;
  Bool readPCMSampleFlag = false;

  if(pcCU->getNumSucIPCM() > 0) 
  {
    readPCMSampleFlag = true;
  }
  else
  {
    m_pcTDecBinIf->decodeBinTrm(uiSymbol);
    DTRACE_CABAC_N

    if (uiSymbol)
    {
      readPCMSampleFlag = true;
      m_pcTDecBinIf->decodeNumSubseqIPCM(numSubseqIPCM);
      pcCU->setNumSucIPCM(numSubseqIPCM + 1);
      m_pcTDecBinIf->decodePCMAlignBits();
    }
  }

  if (readPCMSampleFlag == true)
  {
    Bool bIpcmFlag = true;

    pcCU->setPartSizeSubParts  ( SIZE_2Nx2N, uiAbsPartIdx, uiDepth );
    pcCU->setSizeSubParts      ( g_uiMaxCUWidth>>uiDepth, g_uiMaxCUHeight>>uiDepth, uiAbsPartIdx, uiDepth );
    pcCU->setTrIdxSubParts     ( 0, uiAbsPartIdx, uiDepth );
    pcCU->setIPCMFlagSubParts  ( bIpcmFlag, uiAbsPartIdx, uiDepth );

    UInt uiMinCoeffSize = pcCU->getPic()->getMinCUWidth()*pcCU->getPic()->getMinCUHeight();
    UInt uiLumaOffset   = uiMinCoeffSize*uiAbsPartIdx;
    UInt uiChromaOffset = uiLumaOffset>>2;

    Pel* piPCMSample;
    UInt uiWidth;
    UInt uiHeight;
    UInt uiSampleBits;
    UInt uiX, uiY;

    piPCMSample = pcCU->getPCMSampleY() + uiLumaOffset;
    uiWidth = pcCU->getWidth(uiAbsPartIdx);
    uiHeight = pcCU->getHeight(uiAbsPartIdx);
    uiSampleBits = pcCU->getSlice()->getSPS()->getPCMBitDepthLuma();

    for(uiY = 0; uiY < uiHeight; uiY++)
    {
      for(uiX = 0; uiX < uiWidth; uiX++)
      {
        UInt uiSample;
        m_pcTDecBinIf->xReadPCMCode(uiSampleBits, uiSample);
        piPCMSample[uiX] = uiSample;
      }
      piPCMSample += uiWidth;
    }

    piPCMSample = pcCU->getPCMSampleCb() + uiChromaOffset;
    uiWidth = pcCU->getWidth(uiAbsPartIdx)/2;
    uiHeight = pcCU->getHeight(uiAbsPartIdx)/2;
    uiSampleBits = pcCU->getSlice()->getSPS()->getPCMBitDepthChroma();

    for(uiY = 0; uiY < uiHeight; uiY++)
    {
      for(uiX = 0; uiX < uiWidth; uiX++)
      {
        UInt uiSample;
        m_pcTDecBinIf->xReadPCMCode(uiSampleBits, uiSample);
        piPCMSample[uiX] = uiSample;
      }
      piPCMSample += uiWidth;
    }

    piPCMSample = pcCU->getPCMSampleCr() + uiChromaOffset;
    uiWidth = pcCU->getWidth(uiAbsPartIdx)/2;
    uiHeight = pcCU->getHeight(uiAbsPartIdx)/2;
    uiSampleBits = pcCU->getSlice()->getSPS()->getPCMBitDepthChroma();

    for(uiY = 0; uiY < uiHeight; uiY++)
    {
      for(uiX = 0; uiX < uiWidth; uiX++)
      {
        UInt uiSample;
        m_pcTDecBinIf->xReadPCMCode(uiSampleBits, uiSample);
        piPCMSample[uiX] = uiSample;
      }
      piPCMSample += uiWidth;
    }

    pcCU->setNumSucIPCM( pcCU->getNumSucIPCM() - 1);
    if(pcCU->getNumSucIPCM() == 0)
    {
      m_pcTDecBinIf->resetBac();
    }
  }
}

Void TDecSbac::parseCUTransquantBypassFlag( TComDataCU* pcCU, UInt uiAbsPartIdx, UInt uiDepth )
{
  UInt uiSymbol;
  m_pcTDecBinIf->decodeBin( uiSymbol, m_CUTransquantBypassFlagSCModel.get( 0, 0, 0 ) );
  DTRACE_CABAC_T( "\tTransquantBypassFlag\n" )
  pcCU->setCUTransquantBypassSubParts(uiSymbol ? true : false, uiAbsPartIdx, uiDepth);
}

/** parse skip flag
 * \param pcCU
 * \param uiAbsPartIdx 
 * \param uiDepth
 * \returns Void
 */
Void TDecSbac::parseSkipFlag( TComDataCU* pcCU, UInt uiAbsPartIdx, UInt uiDepth )
{
  if( pcCU->getSlice()->isIntra() )
  {
    return;
  }
#if RWTH_RESI_REFINEMENT
  if ( getDecTop()->getMyCodingArgument()[INTER_LAYER_RWTH_RESI_REFINEMENT] )
  {
    return;
  }
#endif
  
  UInt uiSymbol = 0;
  UInt uiCtxSkip = pcCU->getCtxSkipFlag( uiAbsPartIdx );
  m_pcTDecBinIf->decodeBin( uiSymbol, m_cCUSkipFlagSCModel.get( 0, 0, uiCtxSkip ) );
  
  DTRACE_CABAC_T( "\tSkipFlag" );
  DTRACE_CABAC_T( "\tuiCtxSkip=");
  DTRACE_CABAC_V( uiCtxSkip );
  DTRACE_CABAC_T( "\n");
  
  if( uiSymbol )
  {
#if SKIP_FLAG
    pcCU->setSkipFlagSubParts( true,        uiAbsPartIdx, uiDepth );
#endif
    pcCU->setPredModeSubParts( MODE_INTER,  uiAbsPartIdx, uiDepth );
    pcCU->setPartSizeSubParts( SIZE_2Nx2N, uiAbsPartIdx, uiDepth );
    pcCU->setSizeSubParts( g_uiMaxCUWidth>>uiDepth, g_uiMaxCUHeight>>uiDepth, uiAbsPartIdx, uiDepth );
    pcCU->setMergeFlagSubParts( true , uiAbsPartIdx, 0, uiDepth );
  }
}

#if INTRA_BL
Void TDecSbac::parseIntraBLFlag( TComDataCU* pcCU, UInt uiAbsPartIdx, UInt uiPartIdx, UInt uiDepth )
{
  if( pcCU->getLayerId() == 0 )
  {
    return;
  }

  UInt uiSymbol = 0;

  UInt uiCtxIntraBL = pcCU->getCtxIntraBLFlag( uiAbsPartIdx ) ;
  m_pcTDecBinIf->decodeBin( uiSymbol, m_cIntraBLPredFlagSCModel.get( 0, 0, uiCtxIntraBL )); 
  DTRACE_CABAC_VL( g_nSymbolCounter++ );
  DTRACE_CABAC_T( "\tIntrBLFlag" );
  DTRACE_CABAC_T( "\tuiSymbol: ");
  DTRACE_CABAC_V( uiSymbol );
  DTRACE_CABAC_T( "\n");

  if ( uiSymbol )
  {
    pcCU->setPartSizeSubParts( SIZE_2Nx2N, uiAbsPartIdx, uiDepth );
    pcCU->setPredModeSubParts( MODE_INTRA_BL, uiAbsPartIdx, uiDepth );
    pcCU->setTrIdxSubParts( 0, uiAbsPartIdx, uiDepth );
    pcCU->setLumaIntraDirSubParts ( DC_IDX, uiAbsPartIdx, uiDepth );   
  }
}
#endif

/** parse merge flag
 * \param pcCU
 * \param uiAbsPartIdx 
 * \param uiDepth
 * \param uiPUIdx
 * \returns Void
 */
Void TDecSbac::parseMergeFlag ( TComDataCU* pcCU, UInt uiAbsPartIdx, UInt uiDepth, UInt uiPUIdx )
{
  UInt uiSymbol;
  m_pcTDecBinIf->decodeBin( uiSymbol, *m_cCUMergeFlagExtSCModel.get( 0 ) );
  pcCU->setMergeFlagSubParts( uiSymbol ? true : false, uiAbsPartIdx, uiPUIdx, uiDepth );

  DTRACE_CABAC_T( "\tMergeFlag \tAddress=" );
  DTRACE_CABAC_V( pcCU->getAddr() );
  DTRACE_CABAC_T( "\tuiAbsPartIdx=" );
  DTRACE_CABAC_V( uiAbsPartIdx );
  DTRACE_CABAC_N;
}

Void TDecSbac::parseMergeIndex ( TComDataCU* pcCU, UInt& ruiMergeIndex, UInt uiAbsPartIdx, UInt uiDepth )
{
  DTRACE_CABAC_T( "\tStart Code MergeIndex\tuiAbsPartIdx=" );
  DTRACE_CABAC_V( uiAbsPartIdx )
  DTRACE_CABAC_N

  UInt uiUnaryIdx = 0;
  UInt uiNumCand = pcCU->getSlice()->getMaxNumMergeCand();
  if ( uiNumCand > 1 )
  {
    for( ; uiUnaryIdx < uiNumCand - 1; ++uiUnaryIdx )
    {
      UInt uiSymbol = 0;
      if ( uiUnaryIdx==0 )
      {
        m_pcTDecBinIf->decodeBin( uiSymbol, m_cCUMergeIdxExtSCModel.get( 0, 0, 0 ) );
        DTRACE_CABAC_N
      }
      else
      {
        m_pcTDecBinIf->decodeBinEP( uiSymbol );
        DTRACE_CABAC_N
      }
      if( uiSymbol == 0 )
      {
        break;
      }
    }
  }
  ruiMergeIndex = uiUnaryIdx;

  DTRACE_CABAC_T( "\tEnd Code MergeIndex	ruiMergeIndex=" );
  DTRACE_CABAC_V( ruiMergeIndex )
  DTRACE_CABAC_N
}

Void TDecSbac::parseMVPIdx      ( Int& riMVPIdx )
{
  DTRACE_CABAC_T( "\tStart Code MVPIdx\n" )
  UInt uiSymbol;
  xReadUnaryMaxSymbol(uiSymbol, m_cMVPIdxSCModel.get(0), 1, AMVP_MAX_NUM_CANDS-1);
  riMVPIdx = uiSymbol;

#if ENC_DEC_TRACE
  DTRACE_CABAC_T( "\tEnd Code MVPIdx - riMVPIdx=" )
  DTRACE_CABAC_V( riMVPIdx )
  //DTRACE_CABAC_T( "\tAMVP: " )
  //DTRACE_CABAC_T( pcCU->getAMVPString( eRefList, uiAbsPartIdx ).c_str() )
  DTRACE_CABAC_N
#endif
}

Void TDecSbac::parseSplitFlag     ( TComDataCU* pcCU, UInt uiAbsPartIdx, UInt uiDepth )
{
  if( uiDepth == g_uiMaxCUDepth - g_uiAddCUDepth )
  {
    pcCU->setDepthSubParts( uiDepth, uiAbsPartIdx );
    return;
  }

#if RWTH_RESI_REFINEMENT
  if ( getDecTop()->getMyCodingArgument()[INTER_LAYER_RWTH_RESI_REFINEMENT])
  {
    return;
  }
#endif
  
  UInt uiSymbol;
  UInt uiCtx = pcCU->getCtxSplitFlag( uiAbsPartIdx, uiDepth );
  m_pcTDecBinIf->decodeBin( uiSymbol, m_cCUSplitFlagSCModel.get( 0, 0, uiCtx ) );
  DTRACE_CABAC_T( "\tSplitFlag" )
  DTRACE_CABAC_T( "\tuiCtxSplit=");
  DTRACE_CABAC_V( uiCtx );
  DTRACE_CABAC_T( "\tuiAbsPartIdx=");
  DTRACE_CABAC_V( uiAbsPartIdx );
  DTRACE_CABAC_T( "\tuiDepth=");
  DTRACE_CABAC_V( uiDepth );
  DTRACE_CABAC_T( "\n");

  pcCU->setDepthSubParts( uiDepth + uiSymbol, uiAbsPartIdx );
  
  return;
}

/** parse partition size
 * \param pcCU
 * \param uiAbsPartIdx 
 * \param uiDepth
 * \returns Void
 */
Void TDecSbac::parsePartSize( TComDataCU* pcCU, UInt uiAbsPartIdx, UInt uiDepth )
{
  UInt uiSymbol, uiMode = 0;
  PartSize eMode;
  
#if INTRA_BL
  if ( pcCU->isIntraBL( uiAbsPartIdx ) )
  {
    assert( 0 );
  }
#endif

#if RWTH_RESI_REFINEMENT
  if (pcCU->getPredictionMode(uiAbsPartIdx)==MODE_IL_RWTH_RESI_REFINEMENT)
  {
    pcCU->setPartSizeSubParts( SIZE_2Nx2N, uiAbsPartIdx, uiDepth);
    pcCU->setSizeSubParts( g_uiMaxCUWidth>>uiDepth, g_uiMaxCUHeight>>uiDepth, uiAbsPartIdx, uiDepth );
    return;
  }
#endif

  DTRACE_CABAC_T( "\tStart Code PartSize" );
  DTRACE_CABAC_T( "\tuiAbsPartIdx=" );
  DTRACE_CABAC_V( uiAbsPartIdx )
  DTRACE_CABAC_T( "\tuiDepth=" );
  DTRACE_CABAC_V( uiDepth )
  DTRACE_CABAC_N
    
  if ( pcCU->isIntra( uiAbsPartIdx ) )
  {
    uiSymbol = 1;
    if( uiDepth == g_uiMaxCUDepth - g_uiAddCUDepth )
    {
      m_pcTDecBinIf->decodeBin( uiSymbol, m_cCUPartSizeSCModel.get( 0, 0, 0) );
      DTRACE_CABAC_N
    }
    eMode = uiSymbol ? SIZE_2Nx2N : SIZE_NxN;
    UInt uiTrLevel = 0;    
    UInt uiWidthInBit  = g_aucConvertToBit[pcCU->getWidth(uiAbsPartIdx)]+2;
    UInt uiTrSizeInBit = g_aucConvertToBit[pcCU->getSlice()->getSPS()->getMaxTrSize()]+2;
    uiTrLevel          = uiWidthInBit >= uiTrSizeInBit ? uiWidthInBit - uiTrSizeInBit : 0;
    if( eMode == SIZE_NxN )
    {
      pcCU->setTrIdxSubParts( 1+uiTrLevel, uiAbsPartIdx, uiDepth );
    }
    else
    {
      pcCU->setTrIdxSubParts( uiTrLevel, uiAbsPartIdx, uiDepth );
    }
  }
  else
  {
    UInt uiMaxNumBits = 2;
    if( uiDepth == g_uiMaxCUDepth - g_uiAddCUDepth && !( (g_uiMaxCUWidth>>uiDepth) == 8 && (g_uiMaxCUHeight>>uiDepth) == 8 ) )
    {
      uiMaxNumBits ++;
    }
    for ( UInt ui = 0; ui < uiMaxNumBits; ui++ )
    {
      m_pcTDecBinIf->decodeBin( uiSymbol, m_cCUPartSizeSCModel.get( 0, 0, ui) );
      DTRACE_CABAC_N
      if ( uiSymbol )
      {
        break;
      }
      uiMode++;
    }
    eMode = (PartSize) uiMode;
    if ( pcCU->getSlice()->getSPS()->getAMPAcc( uiDepth ) )
    {
      if (eMode == SIZE_2NxN)
      {
        m_pcTDecBinIf->decodeBin(uiSymbol, m_cCUAMPSCModel.get( 0, 0, 0 ));
        DTRACE_CABAC_N

        if (uiSymbol == 0)
        {
          m_pcTDecBinIf->decodeBinEP(uiSymbol);
          DTRACE_CABAC_N

          eMode = (uiSymbol == 0? SIZE_2NxnU : SIZE_2NxnD);
        }
      }
      else if (eMode == SIZE_Nx2N)
      {
        m_pcTDecBinIf->decodeBin(uiSymbol, m_cCUAMPSCModel.get( 0, 0, 0 ));
        DTRACE_CABAC_N

        if (uiSymbol == 0)
        {
          m_pcTDecBinIf->decodeBinEP(uiSymbol);
          DTRACE_CABAC_N

          eMode = (uiSymbol == 0? SIZE_nLx2N : SIZE_nRx2N);
        }
      }
    }
  }
  pcCU->setPartSizeSubParts( eMode, uiAbsPartIdx, uiDepth );
  pcCU->setSizeSubParts( g_uiMaxCUWidth>>uiDepth, g_uiMaxCUHeight>>uiDepth, uiAbsPartIdx, uiDepth );

  char eModeName[12];
  switch (eMode) {
      case SIZE_2Nx2N: sprintf(eModeName, "SIZE_2Nx2N"); break;
      case SIZE_2NxN:  sprintf(eModeName, "SIZE_2NxN" ); break;
      case SIZE_Nx2N:  sprintf(eModeName, "SIZE_Nx2N" ); break;
      case SIZE_NxN:   sprintf(eModeName, "SIZE_NxN"  ); break;
      case SIZE_2NxnU: sprintf(eModeName, "SIZE_2NxnU"); break;
      case SIZE_2NxnD: sprintf(eModeName, "SIZE_2NxnD"); break;
      case SIZE_nLx2N: sprintf(eModeName, "SIZE_nLx2N"); break;
      case SIZE_nRx2N: sprintf(eModeName, "SIZE_nRx2N"); break;
      default:         sprintf(eModeName, "Unknown"   ); break;
  }

  DTRACE_CABAC_T( "\tEnd Code PartSize" );
  DTRACE_CABAC_T( "\teMode=" );
  DTRACE_CABAC_V( eMode )
  DTRACE_CABAC_T( "\teModeName=" );
  DTRACE_CABAC_T( eModeName )
  if (pcCU->isIntra( uiAbsPartIdx )) {
    DTRACE_CABAC_T( "\tpredMode=Intra\n" )
  }
  else {
    DTRACE_CABAC_T( "\tpredMode=Inter\n" )
  }
}

/** parse prediction mode
 * \param pcCU
 * \param uiAbsPartIdx 
 * \param uiDepth
 * \returns Void
 */
Void TDecSbac::parsePredMode( TComDataCU* pcCU, UInt uiAbsPartIdx, UInt uiDepth )
{
#if RWTH_RESI_REFINEMENT
  //todo: J: decode resi refinement mode (note: has to be before split flag)
  if ( getDecTop()->getMyCodingArgument()[INTER_LAYER_RWTH_RESI_REFINEMENT] )
  {
    if ( getDecTop()->getMyCodingArgument()[TEST_L1_INTRA_PREDICTION] &&
        !getDecTop()->getMyCodingArgument()[TEST_L1_INTER_PREDICTION] )
    {
      // Intra mode is allwoed in resirefinement but not Inter
      UInt uiSymbol;
      m_pcTDecBinIf->decodeBin( uiSymbol, m_cCUPredModeSCModel.get( 0, 0, 0 ) );
      DTRACE_CABAC_T( !uiSymbol ? "\tctx=0\tpred_mode_flag - PredMode=Intra\n" :"\tctx=0\tpred_mode_flag - PredMode=RESI_REFINE\n" );
      pcCU->setPredModeSubParts( !uiSymbol ? MODE_INTRA : MODE_IL_RWTH_RESI_REFINEMENT, uiAbsPartIdx, uiDepth);
    }
    else if (!getDecTop()->getMyCodingArgument()[TEST_L1_INTRA_PREDICTION] &&
              getDecTop()->getMyCodingArgument()[TEST_L1_INTER_PREDICTION] )
    {
      // Inter mode is allowed in resirefinement but not Intra
      UInt uiSymbol = 1;
      if (pcCU->getSlice()->getSliceType() != I_SLICE)
      {
        m_pcTDecBinIf->decodeBin( uiSymbol, m_cCUPredModeSCModel.get( 0, 0, 1 ) );
        DTRACE_CABAC_T( uiSymbol ? "\tctx=1\tpred_mode_flag - PredMode=RESI_REFINE\n" : "\tctx=1\tpred_mode_flag - PredMode=Inter\n" );
      }
      pcCU->setPredModeSubParts( uiSymbol ? MODE_IL_RWTH_RESI_REFINEMENT : MODE_INTER, uiAbsPartIdx, uiDepth);
    }
    else if (getDecTop()->getMyCodingArgument()[TEST_L1_INTRA_PREDICTION] &&
             getDecTop()->getMyCodingArgument()[TEST_L1_INTER_PREDICTION] )
    {
      // Intra and Inter are allowed in resirefinement
      UInt uiSymbol;
      m_pcTDecBinIf->decodeBin( uiSymbol, m_cCUPredModeSCModel.get( 0, 0, 0 ) );
      if (pcCU->getSlice()->getSliceType() == I_SLICE)
      {
        // Only Intra and refine allowed in if lower layer is I_SLICE
        DTRACE_CABAC_T( uiSymbol ? "\tctx=0\tpred_mode_flag - PredMode=Intra\n" : "\tctx=0\tpred_mode_flag - PredMode=RESI_REFINE\n" );
        pcCU->setPredModeSubParts( uiSymbol ? MODE_INTRA : MODE_IL_RWTH_RESI_REFINEMENT, uiAbsPartIdx, uiDepth);
      }
      else
      {
        // Inter, Intra and refine allowed
        DTRACE_CABAC_T( uiSymbol ? "\tctx=0\tpred_mode_flag - PredMode=Inter/Intra\n" : "\tctx=0\tpred_mode_flag - PredMode=RESI_REFINE\n" );
        if ( uiSymbol )
        {
          m_pcTDecBinIf->decodeBin( uiSymbol, m_cCUPredModeSCModel.get( 0, 0, 1 ) );
          DTRACE_CABAC_T( uiSymbol ? "\tctx=1\tpred_mode_flag - PredMode=Intra\n" : "\tctx=1\tpred_mode_flag - PredMode=Inter\n" );
          pcCU->setPredModeSubParts( uiSymbol ? MODE_INTRA : MODE_INTER , uiAbsPartIdx, uiDepth);
        }
        else
        {
          pcCU->setPredModeSubParts( MODE_IL_RWTH_RESI_REFINEMENT, uiAbsPartIdx, uiDepth);
        }
      }
    }
    else
    {
      // Only resi refinement mode is allowed. Nothing to code.
      pcCU->setPredModeSubParts( MODE_IL_RWTH_RESI_REFINEMENT , uiAbsPartIdx, uiDepth);
    }
    return;
  }
#endif

  if( pcCU->getSlice()->isIntra() )
  {
    pcCU->setPredModeSubParts( MODE_INTRA, uiAbsPartIdx, uiDepth );
    return;
  }
  
  UInt uiSymbol;
  Int  iPredMode = MODE_INTER;
  m_pcTDecBinIf->decodeBin( uiSymbol, m_cCUPredModeSCModel.get( 0, 0, 0 ) );
  DTRACE_CABAC_T( uiSymbol ? "\tpred_mode_flag\tPredMode=Intra\n" : "\tpred_mode_flag\tPredMode=Inter\n" );
  iPredMode += uiSymbol;
  pcCU->setPredModeSubParts( (PredMode)iPredMode, uiAbsPartIdx, uiDepth );
}

Void TDecSbac::parseIntraDirLumaAng  ( TComDataCU* pcCU, UInt absPartIdx, UInt depth )
{
  PartSize mode = pcCU->getPartitionSize( absPartIdx );
  UInt partNum = mode==SIZE_NxN?4:1;

#if ENC_DEC_TRACE
  DTRACE_CABAC_T( "\tStart Code IntraDirLumaAng" );
  DTRACE_CABAC_T( "\tuiAbsPartIdx=" );
  DTRACE_CABAC_V( absPartIdx )
  DTRACE_CABAC_T( "\tnumParts=" );
  DTRACE_CABAC_V( partNum )
  DTRACE_CABAC_N
  UInt uiInitialDir[4];
#endif
    
  UInt partOffset = ( pcCU->getPic()->getNumPartInCU() >> ( pcCU->getDepth(absPartIdx) << 1 ) ) >> 2;
  UInt mpmPred[4],symbol;
  Int j,intraPredMode;    
  if (mode==SIZE_NxN)
  {
    depth++;
  }
  for (j=0;j<partNum;j++)
  {
    m_pcTDecBinIf->decodeBin( symbol, m_cCUIntraPredSCModel.get( 0, 0, 0) );
    DTRACE_CABAC_T( "\tprev_intra_luma_pred_flag\n" );
    mpmPred[j] = symbol;
  }
  for (j=0;j<partNum;j++)
  {
    Int preds[3] = {-1, -1, -1};
    Int predNum = pcCU->getIntraDirLumaPredictor(absPartIdx+partOffset*j, preds);  
    if (mpmPred[j])
    {
      m_pcTDecBinIf->decodeBinEP( symbol );
      DTRACE_CABAC_N

      if (symbol)
      {
        m_pcTDecBinIf->decodeBinEP( symbol );
        DTRACE_CABAC_N

        symbol++;
      }
      intraPredMode = preds[symbol];
    }
    else
    {
      intraPredMode = 0;
      m_pcTDecBinIf->decodeBinsEP( symbol, 5 );
      intraPredMode = symbol;
        
      //postponed sorting of MPMs (only in remaining branch)
      if (preds[0] > preds[1])
      { 
        std::swap(preds[0], preds[1]); 
      }
      if (preds[0] > preds[2])
      {
        std::swap(preds[0], preds[2]);
      }
      if (preds[1] > preds[2])
      {
        std::swap(preds[1], preds[2]);
      }
      for ( Int i = 0; i < predNum; i++ )
      {
        intraPredMode += ( intraPredMode >= preds[i] );
      }
    }
    pcCU->setLumaIntraDirSubParts( (UChar)intraPredMode, absPartIdx+partOffset*j, depth );
#if ENC_DEC_TRACE
    uiInitialDir[j] = intraPredMode;
#endif
  }
#if ENC_DEC_TRACE
  DTRACE_CABAC_T( "\tEnd Code IntraDirLumaAng" );
  DTRACE_CABAC_T( "\tintraPredMode=" );
  DTRACE_CABAC_V( uiInitialDir[0] )
  if (partNum == 4)
  {
    DTRACE_CABAC_T( "," )
    DTRACE_CABAC_V( uiInitialDir[1] )
    DTRACE_CABAC_T( "," )
    DTRACE_CABAC_V( uiInitialDir[2] )
    DTRACE_CABAC_T( "," )
    DTRACE_CABAC_V( uiInitialDir[3] )
  }
  DTRACE_CABAC_N
#endif
}

Void TDecSbac::parseIntraDirChroma( TComDataCU* pcCU, UInt uiAbsPartIdx, UInt uiDepth )
{
  UInt uiSymbol;

  m_pcTDecBinIf->decodeBin( uiSymbol, m_cCUChromaPredSCModel.get( 0, 0, 0 ) );
  DTRACE_CABAC_T( "\tCode IntraDirChroma\n" );

  if( uiSymbol == 0 )
  {
    uiSymbol = DM_CHROMA_IDX;
  } 
  else 
  {
#if !REMOVE_LMCHROMA
    if( pcCU->getSlice()->getSPS()->getUseLMChroma() )
    {
      m_pcTDecBinIf->decodeBin( uiSymbol, m_cCUChromaPredSCModel.get( 0, 0, 1 ) );
      DTRACE_CABAC_T( "\tCode IntraDirChroma\n" );
    }
    else
    {
      uiSymbol = 1;
    }

    if( uiSymbol == 0 )
    {
      uiSymbol = LM_CHROMA_IDX;
    } 
    else
#endif
    {
      UInt uiIPredMode;
      m_pcTDecBinIf->decodeBinsEP( uiIPredMode, 2 );
      UInt uiAllowedChromaDir[ NUM_CHROMA_MODE ];
      pcCU->getAllowedChromaDir( uiAbsPartIdx, uiAllowedChromaDir );
      uiSymbol = uiAllowedChromaDir[ uiIPredMode ];
    }
  }
  pcCU->setChromIntraDirSubParts( uiSymbol, uiAbsPartIdx, uiDepth );
  return;
}

Void TDecSbac::parseInterDir( TComDataCU* pcCU, UInt& ruiInterDir, UInt uiAbsPartIdx, UInt uiDepth )
{
  UInt uiSymbol;
  const UInt uiCtx = pcCU->getCtxInterDir( uiAbsPartIdx );
  ContextModel *pCtx = m_cCUInterDirSCModel.get( 0 );
#if DISALLOW_BIPRED_IN_8x4_4x8PUS
  uiSymbol = 0;
  if (pcCU->getPartitionSize(uiAbsPartIdx) == SIZE_2Nx2N || pcCU->getHeight(uiAbsPartIdx) != 8 )
  {
#endif
    m_pcTDecBinIf->decodeBin( uiSymbol, *( pCtx + uiCtx ) );
    DTRACE_CABAC_T( "\tCode InterDir\n" );
#if DISALLOW_BIPRED_IN_8x4_4x8PUS
  }
#endif

  if( uiSymbol )
  {
    uiSymbol = 2;
  }
  else
  {
    m_pcTDecBinIf->decodeBin( uiSymbol, *( pCtx + 4 ) );
    DTRACE_CABAC_T( "\tCode InterDir\n" );
    assert(uiSymbol == 0 || uiSymbol == 1);
  }

  uiSymbol++;
  ruiInterDir = uiSymbol;
  return;
}

Void TDecSbac::parseRefFrmIdx( TComDataCU* pcCU, Int& riRefFrmIdx, UInt uiAbsPartIdx, UInt uiDepth, RefPicList eRefList )
{
  UInt uiSymbol;
  {
    ContextModel *pCtx = m_cCURefPicSCModel.get( 0 );
    m_pcTDecBinIf->decodeBin( uiSymbol, *pCtx );
    DTRACE_CABAC_T( "\tCode RefFrmIdx\n" );

    if( uiSymbol )
    {
#if REF_IDX_BYPASS
      UInt uiRefNum = pcCU->getSlice()->getNumRefIdx( eRefList ) - 2;
      pCtx++;
      UInt ui;
      for( ui = 0; ui < uiRefNum; ++ui )
      {
        if( ui == 0 )
        {
          m_pcTDecBinIf->decodeBin( uiSymbol, *pCtx );
        }
        else
        {
          m_pcTDecBinIf->decodeBinEP( uiSymbol );
        }
        DTRACE_CABAC_T( "\tCode RefFrmIdx\n" );

        if( uiSymbol == 0 )
        {
          break;
        }
      }
      uiSymbol = ui + 1;
#else
      xReadUnaryMaxSymbol( uiSymbol, pCtx + 1, 1, pcCU->getSlice()->getNumRefIdx( eRefList )-2 );
      uiSymbol++;
#endif
    }
    riRefFrmIdx = uiSymbol;
  }

  return;
}

Void TDecSbac::parseMvd( TComDataCU* pcCU, UInt uiAbsPartIdx, UInt uiPartIdx, UInt uiDepth, RefPicList eRefList )
{
  DTRACE_CABAC_T( "\tStart Code Mvd" );
  DTRACE_CABAC_T( "\tuiAbsPartIdx=" );
  DTRACE_CABAC_V( uiAbsPartIdx )
  DTRACE_CABAC_T( "\teRefList=" );
  DTRACE_CABAC_V( eRefList )
  DTRACE_CABAC_N

  UInt uiSymbol;
  UInt uiHorAbs;
  UInt uiVerAbs;
  UInt uiHorSign = 0;
  UInt uiVerSign = 0;
  ContextModel *pCtx = m_cCUMvdSCModel.get( 0 );

  if(pcCU->getSlice()->getMvdL1ZeroFlag() && eRefList == REF_PIC_LIST_1 && pcCU->getInterDir(uiAbsPartIdx)==3)
  {
    uiHorAbs=0;
    uiVerAbs=0;
  }
  else
  {
    m_pcTDecBinIf->decodeBin( uiHorAbs, *pCtx );
    DTRACE_CABAC_N
    m_pcTDecBinIf->decodeBin( uiVerAbs, *pCtx );
    DTRACE_CABAC_N

    const Bool bHorAbsGr0 = uiHorAbs != 0;
    const Bool bVerAbsGr0 = uiVerAbs != 0;
    pCtx++;

    if( bHorAbsGr0 )
    {
      m_pcTDecBinIf->decodeBin( uiSymbol, *pCtx );
      DTRACE_CABAC_N
      uiHorAbs += uiSymbol;
    }

    if( bVerAbsGr0 )
    {
      m_pcTDecBinIf->decodeBin( uiSymbol, *pCtx );
      DTRACE_CABAC_N
      uiVerAbs += uiSymbol;
    }

    if( bHorAbsGr0 )
    {
      if( 2 == uiHorAbs )
      {
        xReadEpExGolomb( uiSymbol, 1 );
        uiHorAbs += uiSymbol;
      }

      m_pcTDecBinIf->decodeBinEP( uiHorSign );
      DTRACE_CABAC_N
    }

    if( bVerAbsGr0 )
    {
      if( 2 == uiVerAbs )
      {
        xReadEpExGolomb( uiSymbol, 1 );
        uiVerAbs += uiSymbol;
      }

      m_pcTDecBinIf->decodeBinEP( uiVerSign );
      DTRACE_CABAC_N
    }

  }

  const TComMv cMv( uiHorSign ? -Int( uiHorAbs ): uiHorAbs, uiVerSign ? -Int( uiVerAbs ) : uiVerAbs );
  pcCU->getCUMvField( eRefList )->setAllMvd( cMv, pcCU->getPartitionSize( uiAbsPartIdx ), uiAbsPartIdx, uiDepth, uiPartIdx );
  
  DTRACE_CABAC_T( "\tEnd Code Mvd" )
  DTRACE_CABAC_T( "\tiHor=" );
  DTRACE_CABAC_V( cMv.getHor() )
  DTRACE_CABAC_T( "\tiVer=" );
  DTRACE_CABAC_V( cMv.getVer() )
  DTRACE_CABAC_N
}


Void TDecSbac::parseTransformSubdivFlag( UInt& ruiSubdivFlag, UInt uiLog2TransformBlockSize )
{
  m_pcTDecBinIf->decodeBin( ruiSubdivFlag, m_cCUTransSubdivFlagSCModel.get( 0, 0, uiLog2TransformBlockSize ) );
  DTRACE_CABAC_T( "\tCode split_transform_flag" )
  DTRACE_CABAC_T( "\tctx=" )
  DTRACE_CABAC_V( uiLog2TransformBlockSize )
  DTRACE_CABAC_N
}

Void TDecSbac::parseQtRootCbf( TComDataCU* pcCU, UInt uiAbsPartIdx, UInt uiDepth, UInt& uiQtRootCbf )
{
  UInt uiSymbol;
  const UInt uiCtx = 0;
  m_pcTDecBinIf->decodeBin( uiSymbol , m_cCUQtRootCbfSCModel.get( 0, 0, uiCtx ) );
  DTRACE_CABAC_T( "\tno_residual_data_flag\tctx=" )
  DTRACE_CABAC_V( uiCtx )
  DTRACE_CABAC_T( "\tuiAbsPartIdx=" )
  DTRACE_CABAC_V( uiAbsPartIdx )
  DTRACE_CABAC_N
  
  uiQtRootCbf = uiSymbol;
}

Void TDecSbac::parseDeltaQP( TComDataCU* pcCU, UInt uiAbsPartIdx, UInt uiDepth )
{
  Int qp;
  UInt uiDQp;
  Int  iDQp;
  
#if CU_DQP_TU_EG
  UInt uiSymbol;

  xReadUnaryMaxSymbol (uiDQp,  &m_cCUDeltaQpSCModel.get( 0, 0, 0 ), 1, CU_DQP_TU_CMAX);

  if( uiDQp >= CU_DQP_TU_CMAX)
  {
    xReadEpExGolomb( uiSymbol, CU_DQP_EG_k );
    uiDQp+=uiSymbol;
  }

  if ( uiDQp > 0 )
  {
    UInt uiSign;
    Int qpBdOffsetY = pcCU->getSlice()->getSPS()->getQpBDOffsetY();
    m_pcTDecBinIf->decodeBinEP(uiSign);
    DTRACE_CABAC_T( "\tDeltaQP\n" )
    iDQp = uiDQp;
    if(uiSign)
    {
      iDQp = -iDQp;
    }
    qp = (((Int) pcCU->getRefQP( uiAbsPartIdx ) + iDQp + 52 + 2*qpBdOffsetY )%(52+qpBdOffsetY)) - qpBdOffsetY;
  }
  else 
  {
    iDQp=0;
    qp = pcCU->getRefQP(uiAbsPartIdx);
  }
#else
  m_pcTDecBinIf->decodeBin( uiDQp, m_cCUDeltaQpSCModel.get( 0, 0, 0 ) );
  DTRACE_CABAC_T( "\tDeltaQP\n" )
  
  if ( uiDQp == 0 )
  {
    qp = pcCU->getRefQP(uiAbsPartIdx);
  }
  else
  {
    UInt uiSign;
    Int qpBdOffsetY = pcCU->getSlice()->getSPS()->getQpBDOffsetY();
    m_pcTDecBinIf->decodeBinEP(uiSign);
    DTRACE_CABAC_T( "\tDeltaQP\n" )

    UInt uiMaxAbsDQpMinus1 = 24 + (qpBdOffsetY/2) + (uiSign);
    UInt uiAbsDQpMinus1;
    xReadUnaryMaxSymbol (uiAbsDQpMinus1,  &m_cCUDeltaQpSCModel.get( 0, 0, 1 ), 1, uiMaxAbsDQpMinus1);

    iDQp = uiAbsDQpMinus1 + 1;

    if(uiSign)
    {
      iDQp = -iDQp;
    }

    qp = (((Int) pcCU->getRefQP( uiAbsPartIdx ) + iDQp + 52 + 2*qpBdOffsetY )%(52+qpBdOffsetY)) - qpBdOffsetY;
  }
#endif
  pcCU->setQPSubParts(qp, uiAbsPartIdx, uiDepth);  
  pcCU->setCodedQP(qp);
}

Void TDecSbac::parseQtCbf( TComDataCU* pcCU, UInt uiAbsPartIdx, TextType eType, UInt uiTrDepth, UInt uiDepth )
{
  UInt uiSymbol;
  const UInt uiCtx = pcCU->getCtxQtCbf( uiAbsPartIdx, eType, uiTrDepth );
  m_pcTDecBinIf->decodeBin( uiSymbol , m_cCUQtCbfSCModel.get( 0, eType ? TEXT_CHROMA: eType, uiCtx ) );

#if ENC_DEC_TRACE
  if ( eType == TEXT_LUMA )
  {
    DTRACE_CABAC_T( "\tcbf_luma" )
  }
  else if ( eType == TEXT_CHROMA_U )
  {
    DTRACE_CABAC_T( "\tcbf_cb" )
  }
  else 
  {
    DTRACE_CABAC_T( "\tcbf_cr" )
  }
  DTRACE_CABAC_T( "\tctx=" )
  DTRACE_CABAC_V( uiCtx )
  DTRACE_CABAC_T( "\tetype=" )
  DTRACE_CABAC_V( eType )
  DTRACE_CABAC_T( "\tuiAbsPartIdx=" )
  DTRACE_CABAC_V( uiAbsPartIdx )
  DTRACE_CABAC_N
#endif
  
  pcCU->setCbfSubParts( uiSymbol << uiTrDepth, eType, uiAbsPartIdx, uiDepth );
}

void TDecSbac::parseTransformSkipFlags (TComDataCU* pcCU, UInt uiAbsPartIdx, UInt width, UInt height, UInt uiDepth, TextType eTType)
{
  if (pcCU->getCUTransquantBypass(uiAbsPartIdx))
  {
    return;
  }
#if !INTER_TRANSFORMSKIP
  if(!pcCU->isIntra(uiAbsPartIdx))
  {
    return;
  }
#endif
  if(width != 4 || height != 4)
  {
    return;
  }
  
  UInt useTransformSkip;
  m_pcTDecBinIf->decodeBin( useTransformSkip , m_cTransformSkipSCModel.get( 0, eTType? TEXT_CHROMA: TEXT_LUMA, 0 ) );
  DTRACE_CABAC_T("\tTransformSkipFlag");
  DTRACE_CABAC_T( "\tsymbol=" )
  DTRACE_CABAC_V( useTransformSkip )
  DTRACE_CABAC_T( "\tAddr=" )
  DTRACE_CABAC_V( pcCU->getAddr() )
  DTRACE_CABAC_T( "\tetype=" )
  DTRACE_CABAC_V( eTType )
  DTRACE_CABAC_T( "\tuiAbsPartIdx=" )
  DTRACE_CABAC_V( uiAbsPartIdx )
  DTRACE_CABAC_T( "\n" )

  if(eTType!= TEXT_LUMA)
  {
    const UInt uiLog2TrafoSize = g_aucConvertToBit[pcCU->getSlice()->getSPS()->getMaxCUWidth()] + 2 - uiDepth;
    if(uiLog2TrafoSize == 2) 
    { 
      uiDepth --;
    }
  }
  pcCU->setTransformSkipSubParts( useTransformSkip, eTType, uiAbsPartIdx, uiDepth);
}

/** Parse (X,Y) position of the last significant coefficient
 * \param uiPosLastX reference to X component of last coefficient
 * \param uiPosLastY reference to Y component of last coefficient
 * \param width  Block width
 * \param height Block height
 * \param eTType plane type / luminance or chrominance
 * \param uiScanIdx scan type (zig-zag, hor, ver)
 *
 * This method decodes the X and Y component within a block of the last significant coefficient.
 */
Void TDecSbac::parseLastSignificantXY( UInt& uiPosLastX, UInt& uiPosLastY, Int width, Int height, TextType eTType, UInt uiScanIdx )
{
  DTRACE_CABAC_T( "\tStart Coding LastSignificantXY\n" )

  UInt uiLast;
  ContextModel *pCtxX = m_cCuCtxLastX.get( 0, eTType );
  ContextModel *pCtxY = m_cCuCtxLastY.get( 0, eTType );

  Int blkSizeOffsetX, blkSizeOffsetY, shiftX, shiftY;
  blkSizeOffsetX = eTType ? 0: (g_aucConvertToBit[ width ] *3 + ((g_aucConvertToBit[ width ] +1)>>2));
  blkSizeOffsetY = eTType ? 0: (g_aucConvertToBit[ height ]*3 + ((g_aucConvertToBit[ height ]+1)>>2));
  shiftX= eTType ? g_aucConvertToBit[ width  ] :((g_aucConvertToBit[ width  ]+3)>>2);
  shiftY= eTType ? g_aucConvertToBit[ height ] :((g_aucConvertToBit[ height ]+3)>>2);
  // posX
  for( uiPosLastX = 0; uiPosLastX < g_uiGroupIdx[ width - 1 ]; uiPosLastX++ )
  {
    m_pcTDecBinIf->decodeBin( uiLast, *( pCtxX + blkSizeOffsetX + (uiPosLastX >> shiftX) ) );
    DTRACE_CABAC_T( "\tctx:" )
    DTRACE_CABAC_V( blkSizeOffsetX + (uiPosLastX >>shiftX) )
    DTRACE_CABAC_T( "\tlast_significant_coeff_x_prefix\n" )

    if( !uiLast )
    {
      break;
    }
  }

  // posY
  for( uiPosLastY = 0; uiPosLastY < g_uiGroupIdx[ height - 1 ]; uiPosLastY++ )
  {
    m_pcTDecBinIf->decodeBin( uiLast, *( pCtxY + blkSizeOffsetY + (uiPosLastY >>shiftY)) );
    DTRACE_CABAC_T( "\tctx:" )
    DTRACE_CABAC_V( blkSizeOffsetY + (uiPosLastY >>shiftY) )
    DTRACE_CABAC_T( "\tlast_significant_coeff_y_prefix\n" )

    if( !uiLast )
    {
      break;
    }
  }
  if ( uiPosLastX > 3 )
  {
    UInt uiTemp  = 0;
    UInt uiCount = ( uiPosLastX - 2 ) >> 1;
    for ( Int i = uiCount - 1; i >= 0; i-- )
    {
      m_pcTDecBinIf->decodeBinEP( uiLast );
      DTRACE_CABAC_T( "\tlast_significant_coeff_x_suffix\n" )
      uiTemp += uiLast << i;
    }
    uiPosLastX = g_uiMinInGroup[ uiPosLastX ] + uiTemp;
  }
  if ( uiPosLastY > 3 )
  {
    UInt uiTemp  = 0;
    UInt uiCount = ( uiPosLastY - 2 ) >> 1;
    for ( Int i = uiCount - 1; i >= 0; i-- )
    {
      m_pcTDecBinIf->decodeBinEP( uiLast );
      DTRACE_CABAC_T( "\tlast_significant_coeff_y_suffix\n" )
      uiTemp += uiLast << i;
    }
    uiPosLastY = g_uiMinInGroup[ uiPosLastY ] + uiTemp;
  }

  if( uiScanIdx == SCAN_VER )
  {
    swap( uiPosLastX, uiPosLastY );
  }

  DTRACE_CABAC_T( "\tEnd LastSignificantXY: (" )
  DTRACE_CABAC_V( uiPosLastX )
  DTRACE_CABAC_T( "," )
  DTRACE_CABAC_V( uiPosLastY )
  DTRACE_CABAC_T( ")\n" )
}

#if RWTH_CODE_LASTXY_DIFF
/** Parse difference position of the last significant coefficient
 * \param uiDiff Diff of last coefficient
 * \param uiRemaningLenth  The remaining length
 * \param eTType plane type / luminance or chrominance
 * This method parses the difference to the lower layer significant position.
 */
Void TDecSbac::parseLastSignificantDiff( UInt& uiPosX, UInt uiPosXRef, UInt& uiPosY, UInt uiPosYRef, UInt uiWidth, UInt uiHeight, TextType eTType, UInt uiScanIdx, const UInt* scan )
{
#if ENC_DEC_TRACE
  DTRACE_CABAC_T( "\tStart Coding LastSignificantXYDiff" )
  DTRACE_CABAC_N
#endif

  const UInt uiLog2BlockSize = g_aucConvertToBit[ uiWidth ] + 2;
  if ( uiScanIdx == SCAN_HOR )
  {
    swap( uiPosXRef, uiPosYRef );
    swap( uiWidth, uiHeight );
  } 

  ContextModel *pCtxSignalling = m_cCuCtxLastDiffSignalling.get( 0, eTType );
  ContextModel *pCtxX = m_cCuCtxLastDiffX.get( 0, eTType );
  ContextModel *pCtxY = m_cCuCtxLastDiffY.get( 0, eTType );

  Int uiRemainingWidth  = uiWidth;
  Int uiRemainingHeight = uiHeight;
  
  UInt uiSymbol;
  Bool bBothDeltaGreaterZero = true;
  Bool bDeltaXSmallerZero;
  UInt uiXLimit = 0;
  UInt uiYLimit = 0;

  UInt uiAbsDeltaX, uiAbsDeltaY;

  // Code (iPosXDelta>=0 && iPosYDelta>=0)
  // In the following cases the flag can be inferred:
  // - if reference pos is (0,0). In this case both deltas are positive
  // - if SCAN_DIAG is used and refPos is one of the last two coefficients in scanning order
  // - if SCAN_DIAG is used and refPos is the first coefficient in the last 4x4 block
  // - if SCAN_VER (or swapped SCAN_HOR) is used and refPosY is 0 and refPosX is a multiple of 4
  // - if SCAN_VER (or swapped SCAN_HOR) is used and refPosX is uiWidth-4 and refPosY is a multiple of 4
  // - if SCAN_VER (or swapped SCAN_HOR) is used and refPos is in the last 4x4 block and refPosX is uiWidth-1
  // - if SCAN_VER (or swapped SCAN_HOR) is used and refPos is in the last 4x4 block and refPosY is uiHeight-4
  if ( 
       ( uiPosXRef == 0 && uiPosYRef == 0 ) ||
       ( uiScanIdx == SCAN_DIAG && uiPosXRef == uiWidth-1 && uiPosYRef >= uiHeight-2 ) ||
       ( uiScanIdx == SCAN_DIAG && uiPosXRef == uiWidth-4 && uiPosYRef == uiHeight-4 ) || 
       ( (uiScanIdx == SCAN_HOR || uiScanIdx == SCAN_VER) && uiPosYRef == 0 && uiPosXRef % 4 == 0 ) ||
       ( (uiScanIdx == SCAN_HOR || uiScanIdx == SCAN_VER) && uiPosXRef == uiWidth-4 && uiPosYRef % 4 == 0 ) ||
       ( (uiScanIdx == SCAN_HOR || uiScanIdx == SCAN_VER) && uiPosXRef == uiWidth-1 && uiPosYRef >= uiHeight-4 ) ||
       ( (uiScanIdx == SCAN_HOR || uiScanIdx == SCAN_VER) && uiPosYRef == uiHeight-4 && uiPosXRef >= uiWidth-4 )
    )
  {
    // last_significant_diff_x>=0_&&_y>=0 has to be true
    bBothDeltaGreaterZero = true;
  }
  else
  {
    UInt ctx = ( uiPosXRef >= (uiWidth/4) && uiPosXRef < (uiWidth/4*3) ) ? 0 : 1;
    m_pcTDecBinIf->decodeBin( uiSymbol, *( pCtxSignalling + ctx ) );
    DTRACE_CABAC_T( "\tctx:" )
    DTRACE_CABAC_V( ctx )
    DTRACE_CABAC_T( "\tlast_significant_diff_x>=0_&&_y>=0\n" )
    bBothDeltaGreaterZero = uiSymbol ? true : false;
  }
       
  if (bBothDeltaGreaterZero)
  {
    uiRemainingWidth = uiWidth - uiPosXRef;
    xParseLastDiffPrefixSuffix( uiAbsDeltaX, uiRemainingWidth, m_pcTDecBinIf, pCtxX, eTType, "deltaX" );
    uiPosX = uiPosXRef + uiAbsDeltaX;
    uiRemainingHeight = uiHeight - uiPosYRef;
    xParseLastDiffPrefixSuffix( uiAbsDeltaY, uiRemainingHeight, m_pcTDecBinIf, pCtxY, eTType, "deltaY" );
    uiPosY = uiPosYRef + uiAbsDeltaY;
  }
  else
  {
    // encode if iPosXDelta<0 or iPosYDelta<0;
    // In the following cases iPosXDelta<0 can be inferred:
    // - if uiPosXRef==0 or uiPosYRef==0.
    // - if uiPosXRef==uiWidth-1
    // - if SCAN_DIAG is used and uiPosYRef >= uiHeight-2
    // - if SCAN_DIAG is used and uiPosYRef+8 >= uiHeight and uiPosXRef is a multiple of 4
    // - if SCAN_VER (or swapped SCAN_HOR) is used and uiPosXRef is a multiple of 4
    // - if SCAN_VER (or swapped SCAN_HOR) is used and uiPosXRef >= uiWidth-4 and uiPosYRef is a multiple of 4
    
    if ( ( uiPosXRef==0 || uiPosYRef==0 ) ||
         ( uiPosXRef==uiWidth-1 ) || 
         ( uiScanIdx == SCAN_DIAG && uiPosYRef >= uiHeight-2) || 
         ( uiScanIdx == SCAN_DIAG && uiPosYRef+8 >= uiHeight && (uiPosXRef % 4) == 0 ) ||
         ( (uiScanIdx == SCAN_VER || uiScanIdx == SCAN_HOR ) && (uiPosXRef % 4) == 0 ) ||
         ( (uiScanIdx == SCAN_VER || uiScanIdx == SCAN_HOR ) && uiPosYRef >= uiWidth-4 && (uiPosXRef % 4) == 0 )
       )
    {
      if ( uiPosYRef==0 || uiPosXRef==uiWidth-1 )
      {
        // last_significant_diff_x<0 has to be true
        bDeltaXSmallerZero = true;
      }
      else
      {
        // last_significant_diff_x<0 has to be false
        bDeltaXSmallerZero = false;
      }
    }
    else
    {
      UInt ctx = 2;
      m_pcTDecBinIf->decodeBin( uiSymbol, *( pCtxSignalling + ctx ) );
      DTRACE_CABAC_T( "\tctx:" )
      DTRACE_CABAC_V( ctx )
      DTRACE_CABAC_T( "\tlast_significant_diff_x<0\n" )
      bDeltaXSmallerZero = uiSymbol ? true : false;
    }

    if (uiScanIdx == SCAN_DIAG)
    {
      // Scan again to determine the limits of deltaX and deltaY
      Int iScanPos = -1;
      Int iScanPosY, iScanPosX, posRef;
      if (bDeltaXSmallerZero)
      {
        // iPosXDelta<0 and iPosYDelta>0
        // coded x value limits the possible values of y (y value depends on x)
        do
        {
          posRef = scan[ ++iScanPos ];
          iScanPosY = posRef >> uiLog2BlockSize;
          iScanPosX = posRef - ( iScanPosY << uiLog2BlockSize );

          if (iScanPosY == uiHeight-1 && iScanPosX >= uiXLimit)
          {
            uiXLimit = iScanPosX + 1;
          }
        }
        while ( ! (iScanPosX==uiPosXRef && iScanPosY==uiPosYRef) );
        uiRemainingWidth  = uiPosXRef - uiXLimit;
        // Decode uiAbsDeltaX
        xParseLastDiffPrefixSuffix( uiAbsDeltaX, uiRemainingWidth, m_pcTDecBinIf, pCtxX, eTType, "deltaX" );
        uiPosX = uiPosXRef - uiAbsDeltaX - 1;

        // Scan again to get the y limit
        iScanPos = -1;
        do
        {
          posRef = scan[ ++iScanPos ];
          iScanPosY = posRef >> uiLog2BlockSize;
          iScanPosX = posRef - ( iScanPosY << uiLog2BlockSize );

          if (iScanPosX == uiPosX && iScanPosY >= uiYLimit)
          {
            uiYLimit = iScanPosY + 1;
          }
        }
        while ( ! (iScanPosX==uiPosXRef && iScanPosY==uiPosYRef) );
        uiRemainingHeight = uiHeight - uiYLimit;
        // Decode uiAbsDeltaY
        xParseLastDiffPrefixSuffix( uiAbsDeltaY, uiRemainingHeight, m_pcTDecBinIf, pCtxY, eTType, "deltaY" );
        uiPosY = uiYLimit + uiAbsDeltaY;
      }
      else
      {
        // iPosYDelta<0 and iPosXDelta>0
        // coded y value limits the possible values of x (x value depends on y)
        do
        {
          posRef = scan[ ++iScanPos ];
          iScanPosY = posRef >> uiLog2BlockSize;
          iScanPosX = posRef - ( iScanPosY << uiLog2BlockSize );

          if (iScanPosX == uiWidth-1 && iScanPosY >= uiYLimit)
          {
            uiYLimit = iScanPosY + 1;
          }
        }
        while ( ! (iScanPosX==uiPosXRef && iScanPosY==uiPosYRef) );
        uiRemainingHeight = uiPosYRef - uiYLimit;
        // Decode uiAbsDeltaY
        xParseLastDiffPrefixSuffix( uiAbsDeltaY, uiRemainingHeight, m_pcTDecBinIf, pCtxY, eTType, "deltaY" );
        uiPosY = uiPosYRef - uiAbsDeltaY - 1;
        
        // Scan again to get the x limit
        iScanPos = -1;
        do
        {
          posRef = scan[ ++iScanPos ];
          iScanPosY = posRef >> uiLog2BlockSize;
          iScanPosX = posRef - ( iScanPosY << uiLog2BlockSize );

          if (iScanPosY == uiPosY && iScanPosX >= uiXLimit)
          {
            uiXLimit = iScanPosX + 1;
          }
        }
        while ( ! (iScanPosX==uiPosXRef && iScanPosY==uiPosYRef) );
        uiRemainingWidth  = uiWidth - uiXLimit;
        // Decode uiAbsDeltaX
        xParseLastDiffPrefixSuffix( uiAbsDeltaX, uiRemainingWidth, m_pcTDecBinIf, pCtxX, eTType, "deltaX" );
        uiPosX = uiXLimit + uiAbsDeltaX;
      }
    }
    else
    {
      // SCAN_VER or swapped SCAN_HOR
      UInt uiNum4x4BlocksStartedY = ((uiPosYRef / 4) + 1);
      if ( bDeltaXSmallerZero )
      {
        uiRemainingWidth  = uiPosXRef;
        uiRemainingHeight = uiHeight - uiNum4x4BlocksStartedY * 4;
      }
      else
      {
        uiRemainingWidth  = uiWidth - uiPosXRef;
        uiRemainingHeight = uiPosYRef;
      }

      xParseLastDiffPrefixSuffix( uiAbsDeltaX, uiRemainingWidth, m_pcTDecBinIf, pCtxX, eTType, "deltaX" );
      uiPosX = bDeltaXSmallerZero ? uiPosXRef - uiAbsDeltaX - 1 : uiPosXRef + uiAbsDeltaX;
      xParseLastDiffPrefixSuffix( uiAbsDeltaY, uiRemainingHeight, m_pcTDecBinIf, pCtxY, eTType, "deltaY" );
      uiPosY = bDeltaXSmallerZero ? uiNum4x4BlocksStartedY * 4 + uiAbsDeltaY : uiPosYRef - uiAbsDeltaY - 1;
    }
  }
 
  if ( uiScanIdx == SCAN_HOR )
  {
    swap( uiPosX, uiPosY );
  } 

#if ENC_DEC_TRACE
  if ( uiScanIdx == SCAN_HOR )
  {
    swap( uiPosXRef, uiPosYRef );
  } 
  DTRACE_CABAC_T( "\tLastSignificantXYDiff: (" )
  DTRACE_CABAC_V( uiPosX )
  DTRACE_CABAC_T( "," )
  DTRACE_CABAC_V( uiPosY )
  DTRACE_CABAC_T( ")\t Ref(" )
  DTRACE_CABAC_V( uiPosXRef )
  DTRACE_CABAC_T( "," )
  DTRACE_CABAC_V( uiPosYRef )
  DTRACE_CABAC_T( ")" )
  DTRACE_CABAC_N
#endif
}

Void TDecSbac::xParseLastDiffPrefixSuffix( UInt &uiAbsDelta, UInt uiRemaining, TDecBinIf* pcBinIf, ContextModel *pCtx, TextType eTType, const char *cTraceText )
{
  if (uiRemaining <= 1)
  {
    // Only one or zero possible value ... nothing to code
    uiAbsDelta = 0;
    return;
  }

  // delta groud ID
  UInt uiLast;
  // Round uiRemaining up to the next possible block width/height (4,8,16,32,64)
  UInt uiRemainingRound = (uiRemaining > 32) ? 64 :
                          (uiRemaining > 16) ? 32 :
                          (uiRemaining > 8) ? 16 :
                          (uiRemaining > 4) ? 8 : 4;
  Int blkSizeOffset = eTType ? 0: (g_aucConvertToBit[ uiRemainingRound ] *3 + ((g_aucConvertToBit[ uiRemainingRound ] +1)>>2));
  Int shift = eTType ? g_aucConvertToBit[ uiRemainingRound ] :((g_aucConvertToBit[ uiRemainingRound ]+3)>>2);
  
  for( uiAbsDelta = 0; uiAbsDelta < g_uiGroupIdx[ uiRemaining - 1 ]; uiAbsDelta++ )
  {
    pcBinIf->decodeBin( uiLast, *( pCtx + blkSizeOffset + (uiAbsDelta >> shift ) ) );
    DTRACE_CABAC_T( "\tctx:" )
    DTRACE_CABAC_V( blkSizeOffset + (uiAbsDelta >> shift ) )
    DTRACE_CABAC_T( "\tlast_significant_coeff_" )
    DTRACE_CABAC_T( cTraceText )
    DTRACE_CABAC_T( "_prefix\n" )
    if( !uiLast )
    {
      break;
    }
  }
  // Suffix
  if ( uiAbsDelta > 3 )
  {
    UInt uiTemp  = 0;
    UInt uiCount = ( uiAbsDelta - 2 ) >> 1;
    if ( uiAbsDelta == g_uiGroupIdx[uiRemaining-1] )
    {
      uiCount = g_uiLog2Round[uiRemaining - g_uiMinInGroup[ uiAbsDelta ]];
    }
    for ( Int i = uiCount - 1; i >= 0; i-- )
    {
      pcBinIf->decodeBinEP( uiLast );
      DTRACE_CABAC_T( "\tlast_significant_coeff_" )
      DTRACE_CABAC_T( cTraceText )
      DTRACE_CABAC_T( "_suffix\n" )
      uiTemp += uiLast << i;
    }
    uiAbsDelta = g_uiMinInGroup[ uiAbsDelta ] + uiTemp;
  }
}
#endif

Void TDecSbac::parseCoeffNxN( TComDataCU* pcCU, TCoeff* pcCoef, UInt uiAbsPartIdx, UInt uiWidth, UInt uiHeight, UInt uiDepth, TextType eTType )
{
  DTRACE_CABAC_T( "\tStart Code CoeffNxN" )
  DTRACE_CABAC_T( "\teType=" )
  DTRACE_CABAC_V( eTType )
  DTRACE_CABAC_T( "\twidth=" )
  DTRACE_CABAC_V( uiWidth )
  DTRACE_CABAC_T( "\theight=" )
  DTRACE_CABAC_V( uiHeight )
  DTRACE_CABAC_T( "\tdepth=" )
  DTRACE_CABAC_V( uiDepth )
  DTRACE_CABAC_T( "\tabspartidx=" )
  DTRACE_CABAC_V( uiAbsPartIdx )
  DTRACE_CABAC_T( "\ttoCU-X=" )
  DTRACE_CABAC_V( pcCU->getCUPelX() )
  DTRACE_CABAC_T( "\ttoCU-Y=" )
  DTRACE_CABAC_V( pcCU->getCUPelY() )
  DTRACE_CABAC_T( "\tCU-addr=" )
  DTRACE_CABAC_V(  pcCU->getAddr() )
  DTRACE_CABAC_T( "\tinCU-X=" )
  DTRACE_CABAC_V( g_auiRasterToPelX[ g_auiZscanToRaster[uiAbsPartIdx] ] )
  DTRACE_CABAC_T( "\tinCU-Y=" )
  DTRACE_CABAC_V( g_auiRasterToPelY[ g_auiZscanToRaster[uiAbsPartIdx] ] )
  DTRACE_CABAC_T( "\tpredmode=" )
  DTRACE_CABAC_V(  pcCU->getPredictionMode( uiAbsPartIdx ) )
  DTRACE_CABAC_N
  
  if( uiWidth > pcCU->getSlice()->getSPS()->getMaxTrSize() )
  {
    uiWidth  = pcCU->getSlice()->getSPS()->getMaxTrSize();
    uiHeight = pcCU->getSlice()->getSPS()->getMaxTrSize();
  }
#if PPS_TS_FLAG
  if(pcCU->getSlice()->getPPS()->getUseTransformSkip())
#else
  if(pcCU->getSlice()->getSPS()->getUseTransformSkip())
#endif
  {
    parseTransformSkipFlags( pcCU, uiAbsPartIdx, uiWidth, uiHeight, uiDepth, eTType);
  }

  eTType = eTType == TEXT_LUMA ? TEXT_LUMA : ( eTType == TEXT_NONE ? TEXT_NONE : TEXT_CHROMA );
  
  //----- parse significance map -----
  const UInt  uiLog2BlockSize   = g_aucConvertToBit[ uiWidth ] + 2;
  const UInt  uiMaxNumCoeff     = uiWidth * uiHeight;
  const UInt  uiMaxNumCoeffM1   = uiMaxNumCoeff - 1;
  UInt uiScanIdx = pcCU->getCoefScanIdx(uiAbsPartIdx, uiWidth, eTType==TEXT_LUMA, pcCU->isIntra(uiAbsPartIdx));
  int blockType = uiLog2BlockSize;
  if (uiWidth != uiHeight)
  {
    uiScanIdx = SCAN_DIAG;
    blockType = 4;
  }
  
  //===== decode last significant =====
  UInt uiPosLastX, uiPosLastY;
  parseLastSignificantXY( uiPosLastX, uiPosLastY, uiWidth, uiHeight, eTType, uiScanIdx );
  UInt uiBlkPosLast      = uiPosLastX + (uiPosLastY<<uiLog2BlockSize);
  pcCoef[ uiBlkPosLast ] = 1;

  //===== decode significance flags =====
  UInt uiScanPosLast   = uiBlkPosLast;
  if (uiScanIdx == SCAN_ZIGZAG)
  {
    // Map zigzag to diagonal scan
    uiScanIdx = SCAN_DIAG;
  }
#if REMOVE_NSQT
  const UInt *scan = g_auiSigLastScan[ uiScanIdx ][ uiLog2BlockSize-1 ];
#else
  const UInt * scan;
  if (uiWidth == uiHeight)
  {
    scan = g_auiSigLastScan[ uiScanIdx ][ uiLog2BlockSize-1 ];
  }
  else
  {
    scan = g_sigScanNSQT[ uiLog2BlockSize - 2 ];
  }
#endif
  for( uiScanPosLast = 0; uiScanPosLast < uiMaxNumCoeffM1; uiScanPosLast++ )
  {
    UInt uiBlkPos = scan[ uiScanPosLast ];
    if( uiBlkPosLast == uiBlkPos )
    {
      break;
    }
  }

#if ENC_DEC_TRACE
  switch (uiScanIdx)
  {
    case 0: DTRACE_CABAC_T("\tscanIdx=SCAN_ZIGZAG\n") break;
    case 1: DTRACE_CABAC_T("\tscanIdx=SCAN_HOR\n") break;
    case 2: DTRACE_CABAC_T("\tscanIdx=SCAN_VER\n") break;
    case 3: DTRACE_CABAC_T("\tscanIdx=SCAN_DIAG\n") break;
  }
#endif

  ContextModel * const baseCoeffGroupCtx = m_cCUSigCoeffGroupSCModel.get( 0, eTType );
  ContextModel * const baseCtx = (eTType==TEXT_LUMA) ? m_cCUSigSCModel.get( 0, 0 ) : m_cCUSigSCModel.get( 0, 0 ) + NUM_SIG_FLAG_CTX_LUMA;

  const Int  iLastScanSet      = uiScanPosLast >> LOG2_SCAN_SET_SIZE;
#if REMOVE_NUM_GREATER1
  UInt c1 = 1;
#else
  UInt uiNumOne                = 0;
#endif
  UInt uiGoRiceParam           = 0;

  Bool beValid; 
  if (pcCU->getCUTransquantBypass(uiAbsPartIdx))
  {
    beValid = false;
  }
  else 
  {
    beValid = pcCU->getSlice()->getPPS()->getSignHideFlag() > 0;
  }
  UInt absSum = 0;

  UInt uiSigCoeffGroupFlag[ MLS_GRP_NUM ];
  ::memset( uiSigCoeffGroupFlag, 0, sizeof(UInt) * MLS_GRP_NUM );
  const UInt uiNumBlkSide = uiWidth >> (MLS_CG_SIZE >> 1);
  const UInt * scanCG;
#if !REMOVE_NSQT
  if (uiWidth == uiHeight)
#endif
  {
    scanCG = g_auiSigLastScan[ uiScanIdx ][ uiLog2BlockSize > 3 ? uiLog2BlockSize-2-1 : 0  ];    
    if( uiLog2BlockSize == 3 )
    {
      scanCG = g_sigLastScan8x8[ uiScanIdx ];
    }
    else if( uiLog2BlockSize == 5 )
    {
      scanCG = g_sigLastScanCG32x32;
    }
  }
#if !REMOVE_NSQT
  else
  {
    scanCG = g_sigCGScanNSQT[ uiLog2BlockSize - 2 ];
  }
#endif
  Int  iScanPosSig             = (Int) uiScanPosLast;
  for( Int iSubSet = iLastScanSet; iSubSet >= 0; iSubSet-- )
  {
    Int  iSubPos     = iSubSet << LOG2_SCAN_SET_SIZE;
    uiGoRiceParam    = 0;
    Int numNonZero = 0;
    
    Int lastNZPosInCG = -1, firstNZPosInCG = SCAN_SET_SIZE;

    Int pos[SCAN_SET_SIZE];
    if( iScanPosSig == (Int) uiScanPosLast )
    {
      lastNZPosInCG  = iScanPosSig;
      firstNZPosInCG = iScanPosSig;
      iScanPosSig--;
      pos[ numNonZero ] = uiBlkPosLast;
      numNonZero = 1;
    }

    // decode significant_coeffgroup_flag
    Int iCGBlkPos = scanCG[ iSubSet ];
    Int iCGPosY   = iCGBlkPos / uiNumBlkSide;
    Int iCGPosX   = iCGBlkPos - (iCGPosY * uiNumBlkSide);
#if !REMOVAL_8x2_2x8_CG
    if( uiWidth == 8 && uiHeight == 8 && (uiScanIdx == SCAN_HOR || uiScanIdx == SCAN_VER) )
    {
      iCGPosY = (uiScanIdx == SCAN_HOR ? iCGBlkPos : 0);
      iCGPosX = (uiScanIdx == SCAN_VER ? iCGBlkPos : 0);
    }
#endif

    DTRACE_CABAC_T( "\tCode SubSet " )
    DTRACE_CABAC_V( iSubSet )
    DTRACE_CABAC_T( " - iCGBlkPos:" )
    DTRACE_CABAC_V( iCGBlkPos )  
    DTRACE_CABAC_T( " (" )
    DTRACE_CABAC_V( iCGPosX )
    DTRACE_CABAC_T( "," )
    DTRACE_CABAC_V( iCGPosY )
    DTRACE_CABAC_T( ")\n" )

    if( iSubSet == iLastScanSet || iSubSet == 0)
    {
      uiSigCoeffGroupFlag[ iCGBlkPos ] = 1;
    }
    else
    {
      UInt uiSigCoeffGroup;
      UInt uiCtxSig  = TComTrQuant::getSigCoeffGroupCtxInc( uiSigCoeffGroupFlag, iCGPosX, iCGPosY, uiScanIdx, uiWidth, uiHeight );
      m_pcTDecBinIf->decodeBin( uiSigCoeffGroup, baseCoeffGroupCtx[ uiCtxSig ] );
      DTRACE_CABAC_T( "\tsignificant_coeff_group_flag\tctx:" )
      DTRACE_CABAC_V( uiCtxSig )
      DTRACE_CABAC_N
      uiSigCoeffGroupFlag[ iCGBlkPos ] = uiSigCoeffGroup;
    }

    // decode significant_coeff_flag
    Int patternSigCtx = TComTrQuant::calcPatternSigCtx( uiSigCoeffGroupFlag, iCGPosX, iCGPosY, uiWidth, uiHeight );
    UInt uiBlkPos, uiPosY, uiPosX, uiSig, uiCtxSig;
    for( ; iScanPosSig >= iSubPos; iScanPosSig-- )
    {
      uiBlkPos  = scan[ iScanPosSig ];
      uiPosY    = uiBlkPos >> uiLog2BlockSize;
      uiPosX    = uiBlkPos - ( uiPosY << uiLog2BlockSize );
      uiSig     = 0;
      
      if( uiSigCoeffGroupFlag[ iCGBlkPos ] )
      {
        if( iScanPosSig > iSubPos || iSubSet == 0  || numNonZero )
        {
#if REMOVAL_8x2_2x8_CG
          uiCtxSig  = TComTrQuant::getSigCtxInc( patternSigCtx, uiScanIdx, uiPosX, uiPosY, blockType, uiWidth, uiHeight, eTType );
#else
          uiCtxSig  = TComTrQuant::getSigCtxInc( patternSigCtx, uiPosX, uiPosY, blockType, uiWidth, uiHeight, eTType );
#endif
          m_pcTDecBinIf->decodeBin( uiSig, baseCtx[ uiCtxSig ] );
          DTRACE_CABAC_T( "\tsignificant_coeff_flag\tctx:" )
          DTRACE_CABAC_V( uiCtxSig )
          DTRACE_CABAC_T( "\tscanPos:" )
          DTRACE_CABAC_V( iScanPosSig )
          DTRACE_CABAC_T( "\tblockPos:" )
          DTRACE_CABAC_V( uiBlkPos )
          DTRACE_CABAC_T( " (" )
          DTRACE_CABAC_V( uiPosX )
          DTRACE_CABAC_T( "," )
          DTRACE_CABAC_V( uiPosY )
          DTRACE_CABAC_T( ")\n" )
        }
        else
        {
          uiSig = 1;
        }
      }
      pcCoef[ uiBlkPos ] = uiSig;
      if( uiSig )
      {
        pos[ numNonZero ] = uiBlkPos;
        numNonZero ++;
        if( lastNZPosInCG == -1 )
        {
          lastNZPosInCG = iScanPosSig;
        }
        firstNZPosInCG = iScanPosSig;
      }
    }
    
    if( numNonZero )
    {
      Bool signHidden = ( lastNZPosInCG - firstNZPosInCG >= SBH_THRESHOLD );
      absSum = 0;
#if !REMOVE_NUM_GREATER1
      UInt c1 = 1;
#endif
      UInt uiCtxSet    = (iSubSet > 0 && eTType==TEXT_LUMA) ? 2 : 0;
      UInt uiBin;
#if REMOVE_NUM_GREATER1
      if( c1 == 0 )
#else
      if( uiNumOne > 0 )
#endif
      {
        uiCtxSet++;
      }
#if REMOVE_NUM_GREATER1
      c1 = 1;
#else
      uiNumOne       >>= 1;
#endif
      ContextModel *baseCtxMod = ( eTType==TEXT_LUMA ) ? m_cCUOneSCModel.get( 0, 0 ) + 4 * uiCtxSet : m_cCUOneSCModel.get( 0, 0 ) + NUM_ONE_FLAG_CTX_LUMA + 4 * uiCtxSet;
      Int absCoeff[SCAN_SET_SIZE];

      for ( Int i = 0; i < numNonZero; i++) absCoeff[i] = 1;   
      Int numC1Flag = min(numNonZero, C1FLAG_NUMBER);
      Int firstC2FlagIdx = -1;

      for( Int idx = 0; idx < numC1Flag; idx++ )
      {
        m_pcTDecBinIf->decodeBin( uiBin, baseCtxMod[c1] );
#if ENC_DEC_TRACE
        DTRACE_CABAC_T( "\tcoeff_abs_level_greater1_flag\tctx:" )
        DTRACE_CABAC_V( c1 )
        DTRACE_CABAC_T( "\tblockPos:" )
        DTRACE_CABAC_V( pos[idx] )
        uiPosY = pos[idx] >> uiLog2BlockSize;
        uiPosX = pos[idx] - ( uiPosY << uiLog2BlockSize );
        DTRACE_CABAC_T( " (" )
        DTRACE_CABAC_V( uiPosX )
        DTRACE_CABAC_T( "," )
        DTRACE_CABAC_V( uiPosY )
        DTRACE_CABAC_T( ")\n" )
#endif
        if( uiBin == 1 )
        {
          c1 = 0;
          if (firstC2FlagIdx == -1)
          {
            firstC2FlagIdx = idx;
          }
        }
        else if( (c1 < 3) && (c1 > 0) )
        {
          c1++;
        }
        absCoeff[ idx ] = uiBin + 1;
      }
      
      if (c1 == 0)
      {
        baseCtxMod = ( eTType==TEXT_LUMA ) ? m_cCUAbsSCModel.get( 0, 0 ) + uiCtxSet : m_cCUAbsSCModel.get( 0, 0 ) + NUM_ABS_FLAG_CTX_LUMA + uiCtxSet;
        if ( firstC2FlagIdx != -1)
        {
          m_pcTDecBinIf->decodeBin( uiBin, baseCtxMod[0] ); 
#if ENC_DEC_TRACE          
          DTRACE_CABAC_T( "\tcoeff_abs_level_greater2_flag\tctx:" )
          DTRACE_CABAC_V( 0 )
          DTRACE_CABAC_T( "\tblockPos:" )
          DTRACE_CABAC_V( pos[firstC2FlagIdx] )
          uiPosY = pos[firstC2FlagIdx] >> uiLog2BlockSize;
          uiPosX = pos[firstC2FlagIdx] - ( uiPosY << uiLog2BlockSize );
          DTRACE_CABAC_T( " (" )
          DTRACE_CABAC_V( uiPosX )
          DTRACE_CABAC_T( "," )
          DTRACE_CABAC_V( uiPosY )
          DTRACE_CABAC_T( ")\n" )
#endif
          absCoeff[ firstC2FlagIdx ] = uiBin + 2;
        }
      }

      UInt coeffSigns;
      DTRACE_CABAC_T( "\tStart Coding signs\n" )
      if ( signHidden && beValid )
      {
        m_pcTDecBinIf->decodeBinsEP( coeffSigns, numNonZero-1 );
      }
      else
      {
        m_pcTDecBinIf->decodeBinsEP( coeffSigns, numNonZero );
      }
      DTRACE_CABAC_T( "\tEnd Coding signs - val:" )
      DTRACE_CABAC_V( coeffSigns )
      DTRACE_CABAC_N

      coeffSigns <<= (signHidden && beValid) ? 32 - (numNonZero-1) : 32 - numNonZero;
      
      Int iFirstCoeff2 = 1;    
      if (c1 == 0 || numNonZero > C1FLAG_NUMBER)
      {
        for( Int idx = 0; idx < numNonZero; idx++ )
        {
          UInt baseLevel  = (idx < C1FLAG_NUMBER)? (2 + iFirstCoeff2) : 1;

          if( absCoeff[ idx ] == baseLevel)
          {
            UInt uiLevel;
            DTRACE_CABAC_T( "\tStart coding coeff_abs_level_remaining\n" )

            xReadCoefRemainExGolomb( uiLevel, uiGoRiceParam );
            absCoeff[ idx ] = uiLevel + baseLevel;
#if ENC_DEC_TRACE            
            DTRACE_CABAC_T( "\tEnd coding coeff_abs_level_remaining\tcoeff_before:" )
            DTRACE_CABAC_V( baseLevel )
            DTRACE_CABAC_T( "\tcoeff_after:" )
            DTRACE_CABAC_V( uiLevel + baseLevel )
            DTRACE_CABAC_T( "\tblockPos:" )
            DTRACE_CABAC_V( pos[idx] )
            uiPosY = pos[idx] >> uiLog2BlockSize;
            uiPosX = pos[idx] - ( uiPosY << uiLog2BlockSize );
            DTRACE_CABAC_T( " (" )
            DTRACE_CABAC_V( uiPosX )
            DTRACE_CABAC_T( "," )
            DTRACE_CABAC_V( uiPosY )
            DTRACE_CABAC_T( ")\n" )
#endif

            
            if(absCoeff[idx]>3*(1<<uiGoRiceParam))
            {
              uiGoRiceParam = min<UInt>(uiGoRiceParam+ 1, 4);
            }
          }

          if(absCoeff[ idx ] >= 2)  
          {
            iFirstCoeff2 = 0;
#if !REMOVE_NUM_GREATER1
            uiNumOne++;
#endif
          }
        }
      }

      for( Int idx = 0; idx < numNonZero; idx++ )
      {
        Int blkPos = pos[ idx ];
        // Signs applied later.
        pcCoef[ blkPos ] = absCoeff[ idx ];
        absSum += absCoeff[ idx ];

        if ( idx == numNonZero-1 && signHidden && beValid )
        {
          // Infer sign of 1st element.
          if (absSum&0x1)
          {
            pcCoef[ blkPos ] = -pcCoef[ blkPos ];
          }
          DTRACE_CABAC_T( "\tInfer hidden sign\tval:" )
          DTRACE_CABAC_V( absSum&0x1 )
          DTRACE_CABAC_T( "\tblockPos:" )
          DTRACE_CABAC_V( blkPos )
          uiPosY = blkPos >> uiLog2BlockSize;
          uiPosX = blkPos - ( uiPosY << uiLog2BlockSize );
          DTRACE_CABAC_T( " (" )
          DTRACE_CABAC_V( uiPosX )
          DTRACE_CABAC_T( "," )
          DTRACE_CABAC_V( uiPosY )
          DTRACE_CABAC_T( ")\n" )
        }
        else
        {
          Int sign = static_cast<Int>( coeffSigns ) >> 31;
          pcCoef[ blkPos ] = ( pcCoef[ blkPos ] ^ sign ) - sign;
          coeffSigns <<= 1;
        }
      }
    }
#if !REMOVE_NUM_GREATER1
    else
    {
      uiNumOne >>= 1;
    }
#endif
  }
  
  DTRACE_CABAC_T( "\tEnd Code CoeffNxN" )
#if ENC_DEC_TRACE_COEFF
  DTRACE_CABAC_T( "\tCoeff: " )
  for (UInt i = 0; i < uiWidth * uiHeight; i++)
  {
    DTRACE_CABAC_V( pcCoef[i] )
    DTRACE_CABAC_T( " " )
  }
#endif
  DTRACE_CABAC_N

  return;
}

/**
  * \param pcCU The current CU
  * \param pcCoef The coefficients of this layer
  * \param pCoefRefLayer The absolut values of the coefficients of the reference layer
  * \param uiBaseLayerQP The QP value used for quantization in the reference layer
  * \param pcColCU The reference layer CU
  */
#if RWTH_RESI_REFINEMENT
Void TDecSbac::parseCoeffRefinement ( TComDataCU* pcCU, TCoeff* pcCoef, TCoeff* pCoefRefLayer, TCoeff* pCoefRefLayerScaled,
#if RWTH_QUANT_USE_REF_COEFF
            UInt uiBaseLayerQP,
#endif                                     
            UInt uiAbsPartIdx, UInt uiWidth, UInt uiHeight, UInt uiDepth, TextType eTType, TComDataCU* pcColCU )
{
  DTRACE_CABAC_T( "\tStart Code Refinement" )
  DTRACE_CABAC_T( "\teType=" )
  DTRACE_CABAC_V( eTType )
  DTRACE_CABAC_T( "\twidth=" )
  DTRACE_CABAC_V( uiWidth )
  DTRACE_CABAC_T( "\theight=" )
  DTRACE_CABAC_V( uiHeight )
  DTRACE_CABAC_T( "\tdepth=" )
  DTRACE_CABAC_V( uiDepth )
  DTRACE_CABAC_T( "\tabspartidx=" )
  DTRACE_CABAC_V( uiAbsPartIdx )
  DTRACE_CABAC_T( "\ttoCU-X=" )
  DTRACE_CABAC_V( pcCU->getCUPelX() )
  DTRACE_CABAC_T( "\ttoCU-Y=" )
  DTRACE_CABAC_V( pcCU->getCUPelY() )
  DTRACE_CABAC_T( "\tCU-addr=" )
  DTRACE_CABAC_V(  pcCU->getAddr() )
  DTRACE_CABAC_T( "\tinCU-X=" )
  DTRACE_CABAC_V( g_auiRasterToPelX[ g_auiZscanToRaster[uiAbsPartIdx] ] )
  DTRACE_CABAC_T( "\tinCU-Y=" )
  DTRACE_CABAC_V( g_auiRasterToPelY[ g_auiZscanToRaster[uiAbsPartIdx] ] )
  DTRACE_CABAC_T( "\tpredmode=" )
  DTRACE_CABAC_V(  pcCU->getPredictionMode( uiAbsPartIdx ) )
  DTRACE_CABAC_N

  if( uiWidth > pcCU->getSlice()->getSPS()->getMaxTrSize() )
  {
    uiWidth  = pcCU->getSlice()->getSPS()->getMaxTrSize();
    uiHeight = pcCU->getSlice()->getSPS()->getMaxTrSize();
  }

  eTType = eTType == TEXT_LUMA ? TEXT_LUMA : ( eTType == TEXT_NONE ? TEXT_NONE : TEXT_CHROMA );

  //----- parse significance map -----
  const UInt  uiLog2BlockSize   = g_aucConvertToBit[ uiWidth ] + 2;
  const UInt  uiMaxNumCoeff     = uiWidth * uiHeight;
  const UInt  uiMaxNumCoeffM1   = uiMaxNumCoeff - 1;
  UInt uiScanIdx;
  if (getDecTop()->getMyCodingArgument()[USE_BL_SCAN_IDX])
  {
    uiScanIdx = pcColCU->getCoefScanIdx(uiAbsPartIdx, uiWidth, eTType==TEXT_LUMA, pcColCU->isIntra(uiAbsPartIdx));
  }
  else
  {
    uiScanIdx = pcCU->getCoefScanIdx(uiAbsPartIdx, uiWidth, eTType==TEXT_LUMA, pcCU->isIntra(uiAbsPartIdx));
  }
  int blockType = uiLog2BlockSize;
  if (uiWidth != uiHeight)
  {
    uiScanIdx = SCAN_DIAG;
    blockType = 4;
  }

  //===== decode significance flags =====
  if (uiScanIdx == SCAN_ZIGZAG)
  {
    // Map zigzag to diagonal scan
    uiScanIdx = SCAN_DIAG;
  }

  const UInt * scan = g_auiSigLastScan[ uiScanIdx ][ uiLog2BlockSize-1 ];
#if !REMOVE_NSQT
  if (uiWidth != uiHeight)
  {
    scan = g_sigScanNSQT[ uiLog2BlockSize - 2 ];
  }
#endif

  //===== decode last significant =====
  UInt uiPosLastX = 0, uiPosLastY = 0;
#if RWTH_CODE_LASTXY_DIFF
  if (getDecTop()->getMyCodingArgument()[CODE_LAST_SIGNIFICANT_DIFF])
  {
    // Get the last significant XY from the lower layer
    Int iScanPostLastLowerLayer = -1;
    Int posLastRef;
    
    // Count lower layer coefficients
    UInt uiNumSigLowerLayer = 0;
    for ( Int i = 0; i < uiWidth*uiHeight; i++ )
    {
      uiNumSigLowerLayer += pCoefRefLayer[i] != 0;
    }

    // Get last significantXY of ref layer
    do
    {
      posLastRef = scan[ ++iScanPostLastLowerLayer ];
      uiNumSigLowerLayer -= ( pCoefRefLayer[ posLastRef ] != 0 );
    }
    while ( uiNumSigLowerLayer > 0 );

    UInt uiPosLastYRef = posLastRef >> uiLog2BlockSize;
    UInt uiPosLastXRef = posLastRef - ( uiPosLastYRef << uiLog2BlockSize );
    
    parseLastSignificantDiff( uiPosLastX, uiPosLastXRef, uiPosLastY, uiPosLastYRef, uiWidth, uiHeight, eTType, uiScanIdx, scan );
  }
  else
#endif
  {
    parseLastSignificantXY( uiPosLastX, uiPosLastY, uiWidth, uiHeight, eTType, uiScanIdx );
  }
  
  UInt uiBlkPosLast      = uiPosLastX + (uiPosLastY<<uiLog2BlockSize);
  UInt uiScanPosLast = uiBlkPosLast;
  pcCoef[ uiBlkPosLast ] = 1;
  for( uiScanPosLast = 0; uiScanPosLast < uiMaxNumCoeffM1; uiScanPosLast++ )
  {
    UInt uiBlkPos = scan[ uiScanPosLast ];
    if( uiBlkPosLast == uiBlkPos )
    {
      break;
    }
  }

#if ENC_DEC_TRACE
  switch (uiScanIdx)
  {
    case 0: DTRACE_CABAC_T("\tscanIdx=SCAN_ZIGZAG\n") break;
    case 1: DTRACE_CABAC_T("\tscanIdx=SCAN_HOR\n") break;
    case 2: DTRACE_CABAC_T("\tscanIdx=SCAN_VER\n") break;
    case 3: DTRACE_CABAC_T("\tscanIdx=SCAN_DIAG\n") break;
  }
#endif

  ContextModel * const baseCoeffGroupCtx = m_cCUSigCoeffGroupSCModel.get( 0, eTType );
  ContextModel * const baseCtx = (eTType==TEXT_LUMA) ? m_cCUSigSCModel.get( 0, 0 ) : m_cCUSigSCModel.get( 0, 0 ) + NUM_SIG_FLAG_CTX_LUMA;

  const Int  iLastScanSet      = uiScanPosLast >> LOG2_SCAN_SET_SIZE;

  // calculate significant group flag of reference layer
  UInt uiNumSig = 0;
  for ( Int i = 0; i < uiWidth * uiHeight; i++ )
  {
    uiNumSig += pCoefRefLayer[i] != 0;
  }
  Int scanPosLastRef = -1;
  Int posLastRef;
  Int iFirstNonSigRefLayer = -1;
  UInt uiSigCoeffGroupFlagRefLayer[ MLS_GRP_NUM ];
  static const UInt uiShift = MLS_CG_SIZE >> 1;
  const UInt uiNumBlkSide = uiWidth >> uiShift;
  ::memset( uiSigCoeffGroupFlagRefLayer,  0, sizeof(UInt) * MLS_GRP_NUM );
  do
  {
    posLastRef = scan[ ++scanPosLastRef ];

    // get L1 sig map
    UInt uiPosY    = posLastRef >> uiLog2BlockSize;
    UInt uiPosX    = posLastRef - ( uiPosY << uiLog2BlockSize );
    UInt uiBlkIdx  = uiNumBlkSide * (uiPosY >> uiShift) + (uiPosX >> uiShift);
#if !REMOVAL_8x2_2x8_CG    
    if( uiWidth == 8 && uiHeight == 8 && (uiScanIdx == SCAN_HOR || uiScanIdx == SCAN_VER) )
    {
      if( uiScanIdx == SCAN_HOR )
      {
        uiBlkIdx = uiPosY >> 1;
      }
      else if( uiScanIdx == SCAN_VER )
      {
        uiBlkIdx = uiPosX >> 1;
      }
    }
#endif
    if( pCoefRefLayer[ posLastRef ] )
    {
      uiSigCoeffGroupFlagRefLayer[ uiBlkIdx ] = 1;
    }
    else
    {
      if (iFirstNonSigRefLayer<0)
      {
        iFirstNonSigRefLayer = scanPosLastRef;
      }
    }

    uiNumSig -= ( pCoefRefLayer[ posLastRef ] != 0 );
  }
  while ( uiNumSig > 0 );

  // decode significant_coeffgroup_flag
  UInt uiSigCoeffGroupFlag[ MLS_GRP_NUM ];
  ::memset( uiSigCoeffGroupFlag,  0, sizeof(UInt) * MLS_GRP_NUM );
  const UInt * scanCG;
#if !REMOVE_NSQT
  if (uiWidth == uiHeight)
#endif
  {
    scanCG = g_auiSigLastScan[ uiScanIdx ][ uiLog2BlockSize > 3 ? uiLog2BlockSize-2-1 : 0  ];
    if( uiLog2BlockSize == 3 )
    {
      scanCG = g_sigLastScan8x8[ uiScanIdx ];
    }
    else if( uiLog2BlockSize == 5 )
    {
      scanCG = g_sigLastScanCG32x32;
    }
  }
#if !REMOVE_NSQT
  else
  {
    scanCG = g_sigCGScanNSQT[ uiLog2BlockSize - 2 ];
  }
#endif

#if RWTH_QUANT_USE_REF_COEFF
  UInt uiQP = pcCU->getQP(uiAbsPartIdx);
#endif
#if !TURN_OFF_CHROMA_QP_SCALE
  UInt uiBaseLayerQPChroma = g_aucChromaScale[ Clip3(0, 51, (Int)uiBaseLayerQP) ];
  UInt uiQPChroma = g_aucChromaScale[ Clip3(0, 51, (Int) uiQP) ];
#endif

  Int  iScanPosSig             = (Int) uiScanPosLast;
  for( Int iSubSet = iLastScanSet; iSubSet >= 0; iSubSet-- )
  {
    Int  iSubPos     = iSubSet << LOG2_SCAN_SET_SIZE;
    Int numNewSig = 0;
    Int pos[SCAN_SET_SIZE];
    Int posRefinement[SCAN_SET_SIZE];
    Int numRefinement = 0;

    if( iScanPosSig == (Int) uiScanPosLast )
    {
      if ( pCoefRefLayer[uiBlkPosLast] == 0)
      {
        pos[ numNewSig ] = uiBlkPosLast;
        numNewSig = 1;
      }
      else
      {
        posRefinement[ numRefinement ] = uiBlkPosLast;
        numRefinement = 1;
      }
      iScanPosSig--;
    }

    Int iCGBlkPos = scanCG[ iSubSet ];
    Int iCGPosY   = iCGBlkPos / uiNumBlkSide;
    Int iCGPosX   = iCGBlkPos - (iCGPosY * uiNumBlkSide);
#if !REMOVAL_8x2_2x8_CG    
    if( uiWidth == 8 && uiHeight == 8 && (uiScanIdx == SCAN_HOR || uiScanIdx == SCAN_VER) )
    {
      iCGPosY = (uiScanIdx == SCAN_HOR ? iCGBlkPos : 0);
      iCGPosX = (uiScanIdx == SCAN_VER ? iCGBlkPos : 0);
    }
#endif
    if( iSubSet == iLastScanSet || iSubSet == 0 || uiSigCoeffGroupFlagRefLayer[ iCGBlkPos ])
    {
      uiSigCoeffGroupFlag[ iCGBlkPos ] = 1;
    }
    else
    {
      UInt uiSigCoeffGroup;
      UInt uiCtxSig  = TComTrQuant::getSigCoeffGroupCtxInc( uiSigCoeffGroupFlag, iCGPosX, iCGPosY, uiScanIdx, uiWidth, uiHeight );
      m_pcTDecBinIf->decodeBin( uiSigCoeffGroup, baseCoeffGroupCtx[ uiCtxSig ] );
      DTRACE_CABAC_T( "\tGroup Significance\t" )
      DTRACE_CABAC_V( uiCtxSig )
      DTRACE_CABAC_N
      uiSigCoeffGroupFlag[ iCGBlkPos ] = uiSigCoeffGroup;
    }

    // decode significant_coeff_flag
    UInt uiBlkPos, uiPosY, uiPosX, uiSig, uiCtxSig, uiSigRefLayer;
    Int patternSigCtx = TComTrQuant::calcPatternSigCtx( uiSigCoeffGroupFlag, iCGPosX, iCGPosY, uiWidth, uiHeight );
    for( ; iScanPosSig >= iSubPos; iScanPosSig-- )
    {
      uiBlkPos  = scan[ iScanPosSig ];
      uiPosY    = uiBlkPos >> uiLog2BlockSize;
      uiPosX    = uiBlkPos - ( uiPosY << uiLog2BlockSize );
      uiSig     = 0;
      uiSigRefLayer = (pCoefRefLayer[ uiBlkPos ] != 0);

      if( uiSigCoeffGroupFlag[ iCGBlkPos ] )
      {
        if( (iScanPosSig > iSubPos || iSubSet == 0  || numNewSig || numRefinement) )
        {
          if (uiSigRefLayer==0 && iScanPosSig>= iFirstNonSigRefLayer)
          {
            uiCtxSig = TComTrQuant::getSigCtxInc( patternSigCtx, uiScanIdx, uiPosX, uiPosY, blockType, uiWidth, uiHeight, eTType );
            m_pcTDecBinIf->decodeBin( uiSig, baseCtx[ uiCtxSig ] );
            DTRACE_CABAC_T( "\tsignificant_coeff_flag\tctx:" )
            DTRACE_CABAC_V( uiCtxSig )
            DTRACE_CABAC_T( "\tscanPos:" )
            DTRACE_CABAC_V( iScanPosSig )
            DTRACE_CABAC_T( "\tblockPos:" )
            DTRACE_CABAC_V( uiBlkPos )
            DTRACE_CABAC_T( " (" )
            DTRACE_CABAC_V( uiPosX )
            DTRACE_CABAC_T( "," )
            DTRACE_CABAC_V( uiPosY )
            DTRACE_CABAC_T( ")\n" )
          }
          else
          {
            uiSig = 1;
          }
        }
        else
        {
          uiSig = 1;
        }

        if (uiSig)
        {
          if( uiSigRefLayer )
          {
            posRefinement[ numRefinement ] = uiBlkPos;
            numRefinement++;
            pcCoef[ uiBlkPos ] = pCoefRefLayer[ uiBlkPos ];
          }
          else
          {
            pos[ numNewSig ] = uiBlkPos;
            numNewSig ++;
            pcCoef[ uiBlkPos ] = uiSig;
          }
        }
      }
    }

    // decode signs of new significant coefficients
    UInt coeffSigns;
    m_pcTDecBinIf->decodeBinsEP( coeffSigns, numNewSig );
    coeffSigns <<= 32 - numNewSig;

    for( Int idx = 0; idx < numNewSig; idx++ )
    {
      Int blkPos = pos[ idx ];

      Int sign = static_cast<Int>( coeffSigns ) >> 31;
      pcCoef[ blkPos ] = ( pcCoef[ blkPos ] ^ sign ) - sign;
      coeffSigns <<= 1;
    }

#if RWTH_CTX_FOR_REFINEMENT
    ContextModel * const baseCoeffRefinement = m_cCURefinementModel.get( 0, eTType? TEXT_CHROMA: TEXT_LUMA );
#endif
    // decode refinement
    for (Int idx = 0; idx < numRefinement; idx++ )
    {
      Int blkPos = posRefinement[idx];
      UInt refinement;
#if RWTH_CTX_FOR_REFINEMENT
      m_pcTDecBinIf->decodeBin( refinement,  baseCoeffRefinement[0] );
#else
      m_pcTDecBinIf->decodeBinEP( refinement );
#endif
#if COLLECT_STAT
      extern double g_refinementStat[2][128];
      extern double g_refinementStatC[2][128];
      if (abs(pCoefRefLayer[blkPos])<128)
      {
        if (eTType==TEXT_LUMA)
          g_refinementStat[refinement][abs(pCoefRefLayer[blkPos])]++;
        else
          g_refinementStatC[refinement][abs(pCoefRefLayer[blkPos])]++;
      }
#endif
      
#if ENC_DEC_TRACE
    {
      DTRACE_CABAC_T( "\tcoeff_refinement\tctx:" )
      DTRACE_CABAC_V( 0 )
      DTRACE_CABAC_T( "\tscanPos:" )
      DTRACE_CABAC_V( idx )
      UInt uiPosY_trace = blkPos >> uiLog2BlockSize;
      UInt uiPosX_trace = blkPos - ( uiPosY_trace << uiLog2BlockSize );
      DTRACE_CABAC_T( "\tblockPos:" )
      DTRACE_CABAC_V( blkPos )
      DTRACE_CABAC_T( " (" )
      DTRACE_CABAC_V( uiPosX_trace )
      DTRACE_CABAC_T( "," )
      DTRACE_CABAC_V( uiPosY_trace )
      DTRACE_CABAC_T( ")\n" )
    }
#endif

      Int sign = ( pCoefRefLayer[blkPos] < 0 ) ? -1 : 1;
#if RWTH_QUANT_USE_REF_COEFF
      Double factor;
      Double scale_BL;
      Double scale_EL;
#if !TURN_OFF_CHROMA_QP_SCALE
      if (eTType != TEXT_LUMA)
      {
        scale_BL = g_invQuantScales[ uiBaseLayerQPChroma%6 ];
        scale_EL = g_invQuantScales[ uiQPChroma%6          ];
      }
      else
#endif
      {
        scale_BL = g_invQuantScales[ uiBaseLayerQP%6 ];
        scale_EL = g_invQuantScales[ uiQP%6          ];
      }
      if (scale_BL <= scale_EL)
        scale_BL *= 2;
      factor = scale_BL/scale_EL;
      pCoefRefLayerScaled[blkPos] = factor*pCoefRefLayer[blkPos];
      Int signedRefinement = refinement*sign;
      pcCoef[blkPos] = signedRefinement + pCoefRefLayerScaled[blkPos];
#else
      pcCoef[blkPos] = refinement*sign + 2*pCoefRefLayer[blkPos];
#endif
    }
  }

#if 0
  printf("decoding pcCoeff\n");
  for (int i = 0; i<uiHeight; i++){
    for (int j = 0; j<uiWidth; j++){
      printf("%d\t", pcCoef[j + i*uiWidth]);
    }
    printf("\n");
  }

  printf("pcCoeffRefLayer\n");
  for (int i = 0; i<uiHeight; i++){
    for (int j = 0; j<uiWidth; j++){
      printf("%d\t", pCoefRefLayer[j + i*uiWidth]);
    }
    printf("\n");
  }
#endif
  DTRACE_CABAC_T( "\tEnd Code Refinement\n" )

#if ENC_DEC_TRACE_COEFF
  // Trace coeffs and reference block coeffs
  DTRACE_CABAC_T( "\t     Coeff: " )
  for (int i = 0; i<uiHeight; i++){
    for (int j = 0; j<uiWidth; j++){
      DTRACE_CABAC_V( pcCoef[j + i*uiWidth] )
      DTRACE_CABAC_T( " " )
    }
  }
  DTRACE_CABAC_N

  UInt uiCoeff;
  DTRACE_CABAC_T( "\t Ref Coeff: " )
  for (int i = 0; i<uiHeight; i++){
    for (int j = 0; j<uiWidth; j++){
      uiCoeff = pCoefRefLayerScaled[j + i*uiWidth];
      DTRACE_CABAC_V( uiCoeff )
      DTRACE_CABAC_T( " " )
    }
  }
  DTRACE_CABAC_N
#endif
}
#endif

Void TDecSbac::parseSaoMaxUvlc ( UInt& val, UInt maxSymbol )
{
  if (maxSymbol == 0)
  {
    val = 0;
    return;
  }

  UInt code;
  Int  i;
#if SAO_ABS_BY_PASS
  m_pcTDecBinIf->decodeBinEP( code );
#else
  m_pcTDecBinIf->decodeBin( code, m_cSaoUvlcSCModel.get( 0, 0, 0 ) );
#endif
  DTRACE_CABAC_T( "\tSaoUvlc\n" )
  if ( code == 0 )
  {
    val = 0;
    return;
  }

  i=1;
  while (1)
  {
#if SAO_ABS_BY_PASS
    m_pcTDecBinIf->decodeBinEP( code );
#else
    m_pcTDecBinIf->decodeBin( code, m_cSaoUvlcSCModel.get( 0, 0, 1 ) );
#endif
    DTRACE_CABAC_T( "\tSaoUvlc\n" )
    if ( code == 0 )
    {
      break;
    }
    i++;
    if (i == maxSymbol) 
    {
      break;
    }
  }

  val = i;
}
#if SAO_TYPE_CODING
Void TDecSbac::parseSaoUflc (UInt uiLength, UInt&  riVal)
{
  m_pcTDecBinIf->decodeBinsEP ( riVal, uiLength );
}
#else
Void TDecSbac::parseSaoUflc (UInt&  riVal)
{
  m_pcTDecBinIf->decodeBinsEP ( riVal, 5 );
}
#endif
#if SAO_MERGE_ONE_CTX
Void TDecSbac::parseSaoMerge (UInt&  ruiVal)
#else
Void TDecSbac::parseSaoMergeLeft (UInt&  ruiVal, UInt uiCompIdx)
#endif
{
  UInt uiCode;
#if SAO_MERGE_ONE_CTX
  m_pcTDecBinIf->decodeBin( uiCode, m_cSaoMergeSCModel.get( 0, 0, 0 ) );
  DTRACE_CABAC_T( "\tSaoMergeLeft\n" )
#else
#if SAO_SINGLE_MERGE
  m_pcTDecBinIf->decodeBin( uiCode, m_cSaoMergeLeftSCModel.get( 0, 0, 0 ) );
  DTRACE_CABAC_T( "\tSaoMergeLeft\n" )
#else
  m_pcTDecBinIf->decodeBin( uiCode, m_cSaoMergeLeftSCModel.get( 0, 0, uiCompIdx ) );
  DTRACE_CABAC_T( "\tSaoMergeLeft\n" )
#endif
#endif
  ruiVal = (Int)uiCode;
}
#if !SAO_MERGE_ONE_CTX
Void TDecSbac::parseSaoMergeUp (UInt&  ruiVal)
{
  UInt uiCode;
  m_pcTDecBinIf->decodeBin( uiCode, m_cSaoMergeUpSCModel.get( 0, 0, 0 ) );
  DTRACE_CABAC_T( "\tSaoMergeUp\n" )
  ruiVal = (Int)uiCode;
}
#endif
Void TDecSbac::parseSaoTypeIdx (UInt&  ruiVal)
{
  UInt uiCode;
#if SAO_TYPE_CODING
  m_pcTDecBinIf->decodeBin( uiCode, m_cSaoTypeIdxSCModel.get( 0, 0, 0 ) );
  DTRACE_CABAC_T( "\tSaoTypeIdx\n" )
  if (uiCode == 0) 
  {
    ruiVal = 0;
  }
  else
  {
    m_pcTDecBinIf->decodeBinEP( uiCode ); 
    DTRACE_CABAC_T( "\tSaoTypeIdx\n" );
    if (uiCode == 0)
    {
      ruiVal = 5;
    }
    else
    {
      ruiVal = 1;
    }
  }
#else
  Int  i;
  m_pcTDecBinIf->decodeBin( uiCode, m_cSaoTypeIdxSCModel.get( 0, 0, 0 ) );
  DTRACE_CABAC_T( "\tSaoTypeIdx\n" )
  if ( uiCode == 0 )
  {
    ruiVal = 0;
    return;
  }
  i=1;
  while (1)
  {
    m_pcTDecBinIf->decodeBin( uiCode, m_cSaoTypeIdxSCModel.get( 0, 0, 1 ) );
    DTRACE_CABAC_T( "\tSaoTypeIdx\n" )

    if ( uiCode == 0 )
    {
      break;
    }
    i++;
  }
  ruiVal = i;
#endif
}

inline Void copySaoOneLcuParam(SaoLcuParam* psDst,  SaoLcuParam* psSrc)
{
  Int i;
  psDst->partIdx = psSrc->partIdx;
  psDst->typeIdx    = psSrc->typeIdx;
  if (psDst->typeIdx != -1)
  {
#if SAO_TYPE_CODING
    psDst->subTypeIdx = psSrc->subTypeIdx ;
#else
    if (psDst->typeIdx == SAO_BO)
    {
      psDst->bandPosition = psSrc->bandPosition ;
    }
    else
    {
      psDst->bandPosition = 0;
    }
#endif
    psDst->length  = psSrc->length;
    for (i=0;i<psDst->length;i++)
    {
      psDst->offset[i] = psSrc->offset[i];
    }
  }
  else
  {
    psDst->length  = 0;
    for (i=0;i<SAO_BO_LEN;i++)
    {
      psDst->offset[i] = 0;
    }
  }
}

#if SAO_TYPE_SHARING
Void TDecSbac::parseSaoOffset(SaoLcuParam* psSaoLcuParam, UInt compIdx)
#else
Void TDecSbac::parseSaoOffset(SaoLcuParam* psSaoLcuParam)
#endif
{
  UInt uiSymbol;
  static Int iTypeLength[MAX_NUM_SAO_TYPE] =
  {
    SAO_EO_LEN,
    SAO_EO_LEN,
    SAO_EO_LEN,
    SAO_EO_LEN,
    SAO_BO_LEN
  }; 

#if SAO_TYPE_SHARING
if (compIdx==2)
{
uiSymbol = (UInt)( psSaoLcuParam->typeIdx + 1);
}
else
{
parseSaoTypeIdx(uiSymbol);
}
#else
  parseSaoTypeIdx(uiSymbol);
#endif
  psSaoLcuParam->typeIdx = (Int)uiSymbol - 1;
  if (uiSymbol)
  {
    psSaoLcuParam->length = iTypeLength[psSaoLcuParam->typeIdx];
#if FULL_NBIT
    Int offsetTh = 1 << ( min((Int)(g_uiBitDepth + (g_uiBitDepth-8)-5),5) );
#else
    Int offsetTh = 1 << ( min((Int)(g_uiBitDepth + g_uiBitIncrement-5),5) );
#endif

    if( psSaoLcuParam->typeIdx == SAO_BO )
    {
#if !SAO_TYPE_CODING
    // Parse Left Band Index
    parseSaoUflc( uiSymbol );
    psSaoLcuParam->bandPosition = uiSymbol;
#endif
      for(Int i=0; i< psSaoLcuParam->length; i++)
      {
        parseSaoMaxUvlc(uiSymbol, offsetTh -1 );
        psSaoLcuParam->offset[i] = uiSymbol;
      }   
      for(Int i=0; i< psSaoLcuParam->length; i++)
      {
        if (psSaoLcuParam->offset[i] != 0) 
        {
          m_pcTDecBinIf->decodeBinEP ( uiSymbol);
          DTRACE_CABAC_T( "\tSAOSign\n" )
          if (uiSymbol)
          {
            psSaoLcuParam->offset[i] = -psSaoLcuParam->offset[i] ;
          }
        }
      }
#if SAO_TYPE_CODING
      parseSaoUflc(5, uiSymbol );
      psSaoLcuParam->subTypeIdx = uiSymbol;
#endif
    }
    else if( psSaoLcuParam->typeIdx < 4 )
    {
      parseSaoMaxUvlc(uiSymbol, offsetTh -1 ); psSaoLcuParam->offset[0] = uiSymbol;
      parseSaoMaxUvlc(uiSymbol, offsetTh -1 ); psSaoLcuParam->offset[1] = uiSymbol;
      parseSaoMaxUvlc(uiSymbol, offsetTh -1 ); psSaoLcuParam->offset[2] = -(Int)uiSymbol;
      parseSaoMaxUvlc(uiSymbol, offsetTh -1 ); psSaoLcuParam->offset[3] = -(Int)uiSymbol;
#if SAO_TYPE_CODING
#if SAO_TYPE_SHARING
     if (compIdx != 2)
     {
       parseSaoUflc(2, uiSymbol );
       psSaoLcuParam->subTypeIdx = uiSymbol;
       psSaoLcuParam->typeIdx += psSaoLcuParam->subTypeIdx;
     }
#else
     parseSaoUflc(2, uiSymbol );
     psSaoLcuParam->subTypeIdx = uiSymbol;
     psSaoLcuParam->typeIdx += psSaoLcuParam->subTypeIdx;
#endif
#endif
   }
  }
  else
  {
    psSaoLcuParam->length = 0;
  }
}

Void TDecSbac::parseSaoOneLcuInterleaving(Int rx, Int ry, SAOParam* pSaoParam, TComDataCU* pcCU, Int iCUAddrInSlice, Int iCUAddrUpInSlice, Int allowMergeLeft, Int allowMergeUp)
{
  //DTRACE_CABAC_T( "\tStart SaoOneLcuInterleaving\n" );

  Int iAddr = pcCU->getAddr();
  UInt uiSymbol;
#if SAO_SINGLE_MERGE
  for (Int iCompIdx=0; iCompIdx<3; iCompIdx++)
  {
    pSaoParam->saoLcuParam[iCompIdx][iAddr].mergeUpFlag    = 0;
    pSaoParam->saoLcuParam[iCompIdx][iAddr].mergeLeftFlag  = 0;
#if SAO_TYPE_CODING
    pSaoParam->saoLcuParam[iCompIdx][iAddr].subTypeIdx     = 0;
#else
    pSaoParam->saoLcuParam[iCompIdx][iAddr].bandPosition   = 0;
#endif
    pSaoParam->saoLcuParam[iCompIdx][iAddr].typeIdx        = -1;
    pSaoParam->saoLcuParam[iCompIdx][iAddr].offset[0]     = 0;
    pSaoParam->saoLcuParam[iCompIdx][iAddr].offset[1]     = 0;
    pSaoParam->saoLcuParam[iCompIdx][iAddr].offset[2]     = 0;
    pSaoParam->saoLcuParam[iCompIdx][iAddr].offset[3]     = 0;

  }
#if SAO_TYPE_SHARING
 if (pSaoParam->bSaoFlag[0] || pSaoParam->bSaoFlag[1] )
#else
 if (pSaoParam->bSaoFlag[0] || pSaoParam->bSaoFlag[1] || pSaoParam->bSaoFlag[2])
#endif
  {
    if (rx>0 && iCUAddrInSlice!=0 && allowMergeLeft)
    {
#if SAO_MERGE_ONE_CTX
      parseSaoMerge(uiSymbol); 
      pSaoParam->saoLcuParam[0][iAddr].mergeLeftFlag = (Bool)uiSymbol;  
#else
      parseSaoMergeLeft(uiSymbol, 0); 
      pSaoParam->saoLcuParam[0][iAddr].mergeLeftFlag = (Bool)uiSymbol;   
#endif
    }
    if (pSaoParam->saoLcuParam[0][iAddr].mergeLeftFlag==0)
    {
      if ((ry > 0) && (iCUAddrUpInSlice>=0) && allowMergeUp)
      {
#if SAO_MERGE_ONE_CTX
        parseSaoMerge(uiSymbol);
        pSaoParam->saoLcuParam[0][iAddr].mergeUpFlag = (Bool)uiSymbol;
#else
        parseSaoMergeUp(uiSymbol);
        pSaoParam->saoLcuParam[0][iAddr].mergeUpFlag = (Bool)uiSymbol;
#endif
      }
    }
  }
#endif

  for (Int iCompIdx=0; iCompIdx<3; iCompIdx++)
  {
#if !SAO_SINGLE_MERGE
    pSaoParam->saoLcuParam[iCompIdx][iAddr].mergeUpFlag    = 0;
    pSaoParam->saoLcuParam[iCompIdx][iAddr].mergeLeftFlag  = 0;
    pSaoParam->saoLcuParam[iCompIdx][iAddr].bandPosition   = 0;
    pSaoParam->saoLcuParam[iCompIdx][iAddr].typeIdx        = -1;
    pSaoParam->saoLcuParam[iCompIdx][iAddr].offset[0]     = 0;
    pSaoParam->saoLcuParam[iCompIdx][iAddr].offset[1]     = 0;
    pSaoParam->saoLcuParam[iCompIdx][iAddr].offset[2]     = 0;
    pSaoParam->saoLcuParam[iCompIdx][iAddr].offset[3]     = 0;
#endif
#if SAO_TYPE_SHARING
    if ((iCompIdx == 0  && pSaoParam->bSaoFlag[0]) || (iCompIdx > 0  && pSaoParam->bSaoFlag[1]) )
#else
    if (pSaoParam->bSaoFlag[iCompIdx])
#endif
    {
      if (rx>0 && iCUAddrInSlice!=0 && allowMergeLeft)
      {
#if SAO_SINGLE_MERGE
        pSaoParam->saoLcuParam[iCompIdx][iAddr].mergeLeftFlag = pSaoParam->saoLcuParam[0][iAddr].mergeLeftFlag;
#else
        parseSaoMergeLeft(uiSymbol,iCompIdx); pSaoParam->saoLcuParam[iCompIdx][iAddr].mergeLeftFlag = (Int)uiSymbol;
#endif
      }
      else
      {
        pSaoParam->saoLcuParam[iCompIdx][iAddr].mergeLeftFlag = 0;
      }

      if (pSaoParam->saoLcuParam[iCompIdx][iAddr].mergeLeftFlag==0)
      {
        if ((ry > 0) && (iCUAddrUpInSlice>=0) && allowMergeUp)
        {
#if SAO_SINGLE_MERGE
          pSaoParam->saoLcuParam[iCompIdx][iAddr].mergeUpFlag = pSaoParam->saoLcuParam[0][iAddr].mergeUpFlag;
#else
          parseSaoMergeUp(uiSymbol);  pSaoParam->saoLcuParam[iCompIdx][iAddr].mergeUpFlag = uiSymbol;
#endif
        }
        else
        {
          pSaoParam->saoLcuParam[iCompIdx][iAddr].mergeUpFlag = 0;
        }
        if (!pSaoParam->saoLcuParam[iCompIdx][iAddr].mergeUpFlag)
        {
#if SAO_TYPE_SHARING
          pSaoParam->saoLcuParam[2][iAddr].typeIdx = pSaoParam->saoLcuParam[1][iAddr].typeIdx;
          parseSaoOffset(&(pSaoParam->saoLcuParam[iCompIdx][iAddr]), iCompIdx);
#else
          parseSaoOffset(&(pSaoParam->saoLcuParam[iCompIdx][iAddr]));
#endif
        }
        else
        {
          copySaoOneLcuParam(&pSaoParam->saoLcuParam[iCompIdx][iAddr], &pSaoParam->saoLcuParam[iCompIdx][iAddr-pSaoParam->numCuInWidth]);
        }
      }
      else
      {
        copySaoOneLcuParam(&pSaoParam->saoLcuParam[iCompIdx][iAddr],  &pSaoParam->saoLcuParam[iCompIdx][iAddr-1]);
      }
    }
    else
    {
      pSaoParam->saoLcuParam[iCompIdx][iAddr].typeIdx = -1;
#if SAO_TYPE_CODING
      pSaoParam->saoLcuParam[iCompIdx][iAddr].subTypeIdx = 0;
#else
      pSaoParam->saoLcuParam[iCompIdx][iAddr].bandPosition = 0;
#endif
    }
  }
  //DTRACE_CABAC_T( "\tEnd SaoOneLcuInterleaving\n" );
}

#if !REMOVE_ALF
Void TDecSbac::parseAlfCtrlFlag (Int compIdx, UInt& code)
{
  UInt decodedSymbol;
  m_pcTDecBinIf->decodeBin( decodedSymbol, m_cCUAlfCtrlFlagSCModel.get( 0, 0, 0 ) );
  code = decodedSymbol;

  DTRACE_CABAC_VL( g_nSymbolCounter++ )
  DTRACE_CABAC_T( "parseAlfCtrlFlag()" )
  DTRACE_CABAC_T( "\tsymbol=" )
  DTRACE_CABAC_V( decodedSymbol )
  DTRACE_CABAC_T( "\tcompIdx=" )
  DTRACE_CABAC_V( compIdx )
  DTRACE_CABAC_T( "\n" )
}
#endif

/**
 - Initialize our contexts from the nominated source.
 .
 \param pSrc Contexts to be copied.
 */
Void TDecSbac::xCopyContextsFrom( TDecSbac* pSrc )
{
  memcpy(m_contextModels, pSrc->m_contextModels, m_numContextModels*sizeof(m_contextModels[0]));
}

Void TDecSbac::xCopyFrom( TDecSbac* pSrc )
{
  m_pcTDecBinIf->copyState( pSrc->m_pcTDecBinIf );

  m_uiLastQp           = pSrc->m_uiLastQp;
  xCopyContextsFrom( pSrc );

}

Void TDecSbac::load ( TDecSbac* pScr )
{
  xCopyFrom(pScr);
}

Void TDecSbac::loadContexts ( TDecSbac* pScr )
{
  xCopyContextsFrom(pScr);
}

Void TDecSbac::decodeFlush ( )
{
  UInt uiBit;
  m_pcTDecBinIf->decodeBinTrm(uiBit);
  DTRACE_CABAC_N

  m_pcTDecBinIf->flush();

}
//! \}
