/* The copyright in this software is being made available under the BSD
 * License, included below. This software may be subject to other third party
 * and contributor rights, including patent rights, and no such rights are
 * granted under this license.  
 *
 * Copyright (c) 2010-2012, ITU/ISO/IEC
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *  * Neither the name of the ITU/ISO/IEC nor the names of its contributors may
 *    be used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

/** \file     TEncSbac.cpp
    \brief    SBAC encoder class
*/

#include "TEncTop.h"
#include "TEncSbac.h"

#include <map>
#include <algorithm>

//! \ingroup TLibEncoder
//! \{

// ====================================================================================================================
// Constructor / destructor / create / destroy
// ====================================================================================================================

TEncSbac::TEncSbac()
// new structure here
: m_pcBitIf                   ( NULL )
, m_pcSlice                   ( NULL )
, m_pcBinIf                   ( NULL )
, m_uiCoeffCost               ( 0 )
, m_numContextModels          ( 0 )
, m_cCUSplitFlagSCModel       ( 1,             1,               NUM_SPLIT_FLAG_CTX            , m_contextModels + m_numContextModels, m_numContextModels )
, m_cCUSkipFlagSCModel        ( 1,             1,               NUM_SKIP_FLAG_CTX             , m_contextModels + m_numContextModels, m_numContextModels)
, m_cCUMergeFlagExtSCModel    ( 1,             1,               NUM_MERGE_FLAG_EXT_CTX        , m_contextModels + m_numContextModels, m_numContextModels)
, m_cCUMergeIdxExtSCModel     ( 1,             1,               NUM_MERGE_IDX_EXT_CTX         , m_contextModels + m_numContextModels, m_numContextModels)
, m_cCUPartSizeSCModel        ( 1,             1,               NUM_PART_SIZE_CTX             , m_contextModels + m_numContextModels, m_numContextModels)
, m_cCUPredModeSCModel        ( 1,             1,               NUM_PRED_MODE_CTX             , m_contextModels + m_numContextModels, m_numContextModels)
, m_cCUAlfCtrlFlagSCModel     ( 1,             1,               NUM_ALF_CTRL_FLAG_CTX         , m_contextModels + m_numContextModels, m_numContextModels)
, m_cCUIntraPredSCModel       ( 1,             1,               NUM_ADI_CTX                   , m_contextModels + m_numContextModels, m_numContextModels)
, m_cCUChromaPredSCModel      ( 1,             1,               NUM_CHROMA_PRED_CTX           , m_contextModels + m_numContextModels, m_numContextModels)
, m_cCUDeltaQpSCModel         ( 1,             1,               NUM_DELTA_QP_CTX              , m_contextModels + m_numContextModels, m_numContextModels)
, m_cCUInterDirSCModel        ( 1,             1,               NUM_INTER_DIR_CTX             , m_contextModels + m_numContextModels, m_numContextModels)
, m_cCURefPicSCModel          ( 1,             1,               NUM_REF_NO_CTX                , m_contextModels + m_numContextModels, m_numContextModels)
, m_cCUMvdSCModel             ( 1,             1,               NUM_MV_RES_CTX                , m_contextModels + m_numContextModels, m_numContextModels)
, m_cCUQtCbfSCModel           ( 1,             2,               NUM_QT_CBF_CTX                , m_contextModels + m_numContextModels, m_numContextModels)
, m_cCUTransSubdivFlagSCModel ( 1,             1,               NUM_TRANS_SUBDIV_FLAG_CTX     , m_contextModels + m_numContextModels, m_numContextModels)
, m_cCUQtRootCbfSCModel       ( 1,             1,               NUM_QT_ROOT_CBF_CTX           , m_contextModels + m_numContextModels, m_numContextModels)
, m_cCUSigCoeffGroupSCModel   ( 1,             2,               NUM_SIG_CG_FLAG_CTX           , m_contextModels + m_numContextModels, m_numContextModels)
, m_cCUSigSCModel             ( 1,             1,               NUM_SIG_FLAG_CTX              , m_contextModels + m_numContextModels, m_numContextModels)
, m_cCuCtxLastX               ( 1,             2,               NUM_CTX_LAST_FLAG_XY          , m_contextModels + m_numContextModels, m_numContextModels)
, m_cCuCtxLastY               ( 1,             2,               NUM_CTX_LAST_FLAG_XY          , m_contextModels + m_numContextModels, m_numContextModels)
, m_cCUOneSCModel             ( 1,             1,               NUM_ONE_FLAG_CTX              , m_contextModels + m_numContextModels, m_numContextModels)
, m_cCUAbsSCModel             ( 1,             1,               NUM_ABS_FLAG_CTX              , m_contextModels + m_numContextModels, m_numContextModels)
, m_cMVPIdxSCModel            ( 1,             1,               NUM_MVP_IDX_CTX               , m_contextModels + m_numContextModels, m_numContextModels)
, m_cALFFlagSCModel           ( 1,             1,               NUM_ALF_FLAG_CTX              , m_contextModels + m_numContextModels, m_numContextModels)
, m_cALFUvlcSCModel           ( 1,             1,               NUM_ALF_UVLC_CTX              , m_contextModels + m_numContextModels, m_numContextModels)
, m_cALFSvlcSCModel           ( 1,             1,               NUM_ALF_SVLC_CTX              , m_contextModels + m_numContextModels, m_numContextModels)
, m_cCUAMPSCModel             ( 1,             1,               NUM_CU_AMP_CTX                , m_contextModels + m_numContextModels, m_numContextModels)
#if !SAO_ABS_BY_PASS
, m_cSaoUvlcSCModel           ( 1,             1,               NUM_SAO_UVLC_CTX              , m_contextModels + m_numContextModels, m_numContextModels)
#endif
#if SAO_MERGE_ONE_CTX
, m_cSaoMergeSCModel          ( 1,             1,               NUM_SAO_MERGE_FLAG_CTX   , m_contextModels + m_numContextModels, m_numContextModels)
#else
, m_cSaoMergeLeftSCModel      ( 1,             1,               NUM_SAO_MERGE_LEFT_FLAG_CTX   , m_contextModels + m_numContextModels, m_numContextModels)
, m_cSaoMergeUpSCModel        ( 1,             1,               NUM_SAO_MERGE_UP_FLAG_CTX     , m_contextModels + m_numContextModels, m_numContextModels)
#endif
#if INTRA_BL
, m_cIntraBLPredFlagSCModel   (1,              1,               NUM_INTRA_BL_PRED_CTX         , m_contextModels + m_numContextModels, m_numContextModels)
#endif
, m_cSaoTypeIdxSCModel        ( 1,             1,               NUM_SAO_TYPE_IDX_CTX          , m_contextModels + m_numContextModels, m_numContextModels)
, m_cTransformSkipSCModel     ( 1,             2,               NUM_TRANSFORMSKIP_FLAG_CTX    , m_contextModels + m_numContextModels, m_numContextModels)
, m_CUTransquantBypassFlagSCModel( 1,          1,               NUM_CU_TRANSQUANT_BYPASS_FLAG_CTX, m_contextModels + m_numContextModels, m_numContextModels)
#if RWTH_CTX_FOR_REFINEMENT
, m_cCURefinementModel        ( 1,             2,               NUM_REFINEMNT_CTX             , m_contextModels + m_numContextModels, m_numContextModels)
#endif
#if RWTH_CODE_LASTXY_DIFF
, m_cCuCtxLastDiffSignalling  ( 1,             2,               NUM_CTX_LAST_FLAG_XY_DIFF_SIG , m_contextModels + m_numContextModels, m_numContextModels)
, m_cCuCtxLastDiffX           ( 1,             2,               NUM_CTX_LAST_FLAG_XY          , m_contextModels + m_numContextModels, m_numContextModels)
, m_cCuCtxLastDiffY           ( 1,             2,               NUM_CTX_LAST_FLAG_XY          , m_contextModels + m_numContextModels, m_numContextModels)
#endif
{
  assert( m_numContextModels <= MAX_NUM_CTX_MOD );
#if !REMOVE_FGS
  m_iSliceGranularity = 0;
#endif
}

TEncSbac::~TEncSbac()
{
}

// ====================================================================================================================
// Public member functions
// ====================================================================================================================

Void TEncSbac::resetEntropy           ()
{
  Int  iQp              = m_pcSlice->getSliceQp();
  SliceType eSliceType  = m_pcSlice->getSliceType();
  
  Int  encCABACTableIdx = m_pcSlice->getPPS()->getEncCABACTableIdx();
  if (!m_pcSlice->isIntra() && (encCABACTableIdx==B_SLICE || encCABACTableIdx==P_SLICE) && m_pcSlice->getPPS()->getCabacInitPresentFlag())
  {
    eSliceType = (SliceType) encCABACTableIdx;
  }

  m_cCUSplitFlagSCModel.initBuffer       ( eSliceType, iQp, (UChar*)INIT_SPLIT_FLAG );
  
  m_cCUSkipFlagSCModel.initBuffer        ( eSliceType, iQp, (UChar*)INIT_SKIP_FLAG );
  m_cCUAlfCtrlFlagSCModel.initBuffer     ( eSliceType, iQp, (UChar*)INIT_ALF_CTRL_FLAG );
  m_cCUMergeFlagExtSCModel.initBuffer    ( eSliceType, iQp, (UChar*)INIT_MERGE_FLAG_EXT);
  m_cCUMergeIdxExtSCModel.initBuffer     ( eSliceType, iQp, (UChar*)INIT_MERGE_IDX_EXT);
  m_cCUPartSizeSCModel.initBuffer        ( eSliceType, iQp, (UChar*)INIT_PART_SIZE );
  m_cCUAMPSCModel.initBuffer             ( eSliceType, iQp, (UChar*)INIT_CU_AMP_POS );
  m_cCUPredModeSCModel.initBuffer        ( eSliceType, iQp, (UChar*)INIT_PRED_MODE );
  m_cCUIntraPredSCModel.initBuffer       ( eSliceType, iQp, (UChar*)INIT_INTRA_PRED_MODE );
  m_cCUChromaPredSCModel.initBuffer      ( eSliceType, iQp, (UChar*)INIT_CHROMA_PRED_MODE );
  m_cCUInterDirSCModel.initBuffer        ( eSliceType, iQp, (UChar*)INIT_INTER_DIR );
  m_cCUMvdSCModel.initBuffer             ( eSliceType, iQp, (UChar*)INIT_MVD );
  m_cCURefPicSCModel.initBuffer          ( eSliceType, iQp, (UChar*)INIT_REF_PIC );
  m_cCUDeltaQpSCModel.initBuffer         ( eSliceType, iQp, (UChar*)INIT_DQP );
  m_cCUQtCbfSCModel.initBuffer           ( eSliceType, iQp, (UChar*)INIT_QT_CBF );
  m_cCUQtRootCbfSCModel.initBuffer       ( eSliceType, iQp, (UChar*)INIT_QT_ROOT_CBF );
  m_cCUSigCoeffGroupSCModel.initBuffer   ( eSliceType, iQp, (UChar*)INIT_SIG_CG_FLAG );
  m_cCUSigSCModel.initBuffer             ( eSliceType, iQp, (UChar*)INIT_SIG_FLAG );
  m_cCuCtxLastX.initBuffer               ( eSliceType, iQp, (UChar*)INIT_LAST );
  m_cCuCtxLastY.initBuffer               ( eSliceType, iQp, (UChar*)INIT_LAST );
  m_cCUOneSCModel.initBuffer             ( eSliceType, iQp, (UChar*)INIT_ONE_FLAG );
  m_cCUAbsSCModel.initBuffer             ( eSliceType, iQp, (UChar*)INIT_ABS_FLAG );
  m_cMVPIdxSCModel.initBuffer            ( eSliceType, iQp, (UChar*)INIT_MVP_IDX );
  m_cALFFlagSCModel.initBuffer           ( eSliceType, iQp, (UChar*)INIT_ALF_FLAG );
  m_cALFUvlcSCModel.initBuffer           ( eSliceType, iQp, (UChar*)INIT_ALF_UVLC );
  m_cALFSvlcSCModel.initBuffer           ( eSliceType, iQp, (UChar*)INIT_ALF_SVLC );
  m_cCUTransSubdivFlagSCModel.initBuffer ( eSliceType, iQp, (UChar*)INIT_TRANS_SUBDIV_FLAG );
#if !SAO_ABS_BY_PASS
  m_cSaoUvlcSCModel.initBuffer           ( eSliceType, iQp, (UChar*)INIT_SAO_UVLC );
#endif
#if SAO_MERGE_ONE_CTX
  m_cSaoMergeSCModel.initBuffer      ( eSliceType, iQp, (UChar*)INIT_SAO_MERGE_FLAG );
#else
  m_cSaoMergeLeftSCModel.initBuffer      ( eSliceType, iQp, (UChar*)INIT_SAO_MERGE_LEFT_FLAG );
  m_cSaoMergeUpSCModel.initBuffer        ( eSliceType, iQp, (UChar*)INIT_SAO_MERGE_UP_FLAG );
#endif
#if INTRA_BL
  m_cIntraBLPredFlagSCModel.initBuffer ( eSliceType, iQp, (UChar*)INIT_INTRA_BL_PRED_FLAG );
#endif
  m_cSaoTypeIdxSCModel.initBuffer        ( eSliceType, iQp, (UChar*)INIT_SAO_TYPE_IDX );
  m_cTransformSkipSCModel.initBuffer     ( eSliceType, iQp, (UChar*)INIT_TRANSFORMSKIP_FLAG );
  m_CUTransquantBypassFlagSCModel.initBuffer( eSliceType, iQp, (UChar*)INIT_CU_TRANSQUANT_BYPASS_FLAG );
#if RWTH_CTX_FOR_REFINEMENT
  m_cCURefinementModel.initBuffer        ( eSliceType, iQp, (UChar*)INIT_REFINEMENT_FLAG );
#endif
#if RWTH_CODE_LASTXY_DIFF
  m_cCuCtxLastDiffSignalling.initBuffer  ( eSliceType, iQp, (UChar*)INIT_LAST_XY_DIFF_SIGNALLING );
  m_cCuCtxLastDiffX.initBuffer           ( eSliceType, iQp, (UChar*)INIT_LAST );
  m_cCuCtxLastDiffY.initBuffer           ( eSliceType, iQp, (UChar*)INIT_LAST );
#endif
  // new structure
  m_uiLastQp = iQp;
  
  m_pcBinIf->start();
  
  return;
}

/** The function does the following: 
 * If current slice type is P/B then it determines the distance of initialisation type 1 and 2 from the current CABAC states and 
 * stores the index of the closest table.  This index is used for the next P/B slice when cabac_init_present_flag is true.
 */
Void TEncSbac::determineCabacInitIdx()
{
  Int  qp              = m_pcSlice->getSliceQp();

  if (!m_pcSlice->isIntra())
  {
    SliceType aSliceTypeChoices[] = {B_SLICE, P_SLICE};

    UInt bestCost             = MAX_UINT;
    SliceType bestSliceType   = aSliceTypeChoices[0];
    for (UInt idx=0; idx<2; idx++)
    {
      UInt curCost          = 0;
      SliceType curSliceType  = aSliceTypeChoices[idx];

      curCost  = m_cCUSplitFlagSCModel.calcCost       ( curSliceType, qp, (UChar*)INIT_SPLIT_FLAG );
      curCost += m_cCUSkipFlagSCModel.calcCost        ( curSliceType, qp, (UChar*)INIT_SKIP_FLAG );
#if INTRA_BL
      curCost += m_cIntraBLPredFlagSCModel.calcCost   ( curSliceType, qp, (UChar*)INIT_INTRA_BL_PRED_FLAG );
#endif
      curCost += m_cCUAlfCtrlFlagSCModel.calcCost     ( curSliceType, qp, (UChar*)INIT_ALF_CTRL_FLAG );
      curCost += m_cCUMergeFlagExtSCModel.calcCost    ( curSliceType, qp, (UChar*)INIT_MERGE_FLAG_EXT);
      curCost += m_cCUMergeIdxExtSCModel.calcCost     ( curSliceType, qp, (UChar*)INIT_MERGE_IDX_EXT);
      curCost += m_cCUPartSizeSCModel.calcCost        ( curSliceType, qp, (UChar*)INIT_PART_SIZE );
      curCost += m_cCUAMPSCModel.calcCost             ( curSliceType, qp, (UChar*)INIT_CU_AMP_POS );
      curCost += m_cCUPredModeSCModel.calcCost        ( curSliceType, qp, (UChar*)INIT_PRED_MODE );
      curCost += m_cCUIntraPredSCModel.calcCost       ( curSliceType, qp, (UChar*)INIT_INTRA_PRED_MODE );
      curCost += m_cCUChromaPredSCModel.calcCost      ( curSliceType, qp, (UChar*)INIT_CHROMA_PRED_MODE );
      curCost += m_cCUInterDirSCModel.calcCost        ( curSliceType, qp, (UChar*)INIT_INTER_DIR );
      curCost += m_cCUMvdSCModel.calcCost             ( curSliceType, qp, (UChar*)INIT_MVD );
      curCost += m_cCURefPicSCModel.calcCost          ( curSliceType, qp, (UChar*)INIT_REF_PIC );
      curCost += m_cCUDeltaQpSCModel.calcCost         ( curSliceType, qp, (UChar*)INIT_DQP );
      curCost += m_cCUQtCbfSCModel.calcCost           ( curSliceType, qp, (UChar*)INIT_QT_CBF );
      curCost += m_cCUQtRootCbfSCModel.calcCost       ( curSliceType, qp, (UChar*)INIT_QT_ROOT_CBF );
      curCost += m_cCUSigCoeffGroupSCModel.calcCost   ( curSliceType, qp, (UChar*)INIT_SIG_CG_FLAG );
      curCost += m_cCUSigSCModel.calcCost             ( curSliceType, qp, (UChar*)INIT_SIG_FLAG );
      curCost += m_cCuCtxLastX.calcCost               ( curSliceType, qp, (UChar*)INIT_LAST );
      curCost += m_cCuCtxLastY.calcCost               ( curSliceType, qp, (UChar*)INIT_LAST );
      curCost += m_cCUOneSCModel.calcCost             ( curSliceType, qp, (UChar*)INIT_ONE_FLAG );
      curCost += m_cCUAbsSCModel.calcCost             ( curSliceType, qp, (UChar*)INIT_ABS_FLAG );
      curCost += m_cMVPIdxSCModel.calcCost            ( curSliceType, qp, (UChar*)INIT_MVP_IDX );
      curCost += m_cALFFlagSCModel.calcCost           ( curSliceType, qp, (UChar*)INIT_ALF_FLAG );
      curCost += m_cALFUvlcSCModel.calcCost           ( curSliceType, qp, (UChar*)INIT_ALF_UVLC );
      curCost += m_cALFSvlcSCModel.calcCost           ( curSliceType, qp, (UChar*)INIT_ALF_SVLC );
      curCost += m_cCUTransSubdivFlagSCModel.calcCost ( curSliceType, qp, (UChar*)INIT_TRANS_SUBDIV_FLAG );
#if !SAO_ABS_BY_PASS
      curCost += m_cSaoUvlcSCModel.calcCost           ( curSliceType, qp, (UChar*)INIT_SAO_UVLC );
#endif
#if SAO_MERGE_ONE_CTX
      curCost += m_cSaoMergeSCModel.calcCost      ( curSliceType, qp, (UChar*)INIT_SAO_MERGE_FLAG );
#else
      curCost += m_cSaoMergeLeftSCModel.calcCost      ( curSliceType, qp, (UChar*)INIT_SAO_MERGE_LEFT_FLAG );
      curCost += m_cSaoMergeUpSCModel.calcCost        ( curSliceType, qp, (UChar*)INIT_SAO_MERGE_UP_FLAG );
#endif
      curCost += m_cSaoTypeIdxSCModel.calcCost        ( curSliceType, qp, (UChar*)INIT_SAO_TYPE_IDX );
      curCost += m_cTransformSkipSCModel.calcCost     ( curSliceType, qp, (UChar*)INIT_TRANSFORMSKIP_FLAG );
      curCost += m_CUTransquantBypassFlagSCModel.calcCost( curSliceType, qp, (UChar*)INIT_CU_TRANSQUANT_BYPASS_FLAG );
#if RWTH_CTX_FOR_REFINEMENT
      curCost += m_cCURefinementModel.calcCost        ( curSliceType, qp, (UChar*)INIT_REFINEMENT_FLAG );
#endif
#if RWTH_CODE_LASTXY_DIFF
      curCost += m_cCuCtxLastDiffSignalling.calcCost  ( curSliceType, qp, (UChar*)INIT_LAST_XY_DIFF_SIGNALLING );
      curCost += m_cCuCtxLastDiffX.calcCost           ( curSliceType, qp, (UChar*)INIT_LAST );
      curCost += m_cCuCtxLastDiffY.calcCost           ( curSliceType, qp, (UChar*)INIT_LAST );
#endif
      if (curCost < bestCost)
      {
        bestSliceType = curSliceType;
        bestCost      = curCost;
      }
    }
    m_pcSlice->getPPS()->setEncCABACTableIdx( bestSliceType );
  }
  else
  {
    m_pcSlice->getPPS()->setEncCABACTableIdx( I_SLICE );
  }  
}


/** The function does the followng: Write out terminate bit. Flush CABAC. Intialize CABAC states. Start CABAC.
 */
Void TEncSbac::updateContextTables( SliceType eSliceType, Int iQp, Bool bExecuteFinish )
{
  codeTerminatingBit( 1 );
  DTRACE_CABAC_N

  if (bExecuteFinish) m_pcBinIf->finish();
  m_cCUSplitFlagSCModel.initBuffer       ( eSliceType, iQp, (UChar*)INIT_SPLIT_FLAG );
  
  m_cCUSkipFlagSCModel.initBuffer        ( eSliceType, iQp, (UChar*)INIT_SKIP_FLAG );
  m_cCUAlfCtrlFlagSCModel.initBuffer     ( eSliceType, iQp, (UChar*)INIT_ALF_CTRL_FLAG );
  m_cCUMergeFlagExtSCModel.initBuffer    ( eSliceType, iQp, (UChar*)INIT_MERGE_FLAG_EXT);
  m_cCUMergeIdxExtSCModel.initBuffer     ( eSliceType, iQp, (UChar*)INIT_MERGE_IDX_EXT);
  m_cCUPartSizeSCModel.initBuffer        ( eSliceType, iQp, (UChar*)INIT_PART_SIZE );
  m_cCUAMPSCModel.initBuffer             ( eSliceType, iQp, (UChar*)INIT_CU_AMP_POS );
  m_cCUPredModeSCModel.initBuffer        ( eSliceType, iQp, (UChar*)INIT_PRED_MODE );
  m_cCUIntraPredSCModel.initBuffer       ( eSliceType, iQp, (UChar*)INIT_INTRA_PRED_MODE );
  m_cCUChromaPredSCModel.initBuffer      ( eSliceType, iQp, (UChar*)INIT_CHROMA_PRED_MODE );
  m_cCUInterDirSCModel.initBuffer        ( eSliceType, iQp, (UChar*)INIT_INTER_DIR );
  m_cCUMvdSCModel.initBuffer             ( eSliceType, iQp, (UChar*)INIT_MVD );
  m_cCURefPicSCModel.initBuffer          ( eSliceType, iQp, (UChar*)INIT_REF_PIC );
  m_cCUDeltaQpSCModel.initBuffer         ( eSliceType, iQp, (UChar*)INIT_DQP );
  m_cCUQtCbfSCModel.initBuffer           ( eSliceType, iQp, (UChar*)INIT_QT_CBF );
  m_cCUQtRootCbfSCModel.initBuffer       ( eSliceType, iQp, (UChar*)INIT_QT_ROOT_CBF );
  m_cCUSigCoeffGroupSCModel.initBuffer   ( eSliceType, iQp, (UChar*)INIT_SIG_CG_FLAG );
  m_cCUSigSCModel.initBuffer             ( eSliceType, iQp, (UChar*)INIT_SIG_FLAG );
  m_cCuCtxLastX.initBuffer               ( eSliceType, iQp, (UChar*)INIT_LAST );
  m_cCuCtxLastY.initBuffer               ( eSliceType, iQp, (UChar*)INIT_LAST );
  m_cCUOneSCModel.initBuffer             ( eSliceType, iQp, (UChar*)INIT_ONE_FLAG );
  m_cCUAbsSCModel.initBuffer             ( eSliceType, iQp, (UChar*)INIT_ABS_FLAG );
  m_cMVPIdxSCModel.initBuffer            ( eSliceType, iQp, (UChar*)INIT_MVP_IDX );
  m_cALFFlagSCModel.initBuffer           ( eSliceType, iQp, (UChar*)INIT_ALF_FLAG );
  m_cALFUvlcSCModel.initBuffer           ( eSliceType, iQp, (UChar*)INIT_ALF_UVLC );
  m_cALFSvlcSCModel.initBuffer           ( eSliceType, iQp, (UChar*)INIT_ALF_SVLC );
  m_cCUTransSubdivFlagSCModel.initBuffer ( eSliceType, iQp, (UChar*)INIT_TRANS_SUBDIV_FLAG );
#if !SAO_ABS_BY_PASS
  m_cSaoUvlcSCModel.initBuffer           ( eSliceType, iQp, (UChar*)INIT_SAO_UVLC );
#endif
#if SAO_MERGE_ONE_CTX
  m_cSaoMergeSCModel.initBuffer      ( eSliceType, iQp, (UChar*)INIT_SAO_MERGE_FLAG );
#else
  m_cSaoMergeLeftSCModel.initBuffer      ( eSliceType, iQp, (UChar*)INIT_SAO_MERGE_LEFT_FLAG );
  m_cSaoMergeUpSCModel.initBuffer        ( eSliceType, iQp, (UChar*)INIT_SAO_MERGE_UP_FLAG );
#endif

#if INTRA_BL
  m_cIntraBLPredFlagSCModel.initBuffer ( eSliceType, iQp, (UChar*)INIT_INTRA_BL_PRED_FLAG );
#endif  
  m_cSaoTypeIdxSCModel.initBuffer        ( eSliceType, iQp, (UChar*)INIT_SAO_TYPE_IDX );
  m_cTransformSkipSCModel.initBuffer     ( eSliceType, iQp, (UChar*)INIT_TRANSFORMSKIP_FLAG );
  m_CUTransquantBypassFlagSCModel.initBuffer( eSliceType, iQp, (UChar*)INIT_CU_TRANSQUANT_BYPASS_FLAG );
#if RWTH_CTX_FOR_REFINEMENT
  m_cCURefinementModel.initBuffer        ( eSliceType, iQp, (UChar*)INIT_REFINEMENT_FLAG );
#endif
#if RWTH_CODE_LASTXY_DIFF
  m_cCuCtxLastDiffSignalling.initBuffer  ( eSliceType, iQp, (UChar*)INIT_LAST_XY_DIFF_SIGNALLING );
  m_cCuCtxLastDiffX.initBuffer           ( eSliceType, iQp, (UChar*)INIT_LAST );
  m_cCuCtxLastDiffY.initBuffer           ( eSliceType, iQp, (UChar*)INIT_LAST );
#endif

  m_pcBinIf->start();
}

Void TEncSbac::codeVPS( TComVPS* pcVPS )
{
  assert (0);
  return;
}

Void TEncSbac::codeSPS( TComSPS* pcSPS )
{
  assert (0);
  return;
}

Void TEncSbac::codePPS( TComPPS* pcPPS )
{
  assert (0);
  return;
}

Void TEncSbac::codeSliceHeader( TComSlice* pcSlice )
{
  assert (0);
  return;
}

Void TEncSbac::codeTilesWPPEntryPoint( TComSlice* pSlice )
{
  assert (0);
  return;
}

Void TEncSbac::codeTerminatingBit( UInt uilsLast )
{
  m_pcBinIf->encodeBinTrm( uilsLast );
  DTRACE_CABAC_N
}

Void TEncSbac::codeSliceFinish()
{
  m_pcBinIf->finish();
}

Void TEncSbac::codeFlush()
{
  m_pcBinIf->flush();
}

Void TEncSbac::encodeStart()
{
  m_pcBinIf->start();
}

Void TEncSbac::xWriteUnarySymbol( UInt uiSymbol, ContextModel* pcSCModel, Int iOffset )
{
#if ENC_DEC_TRACE
  DTRACE_CABAC_T( "\tStart Coding UnarySymbol - iOffset: " )
  DTRACE_CABAC_V( iOffset )
  DTRACE_CABAC_N
  UInt uiInitialSymbol = uiSymbol;
#endif

  m_pcBinIf->encodeBin( uiSymbol ? 1 : 0, pcSCModel[0] );
  DTRACE_CABAC_N
  
  if( 0 == uiSymbol)
  {
    DTRACE_CABAC_T( "\tEnd Coding UnarySymbol - ruiSymbol: 0\n")
    return;
  }
  
  while( uiSymbol-- )
  {
    m_pcBinIf->encodeBin( uiSymbol ? 1 : 0, pcSCModel[ iOffset ] );
    DTRACE_CABAC_T( "\tctx:" )
    DTRACE_CABAC_V( iOffset )
    DTRACE_CABAC_N
  }
  
#if ENC_DEC_TRACE
  DTRACE_CABAC_T( "\tEnd Coding UnarySymbol - ruiSymbol: ")
  DTRACE_CABAC_V( uiInitialSymbol )
  DTRACE_CABAC_N
#endif
}

Void TEncSbac::xWriteUnaryMaxSymbol( UInt uiSymbol, ContextModel* pcSCModel, Int iOffset, UInt uiMaxSymbol )
{
#if ENC_DEC_TRACE
  DTRACE_CABAC_T( "\tStart Coding UnaryMaxSymbol - iOffset: " )
  DTRACE_CABAC_V( iOffset )
  DTRACE_CABAC_T( " - uiMaxSymbol: " )
  DTRACE_CABAC_V( uiMaxSymbol )
  DTRACE_CABAC_N
  UInt uiInitialSymbol = uiSymbol;
#endif

  if (uiMaxSymbol == 0)
  {
    DTRACE_CABAC_T( "\tEnd Coding UnaryMaxSymbol - ruiSymbol: 0\n")
    return;
  }
  
  m_pcBinIf->encodeBin( uiSymbol ? 1 : 0, pcSCModel[ 0 ] );
  DTRACE_CABAC_T( "\tctx:0\n" )
  
  if ( uiSymbol == 0 )
  {
    DTRACE_CABAC_T( "\tEnd Coding UnaryMaxSymbol - ruiSymbol: " )
    DTRACE_CABAC_V( uiSymbol )
    DTRACE_CABAC_N
    return;
  }
  
  Bool bCodeLast = ( uiMaxSymbol > uiSymbol );
  
  while( --uiSymbol )
  {
    m_pcBinIf->encodeBin( 1, pcSCModel[ iOffset ] );
    DTRACE_CABAC_T( "\tctx:" )
    DTRACE_CABAC_V( iOffset )
    DTRACE_CABAC_N
  }
  if( bCodeLast )
  {
    m_pcBinIf->encodeBin( 0, pcSCModel[ iOffset ] );
    DTRACE_CABAC_T( "\tctx:" )
    DTRACE_CABAC_V( iOffset )
    DTRACE_CABAC_N
  }

#if ENC_DEC_TRACE
  DTRACE_CABAC_T( "\tEnd Coding UnaryMaxSymbol - ruiSymbol: ")
  DTRACE_CABAC_V( uiInitialSymbol )
  DTRACE_CABAC_N
#endif
}

Void TEncSbac::xWriteEpExGolomb( UInt uiSymbol, UInt uiCount )
{
#if ENC_DEC_TRACE
  DTRACE_CABAC_T( "\tStart Coding EpExGolomb - uiCount: " )
  DTRACE_CABAC_V( uiCount )
  DTRACE_CABAC_N
  UInt uiInitialSymbol = uiSymbol;
#endif

  UInt bins = 0;
  Int numBins = 0;
  
  while( uiSymbol >= (UInt)(1<<uiCount) )
  {
    bins = 2 * bins + 1;
    numBins++;
    uiSymbol -= 1 << uiCount;
    uiCount  ++;
  }
  bins = 2 * bins + 0;
  numBins++;
  
  bins = (bins << uiCount) | uiSymbol;
  numBins += uiCount;
  
  assert( numBins <= 32 );
  m_pcBinIf->encodeBinsEP( bins, numBins );

#if ENC_DEC_TRACE
  DTRACE_CABAC_T( "\tEnd Coding EpExGolomb - ruiSymbol: ")
  DTRACE_CABAC_V( uiInitialSymbol )
  DTRACE_CABAC_N
#endif
}

/** Coding of coeff_abs_level_minus3
 * \param uiSymbol value of coeff_abs_level_minus3
 * \param ruiGoRiceParam reference to Rice parameter
 * \returns Void
 */
Void TEncSbac::xWriteCoefRemainExGolomb ( UInt symbol, UInt &rParam )
{
#if ENC_DEC_TRACE
  DTRACE_CABAC_T( "\tStart Coding GoRiceExGolomb - ruiGoRiceParam: " )
  DTRACE_CABAC_V( rParam )
  DTRACE_CABAC_N
#endif

  Int codeNumber  = (Int)symbol;
  UInt length;
#if COEF_REMAIN_BIN_REDUCTION
  if (codeNumber < (COEF_REMAIN_BIN_REDUCTION << rParam))
#else
  if (codeNumber < (8 << rParam))
#endif
  {
    length = codeNumber>>rParam;
    m_pcBinIf->encodeBinsEP( (1<<(length+1))-2 , length+1);
    m_pcBinIf->encodeBinsEP((codeNumber%(1<<rParam)),rParam);
  }
  else
  {
    length = rParam;
#if COEF_REMAIN_BIN_REDUCTION
    codeNumber  = codeNumber - ( COEF_REMAIN_BIN_REDUCTION << rParam);
#else
    codeNumber  = codeNumber - ( 8 << rParam);    
#endif
    while (codeNumber >= (1<<length))
    {
      codeNumber -=  (1<<(length++));    
    }
#if COEF_REMAIN_BIN_REDUCTION
    m_pcBinIf->encodeBinsEP((1<<(COEF_REMAIN_BIN_REDUCTION+length+1-rParam))-2,COEF_REMAIN_BIN_REDUCTION+length+1-rParam);
#else
    m_pcBinIf->encodeBinsEP((1<<(8+length+1-rParam))-2,8+length+1-rParam);
#endif
    m_pcBinIf->encodeBinsEP(codeNumber,length);
  }
  DTRACE_CABAC_T( "\tEnd Coding GoRiceExGolomb - ruiGoRiceParam: " )
  DTRACE_CABAC_V( rParam )
  DTRACE_CABAC_T( "\truiSymbol: ")
  DTRACE_CABAC_V( symbol )
  DTRACE_CABAC_N
}

// SBAC RD
Void  TEncSbac::load ( TEncSbac* pSrc)
{
  this->xCopyFrom(pSrc);
}

Void  TEncSbac::loadIntraDirModeLuma( TEncSbac* pSrc)
{
  m_pcBinIf->copyState( pSrc->m_pcBinIf );
  
  this->m_cCUIntraPredSCModel      .copyFrom( &pSrc->m_cCUIntraPredSCModel       );
}


Void  TEncSbac::store( TEncSbac* pDest)
{
  pDest->xCopyFrom( this );
}


Void TEncSbac::xCopyFrom( TEncSbac* pSrc )
{
  m_pcBinIf->copyState( pSrc->m_pcBinIf );
  
  this->m_uiCoeffCost = pSrc->m_uiCoeffCost;
  this->m_uiLastQp    = pSrc->m_uiLastQp;
  
  memcpy( m_contextModels, pSrc->m_contextModels, m_numContextModels * sizeof( ContextModel ) );
}

Void TEncSbac::codeMVPIdx ( TComDataCU* pcCU, UInt uiAbsPartIdx, RefPicList eRefList )
{
  DTRACE_CABAC_T( "\tStart Code MVPIdx\n" )

  Int iSymbol = pcCU->getMVPIdx(eRefList, uiAbsPartIdx);
  Int iNum = AMVP_MAX_NUM_CANDS;

  xWriteUnaryMaxSymbol(iSymbol, m_cMVPIdxSCModel.get(0), 1, iNum-1);
  
#if ENC_DEC_TRACE
  DTRACE_CABAC_T( "\tEnd Code MVPIdx - riMVPIdx=" )
  DTRACE_CABAC_V( iSymbol )
  //DTRACE_CABAC_T( "\tAMVP: " )
  //DTRACE_CABAC_T( pcCU->getAMVPString( eRefList, uiAbsPartIdx ).c_str() )
  DTRACE_CABAC_N
#endif

}

Void TEncSbac::codePartSize( TComDataCU* pcCU, UInt uiAbsPartIdx, UInt uiDepth )
{
#if RWTH_RESI_REFINEMENT
  if (pcCU->getPredictionMode(uiAbsPartIdx)==MODE_IL_RWTH_RESI_REFINEMENT || pcCU->getPredictionMode(uiAbsPartIdx)==MODE_IL_RESI_NO_REFINEMENT)
  {
    return;
  }
#endif

  DTRACE_CABAC_T( "\tStart Code PartSize" );
  DTRACE_CABAC_T( "\tuiAbsPartIdx=" );
  DTRACE_CABAC_V( uiAbsPartIdx )
  DTRACE_CABAC_T( "\tuiDepth=" );
  DTRACE_CABAC_V( uiDepth )
  DTRACE_CABAC_N

  PartSize eSize = pcCU->getPartitionSize( uiAbsPartIdx );
#if INTRA_BL
  assert( !pcCU->isIntraBL( uiAbsPartIdx ) );
#endif

  if ( pcCU->isIntra( uiAbsPartIdx ) )
  {
    if( uiDepth == g_uiMaxCUDepth - g_uiAddCUDepth )
    {
      m_pcBinIf->encodeBin( eSize == SIZE_2Nx2N? 1 : 0, m_cCUPartSizeSCModel.get( 0, 0, 0 ) );
      DTRACE_CABAC_N
    }
  }
  else // Inter
  {
    switch(eSize)
    {
      case SIZE_2Nx2N:
      {
        m_pcBinIf->encodeBin( 1, m_cCUPartSizeSCModel.get( 0, 0, 0) );
        DTRACE_CABAC_N
        break;
      }
      case SIZE_2NxN:
      case SIZE_2NxnU:
      case SIZE_2NxnD:
      {
        m_pcBinIf->encodeBin( 0, m_cCUPartSizeSCModel.get( 0, 0, 0) );
          DTRACE_CABAC_N
        m_pcBinIf->encodeBin( 1, m_cCUPartSizeSCModel.get( 0, 0, 1) );
          DTRACE_CABAC_N
        if ( pcCU->getSlice()->getSPS()->getAMPAcc( uiDepth ) )
        {
          if (eSize == SIZE_2NxN)
          {
            m_pcBinIf->encodeBin(1, m_cCUAMPSCModel.get( 0, 0, 0 ));
            DTRACE_CABAC_N
          }
          else
          {
            m_pcBinIf->encodeBin(0, m_cCUAMPSCModel.get( 0, 0, 0 ));          
            DTRACE_CABAC_N
            m_pcBinIf->encodeBinEP((eSize == SIZE_2NxnU? 0: 1));
            DTRACE_CABAC_N
          }
        }
        break;
      }
      case SIZE_Nx2N:
      case SIZE_nLx2N:
      case SIZE_nRx2N:
      {
        m_pcBinIf->encodeBin( 0, m_cCUPartSizeSCModel.get( 0, 0, 0) );
        DTRACE_CABAC_N
        m_pcBinIf->encodeBin( 0, m_cCUPartSizeSCModel.get( 0, 0, 1) );
        DTRACE_CABAC_N
        if( uiDepth == g_uiMaxCUDepth - g_uiAddCUDepth && !( pcCU->getWidth(uiAbsPartIdx) == 8 && pcCU->getHeight(uiAbsPartIdx) == 8 ) )
        {
          m_pcBinIf->encodeBin( 1, m_cCUPartSizeSCModel.get( 0, 0, 2) );
        }
        if ( pcCU->getSlice()->getSPS()->getAMPAcc( uiDepth ) )
        {
          if (eSize == SIZE_Nx2N)
          {
            m_pcBinIf->encodeBin(1, m_cCUAMPSCModel.get( 0, 0, 0 ));
            DTRACE_CABAC_N
          }
          else
          {
            m_pcBinIf->encodeBin(0, m_cCUAMPSCModel.get( 0, 0, 0 ));
              DTRACE_CABAC_N
            m_pcBinIf->encodeBinEP((eSize == SIZE_nLx2N? 0: 1));
            DTRACE_CABAC_N
          }
        }
        break;
      }
      case SIZE_NxN:
      {
        if( uiDepth == g_uiMaxCUDepth - g_uiAddCUDepth && !( pcCU->getWidth(uiAbsPartIdx) == 8 && pcCU->getHeight(uiAbsPartIdx) == 8 ) )
        {
          m_pcBinIf->encodeBin( 0, m_cCUPartSizeSCModel.get( 0, 0, 0) );
            DTRACE_CABAC_N
          m_pcBinIf->encodeBin( 0, m_cCUPartSizeSCModel.get( 0, 0, 1) );
            DTRACE_CABAC_N
          m_pcBinIf->encodeBin( 0, m_cCUPartSizeSCModel.get( 0, 0, 2) );
            DTRACE_CABAC_N
        }
        break;
      }
      default:
      {
        assert(0);
      }
    }
  }

  char eModeName[12];
  switch (eSize)
  {
      case SIZE_2Nx2N: sprintf(eModeName, "SIZE_2Nx2N"); break;
      case SIZE_2NxN:  sprintf(eModeName, "SIZE_2NxN" ); break;
      case SIZE_Nx2N:  sprintf(eModeName, "SIZE_Nx2N" ); break;
      case SIZE_NxN:   sprintf(eModeName, "SIZE_NxN"  ); break;
      case SIZE_2NxnU: sprintf(eModeName, "SIZE_2NxnU"); break;
      case SIZE_2NxnD: sprintf(eModeName, "SIZE_2NxnD"); break;
      case SIZE_nLx2N: sprintf(eModeName, "SIZE_nLx2N"); break;
      case SIZE_nRx2N: sprintf(eModeName, "SIZE_nRx2N"); break;
      default:         sprintf(eModeName, "Unknown"   ); break;
  }

  DTRACE_CABAC_T( "\tEnd Code PartSize" );
  DTRACE_CABAC_T( "\teMode=" );
  DTRACE_CABAC_V( eSize )
  DTRACE_CABAC_T( "\teModeName=" );
  DTRACE_CABAC_T( eModeName )
  if (pcCU->isIntra( uiAbsPartIdx ))
  {
    DTRACE_CABAC_T( "\tpredMode=Intra\n" )
  }
  else
  {
    DTRACE_CABAC_T( "\tpredMode=Inter\n" )
  }
}

/** code prediction mode
 * \param pcCU
 * \param uiAbsPartIdx  
 * \returns Void
 */
Void TEncSbac::codePredMode( TComDataCU* pcCU, UInt uiAbsPartIdx )
{
  Int iPredMode = pcCU->getPredictionMode( uiAbsPartIdx );

#if RWTH_RESI_REFINEMENT
  //todo: J: encode resi refinement mode (note: has to be before split flag)
  if(getEncCfg()->getMyCodingArgument()[INTER_LAYER_RWTH_RESI_REFINEMENT])
  {
    if (getEncCfg()->getMyCodingArgument()[TEST_L1_INTRA_PREDICTION] &&
        !getEncCfg()->getMyCodingArgument()[TEST_L1_INTER_PREDICTION] )
    {
      // Intra mode is allwoed in resirefinement but not Inter
      m_pcBinIf->encodeBin( iPredMode != MODE_INTRA ? 1 : 0, m_cCUPredModeSCModel.get( 0, 0, 0 ) );
      DTRACE_CABAC_T( iPredMode != MODE_INTRA ? "\tctx=0\tpred_mode_flag - PredMode=Intra\n" : "\tctx=0\tpred_mode_flag - PredMode=RESI_REFINE\n" );
    }
    else if (!getEncCfg()->getMyCodingArgument()[TEST_L1_INTRA_PREDICTION] &&
              getEncCfg()->getMyCodingArgument()[TEST_L1_INTER_PREDICTION] )
    {
      // Inter mode is allowed in resirefinement but not Intra
      if (pcCU->getSlice()->getSliceType() != I_SLICE)
      {
        // No inter prediction in I_SLICE
        m_pcBinIf->encodeBin( iPredMode == MODE_INTER ? 0 : 1, m_cCUPredModeSCModel.get( 0, 0, 1 ) );
        DTRACE_CABAC_T( iPredMode == MODE_INTER ? "\tctx=1\tpred_mode_flag - PredMode=Inter\n" : "\tctx=1\tpred_mode_flag - PredMode=RESI_REFINE\n" );
      }
    }
    else if (getEncCfg()->getMyCodingArgument()[TEST_L1_INTRA_PREDICTION] &&
             getEncCfg()->getMyCodingArgument()[TEST_L1_INTER_PREDICTION] )
    {
      // Intra and Inter are allower
      Bool bResiRefine = (iPredMode == MODE_IL_RESI_NO_REFINEMENT || iPredMode == MODE_IL_RWTH_RESI_REFINEMENT);
      m_pcBinIf->encodeBin( bResiRefine ? 0 : 1, m_cCUPredModeSCModel.get( 0, 0, 0 ) );
      if (pcCU->getSlice()->getSliceType() == I_SLICE)
      {
        DTRACE_CABAC_T( bResiRefine ? "\tctx=0\tpred_mode_flag - PredMode=RESI_REFINE\n" : "\tctx=0\tpred_mode_flag - PredMode=Intra\n" );
      }
      else
      {
        DTRACE_CABAC_T( bResiRefine ? "\tctx=0\tpred_mode_flag - PredMode=RESI_REFINE\n" : "\tctx=0\tpred_mode_flag - PredMode=Inter/Intra\n" );
        if (!bResiRefine)
        {
          m_pcBinIf->encodeBin( iPredMode == MODE_INTER ? 0 : 1, m_cCUPredModeSCModel.get( 0, 0, 1 ) );
          DTRACE_CABAC_T( iPredMode == MODE_INTER ? "\tctx=1\tpred_mode_flag - PredMode=Inter\n" : "\tctx=1\tpred_mode_flag - PredMode=Intra\n" );
        }
      }
    }
    else
    {
      // Only resi refinement mode is allowed. Nothing to code.
    }
    return;
  }
#endif

  // get context function is here
  m_pcBinIf->encodeBin( iPredMode == MODE_INTER ? 0 : 1, m_cCUPredModeSCModel.get( 0, 0, 0 ) );
  DTRACE_CABAC_T( iPredMode == MODE_INTRA ? "\tpred_mode_flag\tPredMode=Intra\n" : "\tpred_mode_flag\tPredMode=Inter\n" );
}

Void TEncSbac::codeCUTransquantBypassFlag( TComDataCU* pcCU, UInt uiAbsPartIdx )
{
  UInt uiSymbol = pcCU->getCUTransquantBypass(uiAbsPartIdx);
  m_pcBinIf->encodeBin( uiSymbol, m_CUTransquantBypassFlagSCModel.get( 0, 0, 0 ) );
  DTRACE_CABAC_T( "\tTransquantBypassFlag\n" )
}

/** code skip flag
 * \param pcCU
 * \param uiAbsPartIdx 
 * \returns Void
 */
Void TEncSbac::codeSkipFlag( TComDataCU* pcCU, UInt uiAbsPartIdx )
{
  // get context function is here
  UInt uiSymbol = pcCU->isSkipped( uiAbsPartIdx ) ? 1 : 0;
  UInt uiCtxSkip = pcCU->getCtxSkipFlag( uiAbsPartIdx ) ;
  m_pcBinIf->encodeBin( uiSymbol, m_cCUSkipFlagSCModel.get( 0, 0, uiCtxSkip ) );

  DTRACE_CABAC_T( "\tSkipFlag" )
  DTRACE_CABAC_T( "\tuiCtxSkip=")
  DTRACE_CABAC_V( uiCtxSkip )
  DTRACE_CABAC_T( "\n")
}

#if INTRA_BL
/** code intra_bl flag
 * \param pcCU
 * \param uiAbsPartIdx 
 * \returns Void
 */
Void TEncSbac::codeIntraBLFlag( TComDataCU* pcCU, UInt uiAbsPartIdx )
{
  // get context function is here
  UInt uiSymbol = pcCU->isIntraBL( uiAbsPartIdx ) ? 1 : 0;

  UInt uiCtxIntraBL = pcCU->getCtxIntraBLFlag( uiAbsPartIdx ) ;
  m_pcBinIf->encodeBin(uiSymbol, m_cIntraBLPredFlagSCModel.get( 0, 0, uiCtxIntraBL )); 

  DTRACE_CABAC_T( "\tIntraBLFlag" );
  DTRACE_CABAC_T( "\tuiCtxIntraBL: ");
  DTRACE_CABAC_V( uiCtxIntraBL );
  DTRACE_CABAC_T( "\tuiSymbol: ");
  DTRACE_CABAC_V( uiSymbol );
  DTRACE_CABAC_T( "\n");
}
#endif

/** code merge flag
 * \param pcCU
 * \param uiAbsPartIdx 
 * \returns Void
 */
Void TEncSbac::codeMergeFlag( TComDataCU* pcCU, UInt uiAbsPartIdx )
{
  const UInt uiSymbol = pcCU->getMergeFlag( uiAbsPartIdx ) ? 1 : 0;
  m_pcBinIf->encodeBin( uiSymbol, *m_cCUMergeFlagExtSCModel.get( 0 ) );

  DTRACE_CABAC_T( "\tMergeFlag \tAddress=" );
  DTRACE_CABAC_V( pcCU->getAddr() );
  DTRACE_CABAC_T( "\tuiAbsPartIdx=" );
  DTRACE_CABAC_V( uiAbsPartIdx );
  DTRACE_CABAC_N;
}

/** code merge index
 * \param pcCU
 * \param uiAbsPartIdx 
 * \returns Void
 */
Void TEncSbac::codeMergeIndex( TComDataCU* pcCU, UInt uiAbsPartIdx )
{
  DTRACE_CABAC_T( "\tStart Code MergeIndex\tuiAbsPartIdx=" )
  DTRACE_CABAC_V( uiAbsPartIdx )
  DTRACE_CABAC_N

  UInt uiUnaryIdx = pcCU->getMergeIndex( uiAbsPartIdx );
  UInt uiNumCand = pcCU->getSlice()->getMaxNumMergeCand();
  if ( uiNumCand > 1 )
  {
    for( UInt ui = 0; ui < uiNumCand - 1; ++ui )
    {
      const UInt uiSymbol = ui == uiUnaryIdx ? 0 : 1;
      if ( ui==0 )
      {
        m_pcBinIf->encodeBin( uiSymbol, m_cCUMergeIdxExtSCModel.get( 0, 0, 0 ) );
        DTRACE_CABAC_N
      }
      else
      {
        m_pcBinIf->encodeBinEP( uiSymbol );
        DTRACE_CABAC_N
      }
      if( uiSymbol == 0 )
      {
        break;
      }
    }
  }

#if ENC_DEC_TRACE
  DTRACE_CABAC_T( "\tEnd Code MergeIndex\truiMergeIndex=" )
  DTRACE_CABAC_V( uiUnaryIdx )
  DTRACE_CABAC_N
#endif
}

Void TEncSbac::codeSplitFlag   ( TComDataCU* pcCU, UInt uiAbsPartIdx, UInt uiDepth )
{
  if( uiDepth == g_uiMaxCUDepth - g_uiAddCUDepth )
  {
    return;
  }

#if RWTH_RESI_REFINEMENT
  if (getEncCfg()->getMyCodingArgument()[INHERIT_CU_TREE])
  {
    return;
  }
#endif
  
  UInt uiCtx           = pcCU->getCtxSplitFlag( uiAbsPartIdx, uiDepth );
  UInt uiCurrSplitFlag = ( pcCU->getDepth( uiAbsPartIdx ) > uiDepth ) ? 1 : 0;
  
  assert( uiCtx < 3 );
  m_pcBinIf->encodeBin( uiCurrSplitFlag, m_cCUSplitFlagSCModel.get( 0, 0, uiCtx ) );
  
  DTRACE_CABAC_T( "\tSplitFlag" )
  DTRACE_CABAC_T( "\tuiCtxSplit=")
  DTRACE_CABAC_V( uiCtx )
  DTRACE_CABAC_T( "\tuiAbsPartIdx=")
  DTRACE_CABAC_V( uiAbsPartIdx )
  DTRACE_CABAC_T( "\tuiDepth=")
  DTRACE_CABAC_V( uiDepth )
  DTRACE_CABAC_T( "\n")
}

Void TEncSbac::codeTransformSubdivFlag( UInt uiSymbol, UInt uiCtx )
{
  m_pcBinIf->encodeBin( uiSymbol, m_cCUTransSubdivFlagSCModel.get( 0, 0, uiCtx ) );
  DTRACE_CABAC_T( "\tCode split_transform_flag" )
  DTRACE_CABAC_T( "\tctx=" )
  DTRACE_CABAC_V( uiCtx )
  DTRACE_CABAC_N
}

Void TEncSbac::codeIntraDirLumaAng( TComDataCU* pcCU, UInt absPartIdx, Bool isMultiple)
{
  PartSize mode = pcCU->getPartitionSize( absPartIdx );
  UInt partNum = isMultiple?(mode==SIZE_NxN?4:1):1;

#if ENC_DEC_TRACE
  DTRACE_CABAC_T( "\tStart Code IntraDirLumaAng" );
  DTRACE_CABAC_T( "\tuiAbsPartIdx=" );
  DTRACE_CABAC_V( absPartIdx )
  DTRACE_CABAC_T( "\tnumParts=" );
  DTRACE_CABAC_V( partNum )
  DTRACE_CABAC_N
  UInt uiInitialDir[4];
#endif

  UInt dir[4],j;
  Int preds[4][3] = {{-1, -1, -1},{-1, -1, -1},{-1, -1, -1},{-1, -1, -1}};
  Int predNum[4], predIdx[4] ={ -1,-1,-1,-1};
  UInt partOffset = ( pcCU->getPic()->getNumPartInCU() >> ( pcCU->getDepth(absPartIdx) << 1 ) ) >> 2;
  for (j=0;j<partNum;j++)
  {
    dir[j] = pcCU->getLumaIntraDir( absPartIdx+partOffset*j );
#if ENC_DEC_TRACE
    uiInitialDir[j] = dir[j];
#endif
    predNum[j] = pcCU->getIntraDirLumaPredictor(absPartIdx+partOffset*j, preds[j]);  
    for(UInt i = 0; i < predNum[j]; i++)
    {
      if(dir[j] == preds[j][i])
      {
        predIdx[j] = i;
      }
    }
    m_pcBinIf->encodeBin((predIdx[j] != -1)? 1 : 0, m_cCUIntraPredSCModel.get( 0, 0, 0 ) );
    DTRACE_CABAC_T( "\tprev_intra_luma_pred_flag\n" );
  }  
  for (j=0;j<partNum;j++)
  {
    if(predIdx[j] != -1)
    {
      m_pcBinIf->encodeBinEP( predIdx[j] ? 1 : 0 );
      DTRACE_CABAC_N
      if (predIdx[j])
      {
        m_pcBinIf->encodeBinEP( predIdx[j]-1 );
        DTRACE_CABAC_N

      }
    }
    else
    {
      if (preds[j][0] > preds[j][1])
      { 
        std::swap(preds[j][0], preds[j][1]); 
      }
      if (preds[j][0] > preds[j][2])
      {
        std::swap(preds[j][0], preds[j][2]);
      }
      if (preds[j][1] > preds[j][2])
      {
        std::swap(preds[j][1], preds[j][2]);
      }
      for(Int i = (predNum[j] - 1); i >= 0; i--)
      {
        dir[j] = dir[j] > preds[j][i] ? dir[j] - 1 : dir[j];
      }
      m_pcBinIf->encodeBinsEP( dir[j], 5 );
    }
  }
#if ENC_DEC_TRACE
  DTRACE_CABAC_T( "\tEnd Code IntraDirLumaAng" );
  DTRACE_CABAC_T( "\tintraPredMode=" );
  DTRACE_CABAC_V( uiInitialDir[0] )
  if (partNum == 4)
  {
    DTRACE_CABAC_T( "," )
    DTRACE_CABAC_V( uiInitialDir[1] )
    DTRACE_CABAC_T( "," )
    DTRACE_CABAC_V( uiInitialDir[2] )
    DTRACE_CABAC_T( "," )
    DTRACE_CABAC_V( uiInitialDir[3] )
  }
  DTRACE_CABAC_N
#endif

  return;
}

Void TEncSbac::codeIntraDirChroma( TComDataCU* pcCU, UInt uiAbsPartIdx )
{
  UInt uiIntraDirChroma = pcCU->getChromaIntraDir( uiAbsPartIdx );

  if( uiIntraDirChroma == DM_CHROMA_IDX ) 
  {
    m_pcBinIf->encodeBin( 0, m_cCUChromaPredSCModel.get( 0, 0, 0 ) );
    DTRACE_CABAC_T( "\tCode IntraDirChroma\n" );
  }
#if !REMOVE_LMCHROMA
  else if( uiIntraDirChroma == LM_CHROMA_IDX )
  {
    m_pcBinIf->encodeBin( 1, m_cCUChromaPredSCModel.get( 0, 0, 0 ) );
    DTRACE_CABAC_T( "\tCode IntraDirChroma\n" );
    m_pcBinIf->encodeBin( 0, m_cCUChromaPredSCModel.get( 0, 0, 1 ) );
    DTRACE_CABAC_T( "\tCode IntraDirChroma\n" );
  }
#endif
  else
  { 
    UInt uiAllowedChromaDir[ NUM_CHROMA_MODE ];
    pcCU->getAllowedChromaDir( uiAbsPartIdx, uiAllowedChromaDir );

#if REMOVE_LMCHROMA
    for( Int i = 0; i < NUM_CHROMA_MODE - 1; i++ )
#else
    for( Int i = 0; i < NUM_CHROMA_MODE - 2; i++ )
#endif
    {
      if( uiIntraDirChroma == uiAllowedChromaDir[i] )
      {
        uiIntraDirChroma = i;
        break;
      }
    }
    m_pcBinIf->encodeBin( 1, m_cCUChromaPredSCModel.get( 0, 0, 0 ) );
    DTRACE_CABAC_T( "\tCode IntraDirChroma\n" );

#if !REMOVE_LMCHROMA
    if (pcCU->getSlice()->getSPS()->getUseLMChroma())
    {
      m_pcBinIf->encodeBin( 1, m_cCUChromaPredSCModel.get( 0, 0, 1 ));
      DTRACE_CABAC_T( "\tCode IntraDirChroma\n" );
    }
#endif
    m_pcBinIf->encodeBinsEP( uiIntraDirChroma, 2 );
  }
  return;
}

Void TEncSbac::codeInterDir( TComDataCU* pcCU, UInt uiAbsPartIdx )
{
  const UInt uiInterDir = pcCU->getInterDir( uiAbsPartIdx ) - 1;
  const UInt uiCtx      = pcCU->getCtxInterDir( uiAbsPartIdx );
  ContextModel *pCtx    = m_cCUInterDirSCModel.get( 0 );
#if DISALLOW_BIPRED_IN_8x4_4x8PUS
  if (pcCU->getPartitionSize(uiAbsPartIdx) == SIZE_2Nx2N || pcCU->getHeight(uiAbsPartIdx) != 8 )
  {
#endif
    m_pcBinIf->encodeBin( uiInterDir == 2 ? 1 : 0, *( pCtx + uiCtx ) );
    DTRACE_CABAC_T( "\tCode InterDir\n" );
#if DISALLOW_BIPRED_IN_8x4_4x8PUS
  }
#endif
  if (uiInterDir < 2)
  {
    m_pcBinIf->encodeBin( uiInterDir, *( pCtx + 4 ) );
    DTRACE_CABAC_T( "\tCode InterDir\n" );
  }
  return;
}

Void TEncSbac::codeRefFrmIdx( TComDataCU* pcCU, UInt uiAbsPartIdx, RefPicList eRefList )
{
  {
    Int iRefFrame = pcCU->getCUMvField( eRefList )->getRefIdx( uiAbsPartIdx );
    ContextModel *pCtx = m_cCURefPicSCModel.get( 0 );
    m_pcBinIf->encodeBin( ( iRefFrame == 0 ? 0 : 1 ), *pCtx );
    DTRACE_CABAC_T( "\tCode RefFrmIdx\n" );
    
    if( iRefFrame > 0 )
    {
#if REF_IDX_BYPASS
      UInt uiRefNum = pcCU->getSlice()->getNumRefIdx( eRefList ) - 2;
      pCtx++;
      iRefFrame--;
      for( UInt ui = 0; ui < uiRefNum; ++ui )
      {
        const UInt uiSymbol = ui == iRefFrame ? 0 : 1;
        if( ui == 0 )
        {
          m_pcBinIf->encodeBin( uiSymbol, *pCtx ); 
          DTRACE_CABAC_T( "\tCode RefFrmIdx\n" );      
        }
        else
        {
          m_pcBinIf->encodeBinEP( uiSymbol );
          DTRACE_CABAC_T( "\tCode RefFrmIdx\n" );
        }
        if( uiSymbol == 0 )
        {
          break;
        }
      }
#else
      xWriteUnaryMaxSymbol( iRefFrame - 1, pCtx + 1, 1, pcCU->getSlice()->getNumRefIdx( eRefList )-2 );
#endif
    }
  }
  return;
}

Void TEncSbac::codeMvd( TComDataCU* pcCU, UInt uiAbsPartIdx, RefPicList eRefList )
{
  DTRACE_CABAC_T( "\tStart Code Mvd" );
  DTRACE_CABAC_T( "\tuiAbsPartIdx=" );
  DTRACE_CABAC_V( uiAbsPartIdx )
  DTRACE_CABAC_T( "\teRefList=" );
  DTRACE_CABAC_V( eRefList )
  DTRACE_CABAC_N

  if(pcCU->getSlice()->getMvdL1ZeroFlag() && eRefList == REF_PIC_LIST_1 && pcCU->getInterDir(uiAbsPartIdx)==3)
  {
    DTRACE_CABAC_T( "\tEnd Code Mvd\tiHor=0\tiVer=0\n" )
    return;
  }

  const TComCUMvField* pcCUMvField = pcCU->getCUMvField( eRefList );
  const Int iHor = pcCUMvField->getMvd( uiAbsPartIdx ).getHor();
  const Int iVer = pcCUMvField->getMvd( uiAbsPartIdx ).getVer();
  ContextModel* pCtx = m_cCUMvdSCModel.get( 0 );

  m_pcBinIf->encodeBin( iHor != 0 ? 1 : 0, *pCtx );
  DTRACE_CABAC_N
  m_pcBinIf->encodeBin( iVer != 0 ? 1 : 0, *pCtx );
  DTRACE_CABAC_N

  const Bool bHorAbsGr0 = iHor != 0;
  const Bool bVerAbsGr0 = iVer != 0;
  const UInt uiHorAbs   = 0 > iHor ? -iHor : iHor;
  const UInt uiVerAbs   = 0 > iVer ? -iVer : iVer;
  pCtx++;

  if( bHorAbsGr0 )
  {
    m_pcBinIf->encodeBin( uiHorAbs > 1 ? 1 : 0, *pCtx );
    DTRACE_CABAC_N
  }

  if( bVerAbsGr0 )
  {
    m_pcBinIf->encodeBin( uiVerAbs > 1 ? 1 : 0, *pCtx );
    DTRACE_CABAC_N
  }

  if( bHorAbsGr0 )
  {
    if( uiHorAbs > 1 )
    {
      xWriteEpExGolomb( uiHorAbs-2, 1 );
    }

    m_pcBinIf->encodeBinEP( 0 > iHor ? 1 : 0 );
    DTRACE_CABAC_N
  }

  if( bVerAbsGr0 )
  {
    if( uiVerAbs > 1 )
    {
      xWriteEpExGolomb( uiVerAbs-2, 1 );
    }

    m_pcBinIf->encodeBinEP( 0 > iVer ? 1 : 0 );
    DTRACE_CABAC_N
  }
  
  DTRACE_CABAC_T( "\tEnd Code Mvd" )
  DTRACE_CABAC_T( "\tiHor=" );
  DTRACE_CABAC_V( iHor )
  DTRACE_CABAC_T( "\tiVer=" );
  DTRACE_CABAC_V( iVer )
  DTRACE_CABAC_N

}

Void TEncSbac::codeDeltaQP( TComDataCU* pcCU, UInt uiAbsPartIdx )
{
  Int iDQp  = pcCU->getQP( uiAbsPartIdx ) - pcCU->getRefQP( uiAbsPartIdx );
  
  Int qpBdOffsetY =  pcCU->getSlice()->getSPS()->getQpBDOffsetY();
  iDQp = (iDQp + 78 + qpBdOffsetY + (qpBdOffsetY/2)) % (52 + qpBdOffsetY) - 26 - (qpBdOffsetY/2);

#if CU_DQP_TU_EG
  UInt uiAbsDQp = (UInt)((iDQp > 0)? iDQp  : (-iDQp));
  UInt TUValue = min((Int)uiAbsDQp, CU_DQP_TU_CMAX);
  xWriteUnaryMaxSymbol( TUValue, &m_cCUDeltaQpSCModel.get( 0, 0, 0 ), 1, CU_DQP_TU_CMAX);
  if( uiAbsDQp >= CU_DQP_TU_CMAX )
  {
    xWriteEpExGolomb( uiAbsDQp - CU_DQP_TU_CMAX, CU_DQP_EG_k );
  }

  if ( uiAbsDQp > 0)
  {
    UInt uiSign = (iDQp > 0 ? 0 : 1);
    m_pcBinIf->encodeBinEP(uiSign);
    DTRACE_CABAC_T( "\tDeltaQP\n" )
  }
#else
  if ( iDQp == 0 )
  {
    m_pcBinIf->encodeBin( 0, m_cCUDeltaQpSCModel.get( 0, 0, 0 ) );
    DTRACE_CABAC_T( "\tDeltaQP\n" )
  }
  else
  {
    m_pcBinIf->encodeBin( 1, m_cCUDeltaQpSCModel.get( 0, 0, 0 ) );
    DTRACE_CABAC_T( "\tDeltaQP\n" )
   
    UInt uiSign = (iDQp > 0 ? 0 : 1);

    m_pcBinIf->encodeBinEP(uiSign);
    DTRACE_CABAC_T( "\tDeltaQP\n" )

    assert(iDQp >= -(26+(qpBdOffsetY/2)));
    assert(iDQp <=  (25+(qpBdOffsetY/2)));

    UInt uiMaxAbsDQpMinus1 = 24 + (qpBdOffsetY/2) + (uiSign);
    UInt uiAbsDQpMinus1 = (UInt)((iDQp > 0)? iDQp  : (-iDQp)) - 1;
    xWriteUnaryMaxSymbol( uiAbsDQpMinus1, &m_cCUDeltaQpSCModel.get( 0, 0, 1 ), 1, uiMaxAbsDQpMinus1);
  }
#endif

  return;
}

Void TEncSbac::codeQtCbf( TComDataCU* pcCU, UInt uiAbsPartIdx, TextType eType, UInt uiTrDepth )
{
#if RWTH_RESI_REFINEMENT && 0
  //todo: J:encode cbf
  if (pcCU->getPredictionMode(uiAbsPartIdx)==MODE_IL_RWTH_RESI_REFINEMENT)
    return;
#endif

  UInt uiCbf = pcCU->getCbf     ( uiAbsPartIdx, eType, uiTrDepth );
  UInt uiCtx = pcCU->getCtxQtCbf( uiAbsPartIdx, eType, uiTrDepth );
  m_pcBinIf->encodeBin( uiCbf , m_cCUQtCbfSCModel.get( 0, eType ? TEXT_CHROMA : eType, uiCtx ) );
  
#if ENC_DEC_TRACE
  if ( eType == TEXT_LUMA )
  {
    DTRACE_CABAC_T( "\tcbf_luma" )
  }
  else if ( eType == TEXT_CHROMA_U )
  {
    DTRACE_CABAC_T( "\tcbf_cb" )
  }
  else 
  {
    DTRACE_CABAC_T( "\tcbf_cr" )
  }
  DTRACE_CABAC_T( "\tctx=" )
  DTRACE_CABAC_V( uiCtx )
  DTRACE_CABAC_T( "\tetype=" )
  DTRACE_CABAC_V( eType )
  DTRACE_CABAC_T( "\tuiAbsPartIdx=" )
  DTRACE_CABAC_V( uiAbsPartIdx )
  DTRACE_CABAC_N
#endif
}

void TEncSbac::codeTransformSkipFlags (TComDataCU* pcCU, UInt uiAbsPartIdx, UInt width, UInt height, UInt uiDepth, TextType eTType )
{
  if (pcCU->getCUTransquantBypass(uiAbsPartIdx))
  {
    return;
  }
#if !INTER_TRANSFORMSKIP
  if(!pcCU->isIntra(uiAbsPartIdx))
  {
    return;
  }
#endif
  if(width != 4 || height != 4)
  {
    return;
  }

  UInt useTransformSkip = pcCU->getTransformSkip( uiAbsPartIdx,eTType);
  m_pcBinIf->encodeBin( useTransformSkip, m_cTransformSkipSCModel.get( 0, eTType? TEXT_CHROMA: TEXT_LUMA, 0 ) );
  DTRACE_CABAC_T("\tTransformSkipFlag");
  DTRACE_CABAC_T( "\tsymbol=" )
  DTRACE_CABAC_V( useTransformSkip )
  DTRACE_CABAC_T( "\tAddr=" )
  DTRACE_CABAC_V( pcCU->getAddr() )
  DTRACE_CABAC_T( "\tetype=" )
  DTRACE_CABAC_V( eTType )
  DTRACE_CABAC_T( "\tuiAbsPartIdx=" )
  DTRACE_CABAC_V( uiAbsPartIdx )
  DTRACE_CABAC_T( "\n" )
}

/** Code I_PCM information. 
 * \param pcCU pointer to CU
 * \param uiAbsPartIdx CU index
 * \param numIPCM the number of succesive IPCM blocks with the same size 
 * \param firstIPCMFlag 
 * \returns Void
 */
Void TEncSbac::codeIPCMInfo( TComDataCU* pcCU, UInt uiAbsPartIdx, Int numIPCM, Bool firstIPCMFlag)
{
  UInt uiIPCM = (pcCU->getIPCMFlag(uiAbsPartIdx) == true)? 1 : 0;

  Bool writePCMSampleFlag = pcCU->getIPCMFlag(uiAbsPartIdx);

  if( uiIPCM == 0 || firstIPCMFlag)
  {
    m_pcBinIf->encodeBinTrm (uiIPCM);
    DTRACE_CABAC_N

    if ( firstIPCMFlag )
    {
      m_pcBinIf->encodeNumSubseqIPCM( numIPCM - 1 );
      m_pcBinIf->encodePCMAlignBits();
    }
  }

  if (writePCMSampleFlag)
  {
    UInt uiMinCoeffSize = pcCU->getPic()->getMinCUWidth()*pcCU->getPic()->getMinCUHeight();
    UInt uiLumaOffset   = uiMinCoeffSize*uiAbsPartIdx;
    UInt uiChromaOffset = uiLumaOffset>>2;
    Pel* piPCMSample;
    UInt uiWidth;
    UInt uiHeight;
    UInt uiSampleBits;
    UInt uiX, uiY;

    piPCMSample = pcCU->getPCMSampleY() + uiLumaOffset;
    uiWidth = pcCU->getWidth(uiAbsPartIdx);
    uiHeight = pcCU->getHeight(uiAbsPartIdx);
    uiSampleBits = pcCU->getSlice()->getSPS()->getPCMBitDepthLuma();

    for(uiY = 0; uiY < uiHeight; uiY++)
    {
      for(uiX = 0; uiX < uiWidth; uiX++)
      {
        UInt uiSample = piPCMSample[uiX];

        m_pcBinIf->xWritePCMCode(uiSample, uiSampleBits);
      }
      piPCMSample += uiWidth;
    }

    piPCMSample = pcCU->getPCMSampleCb() + uiChromaOffset;
    uiWidth = pcCU->getWidth(uiAbsPartIdx)/2;
    uiHeight = pcCU->getHeight(uiAbsPartIdx)/2;
    uiSampleBits = pcCU->getSlice()->getSPS()->getPCMBitDepthChroma();

    for(uiY = 0; uiY < uiHeight; uiY++)
    {
      for(uiX = 0; uiX < uiWidth; uiX++)
      {
        UInt uiSample = piPCMSample[uiX];

        m_pcBinIf->xWritePCMCode(uiSample, uiSampleBits);
      }
      piPCMSample += uiWidth;
    }

    piPCMSample = pcCU->getPCMSampleCr() + uiChromaOffset;
    uiWidth = pcCU->getWidth(uiAbsPartIdx)/2;
    uiHeight = pcCU->getHeight(uiAbsPartIdx)/2;
    uiSampleBits = pcCU->getSlice()->getSPS()->getPCMBitDepthChroma();

    for(uiY = 0; uiY < uiHeight; uiY++)
    {
      for(uiX = 0; uiX < uiWidth; uiX++)
      {
        UInt uiSample = piPCMSample[uiX];

        m_pcBinIf->xWritePCMCode(uiSample, uiSampleBits);
      }
      piPCMSample += uiWidth;
    }
    numIPCM--;
    if(numIPCM == 0)
    {
      m_pcBinIf->resetBac();
    }
  }
}

Void TEncSbac::codeQtRootCbf( TComDataCU* pcCU, UInt uiAbsPartIdx )
{
  UInt uiCbf = pcCU->getQtRootCbf( uiAbsPartIdx );

#if RWTH_RESI_REFINEMENT
  if (pcCU->getPredictionMode(uiAbsPartIdx)==MODE_IL_RESI_NO_REFINEMENT)
  {
    uiCbf = 0;
  }
#endif

  UInt uiCtx = 0;
  m_pcBinIf->encodeBin( uiCbf , m_cCUQtRootCbfSCModel.get( 0, 0, uiCtx ) );
  DTRACE_CABAC_T( "\tno_residual_data_flag\tctx=" )
  DTRACE_CABAC_V( uiCtx )
  DTRACE_CABAC_T( "\tuiAbsPartIdx=" )
  DTRACE_CABAC_V( uiAbsPartIdx )
  DTRACE_CABAC_T( "\n" )
}

#if TU_ZERO_CBF_RDO
Void TEncSbac::codeQtCbfZero( TComDataCU* pcCU, UInt uiAbsPartIdx, TextType eType, UInt uiTrDepth )
{
  // this function is only used to estimate the bits when cbf is 0
  // and will never be called when writing the bistream. do not need to write log
  UInt uiCbf = 0;
  UInt uiCtx = pcCU->getCtxQtCbf( uiAbsPartIdx, eType, uiTrDepth );
  m_pcBinIf->encodeBin( uiCbf , m_cCUQtCbfSCModel.get( 0, eType ? TEXT_CHROMA : eType, uiCtx ) );
  DTRACE_CABAC_T( "\tQtCbfZero\n" )
}

Void TEncSbac::codeQtRootCbfZero( TComDataCU* pcCU, UInt uiAbsPartIdx )
{
  // this function is only used to estimate the bits when cbf is 0
  // and will never be called when writing the bistream. do not need to write log
  UInt uiCbf = 0;
  UInt uiCtx = 0;
  m_pcBinIf->encodeBin( uiCbf , m_cCUQtRootCbfSCModel.get( 0, 0, uiCtx ) );
  DTRACE_CABAC_T( "\tQtRootCbfZero\n" )
}
#endif

/** Encode (X,Y) position of the last significant coefficient
 * \param uiPosX X component of last coefficient
 * \param uiPosY Y component of last coefficient
 * \param width  Block width
 * \param height Block height
 * \param eTType plane type / luminance or chrominance
 * \param uiScanIdx scan type (zig-zag, hor, ver)
 * This method encodes the X and Y component within a block of the last significant coefficient.
 */
Void TEncSbac::codeLastSignificantXY( UInt uiPosX, UInt uiPosY, Int width, Int height, TextType eTType, UInt uiScanIdx )
{  
#if ENC_DEC_TRACE
  DTRACE_CABAC_T( "\tStart Coding LastSignificantXY\n" )
  UInt uiInitialPosX = uiPosX;
  UInt uiInitialPosY = uiPosY;
#endif

  // swap
  if( uiScanIdx == SCAN_VER )
  {
    swap( uiPosX, uiPosY );
  }

  UInt uiCtxLast;
  ContextModel *pCtxX = m_cCuCtxLastX.get( 0, eTType );
  ContextModel *pCtxY = m_cCuCtxLastY.get( 0, eTType );
  UInt uiGroupIdxX    = g_uiGroupIdx[ uiPosX ];
  UInt uiGroupIdxY    = g_uiGroupIdx[ uiPosY ];

  Int blkSizeOffsetX, blkSizeOffsetY, shiftX, shiftY;
  blkSizeOffsetX = eTType ? 0: (g_aucConvertToBit[ width ] *3 + ((g_aucConvertToBit[ width ] +1)>>2));
  blkSizeOffsetY = eTType ? 0: (g_aucConvertToBit[ height ]*3 + ((g_aucConvertToBit[ height ]+1)>>2));
  shiftX= eTType ? g_aucConvertToBit[ width  ] :((g_aucConvertToBit[ width  ]+3)>>2);
  shiftY= eTType ? g_aucConvertToBit[ height ] :((g_aucConvertToBit[ height ]+3)>>2);
  // posX
  for( uiCtxLast = 0; uiCtxLast < uiGroupIdxX; uiCtxLast++ )
  {
      m_pcBinIf->encodeBin( 1, *( pCtxX + blkSizeOffsetX + (uiCtxLast >>shiftX) ) );
      DTRACE_CABAC_T( "\tctx:" )
      DTRACE_CABAC_V( blkSizeOffsetX + (uiCtxLast >>shiftX) )
      DTRACE_CABAC_T( "\tlast_significant_coeff_x_prefix\n" )
  }
  if( uiGroupIdxX < g_uiGroupIdx[ width - 1 ])
  {
      m_pcBinIf->encodeBin( 0, *( pCtxX + blkSizeOffsetX + (uiCtxLast >>shiftX) ) );
      DTRACE_CABAC_T( "\tctx:" )
      DTRACE_CABAC_V( blkSizeOffsetX + (uiCtxLast >>shiftX) )
      DTRACE_CABAC_T( "\tlast_significant_coeff_x_prefix\n" )
  }

  // posY
  for( uiCtxLast = 0; uiCtxLast < uiGroupIdxY; uiCtxLast++ )
  {
    m_pcBinIf->encodeBin( 1, *( pCtxY + blkSizeOffsetY + (uiCtxLast >>shiftY) ) );
    DTRACE_CABAC_T( "\tctx:" )
    DTRACE_CABAC_V( blkSizeOffsetY + (uiCtxLast >>shiftY) )
    DTRACE_CABAC_T( "\tlast_significant_coeff_y_prefix\n" )

  }
  if( uiGroupIdxY < g_uiGroupIdx[ height - 1 ])
  {
    m_pcBinIf->encodeBin( 0, *( pCtxY + blkSizeOffsetY + (uiCtxLast >>shiftY) ) );
    DTRACE_CABAC_T( "\tctx:" )
    DTRACE_CABAC_V( blkSizeOffsetY + (uiCtxLast >>shiftY) )
    DTRACE_CABAC_T( "\tlast_significant_coeff_y_prefix\n" )

  }
  if ( uiGroupIdxX > 3 )
  {      
    UInt uiCount = ( uiGroupIdxX - 2 ) >> 1;
    uiPosX       = uiPosX - g_uiMinInGroup[ uiGroupIdxX ];
    for (Int i = uiCount - 1 ; i >= 0; i-- )
    {
      m_pcBinIf->encodeBinEP( ( uiPosX >> i ) & 1 );
      DTRACE_CABAC_T( "\tlast_significant_coeff_x_suffix\n" )
    }
  }
  if ( uiGroupIdxY > 3 )
  {      
    UInt uiCount = ( uiGroupIdxY - 2 ) >> 1;
    uiPosY       = uiPosY - g_uiMinInGroup[ uiGroupIdxY ];
    for ( Int i = uiCount - 1 ; i >= 0; i-- )
    {
      m_pcBinIf->encodeBinEP( ( uiPosY >> i ) & 1 );
      DTRACE_CABAC_T( "\tlast_significant_coeff_y_suffix\n" )
    }
  }
  
  DTRACE_CABAC_T( "\tEnd LastSignificantXY: (" )
  DTRACE_CABAC_V( uiInitialPosX )
  DTRACE_CABAC_T( "," )
  DTRACE_CABAC_V( uiInitialPosY )
  DTRACE_CABAC_T( ")\n" )
}

#if RWTH_CODE_LASTXY_DIFF
Void TEncSbac::xCodeLastDiffPrefixSuffix( UInt uiAbsDelta, UInt uiRemaining, TEncBinIf* pcBinIf, TextType eTType, ContextModel *pCtx, const char *cTraceText )
{
  if (uiRemaining <= 1)
  {
    // Only one or zero possible value ... nothing to code
    return;
  }

  // Encode uiAbsDelta
  UInt uiCtxLast;
  UInt uiGroupIdxDiff = g_uiGroupIdx[ uiAbsDelta ];
  
  // Round uiRemaining up to the next possible block width/height (4,8,16,32,64)
  UInt uiRemainingRound = (uiRemaining > 32) ? 64 :
                          (uiRemaining > 16) ? 32 :
                          (uiRemaining > 8) ? 16 :
                          (uiRemaining > 4) ? 8 : 4;
  Int shift = eTType ? g_aucConvertToBit[ uiRemainingRound ] : ((g_aucConvertToBit[ uiRemainingRound ]+3)>>2);
  Int blkSizeOffset = eTType ? 0: (g_aucConvertToBit[ uiRemainingRound ] *3 + ((g_aucConvertToBit[ uiRemainingRound ] +1)>>2));
  // Prefix
  for( uiCtxLast = 0; uiCtxLast < uiGroupIdxDiff; uiCtxLast++ )
  {
    pcBinIf->encodeBin( 1, *( pCtx + blkSizeOffset + (uiCtxLast >> shift) ) );
    DTRACE_CABAC_T( "\tctx:" )
    DTRACE_CABAC_V( blkSizeOffset + (uiCtxLast >>shift) )
    DTRACE_CABAC_T( "\tlast_significant_coeff_" )
    DTRACE_CABAC_T( cTraceText );
    DTRACE_CABAC_T( "_prefix\n" )
  }
  if( uiGroupIdxDiff < g_uiGroupIdx[uiRemaining-1] )
  {
    pcBinIf->encodeBin( 0, *( pCtx + blkSizeOffset + (uiCtxLast >>shift) ) );
    DTRACE_CABAC_T( "\tctx:" )
    DTRACE_CABAC_V( blkSizeOffset + (uiCtxLast >>shift) )
    DTRACE_CABAC_T( "\tlast_significant_coeff_" )
    DTRACE_CABAC_T( cTraceText )
    DTRACE_CABAC_T( "_prefix\n" )
  }
  // Suffix
  if ( uiGroupIdxDiff > 3 )
  {
    UInt uiCount = ( uiGroupIdxDiff - 2 ) >> 1;
    if ( uiGroupIdxDiff == g_uiGroupIdx[uiRemaining-1] )
    {
      uiCount = g_uiLog2Round[uiRemaining - g_uiMinInGroup[ uiGroupIdxDiff ]];
    }
    uiAbsDelta   = uiAbsDelta - g_uiMinInGroup[ uiGroupIdxDiff ];
    for (Int i = uiCount - 1 ; i >= 0; i-- )
    {
      pcBinIf->encodeBinEP( ( uiAbsDelta >> i ) & 1 );
      DTRACE_CABAC_T( "\tlast_significant_coeff_" )
      DTRACE_CABAC_T( cTraceText )
      DTRACE_CABAC_T( "_suffix\n" )
    }
  }
}

/** Encode difference position of the last significant coefficient
 * \param uiPosX X value of last significant coefficient
 * \param uiPosXRef X value of last significant coefficient in reference layer
 * \param uiPosY Y value of last significant coefficient
 * \param uiPosYRef Y value of last significant coefficient in reference layer
 * \param uiWidth Block width
 * \param uiHeight Block height
 * \param eTType plane type / luminance or chrominance
 * \param uiScanIdx The scan index (SCAN_HOR, SCAN_VER, SCAN_DIAG)
 * \param scan The scan used for coefficient scanning
 * This method encodes the difference to the lower layer significant position.
 */
Void TEncSbac::codeLastSignificantDiff( UInt uiPosX, UInt uiPosXRef, UInt uiPosY, UInt uiPosYRef, UInt uiWidth, UInt uiHeight, TextType eTType, UInt uiScanIdx, const UInt* scan )
{
#if ENC_DEC_TRACE
  DTRACE_CABAC_T( "\tStart Coding LastSignificantXYDiff\n" )
#endif

  const UInt uiLog2BlockSize = g_aucConvertToBit[ uiWidth ] + 2;
                
  if ( uiScanIdx == SCAN_HOR )
  {
    swap( uiPosX, uiPosY );
    swap( uiPosXRef, uiPosYRef );
    swap( uiWidth, uiHeight );
  }
  Int iPosXDelta = uiPosX - uiPosXRef;
  Int iPosYDelta = uiPosY - uiPosYRef;

  ContextModel *pCtxSignalling = m_cCuCtxLastDiffSignalling.get( 0, eTType );

  Int uiAbsDeltaX;
  Int uiRemainingWidth;
  Int uiAbsDeltaY;
  Int uiRemainingHeight;
  Bool bCodeXFirst = true;

  // Code (iPosXDelta>=0 && iPosYDelta>=0)
  // In the following cases the flag can be inferred:
  // - if reference pos is (0,0). In this case both deltas are positive
  // - if SCAN_DIAG is used and refPos is one of the last two coefficients in scanning order
  // - if SCAN_DIAG is used and refPos is the first coefficient in the last 4x4 block
  // - if SCAN_VER (or swapped SCAN_HOR) is used and refPosY is 0 and refPosX is a multiple of 4
  // - if SCAN_VER (or swapped SCAN_HOR) is used and refPosX is uiWidth-4 and refPosY is a multiple of 4
  // - if SCAN_VER (or swapped SCAN_HOR) is used and refPos is in the last 4x4 block and refPosX is uiWidth-1
  // - if SCAN_VER (or swapped SCAN_HOR) is used and refPos is in the last 4x4 block and refPosY is uiHeight-4
  if ( 
       ( uiPosXRef == 0 && uiPosYRef == 0 ) ||
       ( uiScanIdx == SCAN_DIAG && uiPosXRef == uiWidth-1 && uiPosYRef >= uiHeight-2 ) ||
       ( uiScanIdx == SCAN_DIAG && uiPosXRef == uiWidth-4 && uiPosYRef == uiHeight-4 ) || 
       ( (uiScanIdx == SCAN_HOR || uiScanIdx == SCAN_VER) && uiPosYRef == 0 && uiPosXRef % 4 == 0 ) ||
       ( (uiScanIdx == SCAN_HOR || uiScanIdx == SCAN_VER) && uiPosXRef == uiWidth-4 && uiPosYRef % 4 == 0 ) ||
       ( (uiScanIdx == SCAN_HOR || uiScanIdx == SCAN_VER) && uiPosXRef == uiWidth-1 && uiPosYRef >= uiHeight-4 ) ||
       ( (uiScanIdx == SCAN_HOR || uiScanIdx == SCAN_VER) && uiPosYRef == uiHeight-4 && uiPosXRef >= uiWidth-4 )
    )
  {
    // last_significant_diff_x>=0_&&_y>=0 has to be true
    assert( iPosXDelta>=0 && iPosYDelta>=0 );
  }
  else
  {
    UInt ctx = ( uiPosXRef >= (uiWidth/4) && uiPosXRef < (uiWidth/4*3) ) ? 0 : 1;
    m_pcBinIf->encodeBin( (iPosXDelta>=0 && iPosYDelta>=0), *( pCtxSignalling + ctx ) );
    DTRACE_CABAC_T( "\tctx:" )
    DTRACE_CABAC_V( ctx )
    DTRACE_CABAC_T( "\tlast_significant_diff_x>=0_&&_y>=0\n" )
  }

  if (iPosXDelta>=0 && iPosYDelta>=0)
  {
    // Code iPosXDelta and iPosYDelta normally
    uiAbsDeltaX       = uiPosX - uiPosXRef;
    uiRemainingWidth  = uiWidth  - uiPosXRef;
    uiAbsDeltaY       = uiPosY - uiPosYRef;
    uiRemainingHeight = uiHeight - uiPosYRef;
  }
  else // iPosXDelta<0 || iPosYDelta<0
  {
    assert (iPosXDelta>=0 || iPosYDelta>=0);
    assert (iPosXDelta<0  || iPosYDelta<0 );
    // encode if iPosXDelta<0 or iPosYDelta<0;
    // In the following cases iPosXDelta<0 can be inferred:
    // - if uiPosXRef==0 or uiPosYRef==0.
    // - if uiPosXRef==uiWidth-1
    // - if SCAN_DIAG is used and uiPosYRef >= uiHeight-2
    // - if SCAN_DIAG is used and uiPosYRef+8 >= uiHeight and uiPosXRef is a multiple of 4
    // - if SCAN_VER (or swapped SCAN_HOR) is used and uiPosXRef is a multiple of 4
    // - if SCAN_VER (or swapped SCAN_HOR) is used and uiPosXRef >= uiWidth-4 and uiPosYRef is a multiple of 4
    
    if ( ( uiPosXRef==0 || uiPosYRef==0 ) ||
         ( uiPosXRef==uiWidth-1 ) || 
         ( uiScanIdx == SCAN_DIAG && uiPosYRef >= uiHeight-2) || 
         ( uiScanIdx == SCAN_DIAG && uiPosYRef+8 >= uiHeight && (uiPosXRef % 4) == 0 ) ||
         ( (uiScanIdx == SCAN_VER || uiScanIdx == SCAN_HOR ) && (uiPosXRef % 4) == 0 ) ||
         ( (uiScanIdx == SCAN_VER || uiScanIdx == SCAN_HOR ) && uiPosYRef >= uiWidth-4 && (uiPosXRef % 4) == 0 )
       )
    {
      // iPosXDelta<0 can be inferred
      if ( uiPosYRef==0 || uiPosXRef==uiWidth-1 )
      {
        // last_significant_diff_x<0 has to be true
        assert ( iPosYDelta>0 );
        assert ( iPosXDelta<0 );        
      }
      else
      {
        // last_significant_diff_x<0 has to be false
        assert ( iPosYDelta<0 );
        assert ( iPosXDelta>0 );
      }
    }
    else
    {
      UInt ctx = 2;
      m_pcBinIf->encodeBin( (iPosXDelta<0), *( pCtxSignalling + ctx ) );
      DTRACE_CABAC_T( "\tctx:" )
      DTRACE_CABAC_V( ctx )
      DTRACE_CABAC_T( "\tlast_significant_diff_x<0\n" )
    }

    if (uiScanIdx == SCAN_DIAG)
    {
      // Scan again
      Int iScanPos = -1;
      Int iScanPosY, iScanPosX, posRef;
      UInt uiXLimit = 0;
      UInt uiYLimit = 0;
      if (iPosXDelta<0)
      {
        // iPosXDelta<0 and iPosYDelta>0
        // coded x value limits the possible values of y (y value depends on x)
        do
        {
          posRef = scan[ ++iScanPos ];
          iScanPosY = posRef >> uiLog2BlockSize;
          iScanPosX = posRef - ( iScanPosY << uiLog2BlockSize );

          if (iScanPosY == uiHeight-1 && iScanPosX >= uiXLimit)
          {
            uiXLimit = iScanPosX + 1;
          }
          if (iScanPosX == uiPosX && iScanPosY >= uiYLimit)
          {
            uiYLimit = iScanPosY + 1;
          }
        }
        while ( ! (iScanPosX==uiPosXRef && iScanPosY==uiPosYRef) );
        uiAbsDeltaX       = abs(iPosXDelta) - 1;
        uiRemainingWidth  = uiPosXRef - uiXLimit;
        uiAbsDeltaY       = uiPosY - uiYLimit;
        uiRemainingHeight = uiHeight - uiYLimit;
      }
      else
      {
        // iPosYDelta<0 and iPosXDelta>0
        // coded y value limits the possible values of x (x value depends on y)
        do
        {
          posRef = scan[ ++iScanPos ];
          iScanPosY = posRef >> uiLog2BlockSize;
          iScanPosX = posRef - ( iScanPosY << uiLog2BlockSize );

          if (iScanPosY == uiPosY && iScanPosX >= uiXLimit)
          {
            uiXLimit = iScanPosX + 1;
          }
          if (iScanPosX == uiWidth-1 && iScanPosY >= uiYLimit)
          {
            uiYLimit = iScanPosY + 1;
          }
        }
        while ( ! (iScanPosX==uiPosXRef && iScanPosY==uiPosYRef) );
        uiAbsDeltaX       = uiPosX - uiXLimit;
        uiRemainingWidth  = uiWidth - uiXLimit;
        uiAbsDeltaY       = abs(iPosYDelta) - 1;
        uiRemainingHeight = uiPosYRef - uiYLimit;
        bCodeXFirst = false;
      }
    }
    else
    {
      // SCAN_VER or swapped SCAN_HOR
      if ( iPosXDelta>0 )
      {
        uiAbsDeltaX       = uiPosX - uiPosXRef;
        uiRemainingWidth  = uiWidth - uiPosXRef;
        uiAbsDeltaY       = abs(iPosYDelta) - 1;
        uiRemainingHeight = uiPosYRef;
      }
      else
      {
        uiAbsDeltaX       = abs(iPosXDelta) - 1;
        uiRemainingWidth  = uiPosXRef;
        UInt uiNum4x4BlocksStartedY = ((uiPosYRef / 4) + 1);
        uiAbsDeltaY       = uiPosY - uiNum4x4BlocksStartedY * 4;
        uiRemainingHeight = uiHeight - uiNum4x4BlocksStartedY * 4;
      }
    }
  }

  // Encode the set values
  assert ( uiAbsDeltaX >= 0 );
  assert ( uiRemainingWidth >= 0 && uiRemainingWidth <= uiWidth );
  assert ( uiAbsDeltaY >= 0 );
  assert ( uiRemainingHeight >= 0 && uiRemainingHeight <= uiHeight );

  ContextModel *pCtxX = m_cCuCtxLastDiffX.get( 0, eTType );
  ContextModel *pCtxY = m_cCuCtxLastDiffY.get( 0, eTType );

  if (bCodeXFirst)
  {
    xCodeLastDiffPrefixSuffix( uiAbsDeltaX, uiRemainingWidth, m_pcBinIf, eTType, pCtxX, "deltaX" );
    xCodeLastDiffPrefixSuffix( uiAbsDeltaY, uiRemainingHeight, m_pcBinIf, eTType, pCtxY, "deltaY" );
  }
  else
  {
    xCodeLastDiffPrefixSuffix( uiAbsDeltaY, uiRemainingHeight, m_pcBinIf, eTType, pCtxY, "deltaY" );
    xCodeLastDiffPrefixSuffix( uiAbsDeltaX, uiRemainingWidth, m_pcBinIf, eTType, pCtxX, "deltaX" );
  }

#if ENC_DEC_TRACE
  if ( uiScanIdx == SCAN_HOR )
  {
    swap( uiPosX, uiPosY );
    swap( uiPosXRef, uiPosYRef );
  }
  DTRACE_CABAC_T( "\tLastSignificantXYDiff: (" )
  DTRACE_CABAC_V( uiPosX )
  DTRACE_CABAC_T( "," )
  DTRACE_CABAC_V( uiPosY )
  DTRACE_CABAC_T( ")\t Ref(" )
  DTRACE_CABAC_V( uiPosXRef )
  DTRACE_CABAC_T( "," )
  DTRACE_CABAC_V( uiPosYRef )
  DTRACE_CABAC_T( ")\n" )
#endif
}
#endif

Void TEncSbac::codeCoeffNxN( TComDataCU* pcCU, TCoeff* pcCoef, UInt uiAbsPartIdx, UInt uiWidth, UInt uiHeight, UInt uiDepth, TextType eTType )
{
  if( uiWidth > m_pcSlice->getSPS()->getMaxTrSize() )
  {
    uiWidth  = m_pcSlice->getSPS()->getMaxTrSize();
    uiHeight = m_pcSlice->getSPS()->getMaxTrSize();
  }

  DTRACE_CABAC_T( "\tStart Code CoeffNxN" )
  DTRACE_CABAC_T( "\teType=" )
  DTRACE_CABAC_V( eTType )
  DTRACE_CABAC_T( "\twidth=" )
  DTRACE_CABAC_V( uiWidth )
  DTRACE_CABAC_T( "\theight=" )
  DTRACE_CABAC_V( uiHeight )
  DTRACE_CABAC_T( "\tdepth=" )
  DTRACE_CABAC_V( uiDepth )
  DTRACE_CABAC_T( "\tabspartidx=" )
  DTRACE_CABAC_V( uiAbsPartIdx )
  DTRACE_CABAC_T( "\ttoCU-X=" )
  DTRACE_CABAC_V( pcCU->getCUPelX() )
  DTRACE_CABAC_T( "\ttoCU-Y=" )
  DTRACE_CABAC_V( pcCU->getCUPelY() )
  DTRACE_CABAC_T( "\tCU-addr=" )
  DTRACE_CABAC_V(  pcCU->getAddr() )
  DTRACE_CABAC_T( "\tinCU-X=" )
  DTRACE_CABAC_V( g_auiRasterToPelX[ g_auiZscanToRaster[uiAbsPartIdx] ] )
  DTRACE_CABAC_T( "\tinCU-Y=" )
  DTRACE_CABAC_V( g_auiRasterToPelY[ g_auiZscanToRaster[uiAbsPartIdx] ] )
  DTRACE_CABAC_T( "\tpredmode=" )
  DTRACE_CABAC_V(  pcCU->getPredictionMode( uiAbsPartIdx ) )
  DTRACE_CABAC_N

#if 0
  printf("encoding pcCoeff\n");
  for (int i = 0; i<uiHeight; i++){
    for (int j = 0; j<uiWidth; j++){
      printf("%d\t", pcCoef[j + i*uiWidth]);
    }
    printf("\n");
  }

#endif
  
  UInt uiNumSig = 0;
  
  // compute number of significant coefficients
  uiNumSig = TEncEntropy::countNonZeroCoeffs(pcCoef, uiWidth * uiHeight);
  
  if ( uiNumSig == 0 )
  {
    // No significant coefficients
    DTRACE_CABAC_T( "\tEnd Code CoeffNxN" )
#if ENC_DEC_TRACE_COEFF
    DTRACE_CABAC_T( "\tCoeff: " )
    for (UInt i = 0; i < uiWidth * uiHeight; i++)
    {
      DTRACE_CABAC_V( pcCoef[i] )
      DTRACE_CABAC_T( " " )
    }
#endif
    DTRACE_CABAC_N
    return;
  }

#if PPS_TS_FLAG
  if(pcCU->getSlice()->getPPS()->getUseTransformSkip())
#else
  if(pcCU->getSlice()->getSPS()->getUseTransformSkip())
#endif
  {
    codeTransformSkipFlags( pcCU,uiAbsPartIdx, uiWidth, uiHeight, uiDepth, eTType );
  }
  eTType = eTType == TEXT_LUMA ? TEXT_LUMA : ( eTType == TEXT_NONE ? TEXT_NONE : TEXT_CHROMA );
  
  //----- encode significance map -----
  const UInt   uiLog2BlockSize   = g_aucConvertToBit[ uiWidth ] + 2;
  UInt uiScanIdx = pcCU->getCoefScanIdx(uiAbsPartIdx, uiWidth, eTType==TEXT_LUMA, pcCU->isIntra(uiAbsPartIdx));
  if (uiScanIdx == SCAN_ZIGZAG)
  {
    // Map zigzag to diagonal scan
    uiScanIdx = SCAN_DIAG;
  }
  Int blockType = uiLog2BlockSize;
  if (uiWidth != uiHeight)
  {
    uiScanIdx = SCAN_DIAG;
    blockType = 4;
  }
  
#if REMOVE_NSQT
  const UInt *scan = g_auiSigLastScan[ uiScanIdx ][ uiLog2BlockSize - 1 ];
#else
  const UInt * scan;
  if (uiWidth == uiHeight)
  {
    scan = g_auiSigLastScan[ uiScanIdx ][ uiLog2BlockSize - 1 ];
  }
  else
  {
    scan = g_sigScanNSQT[ uiLog2BlockSize - 2 ];
  }
#endif
  
  Bool beValid;
  if (pcCU->getCUTransquantBypass(uiAbsPartIdx))
  {
    beValid = false;
  }
  else 
  {
    beValid = pcCU->getSlice()->getPPS()->getSignHideFlag() > 0;
  }

  // Find position of last coefficient
  Int scanPosLast = -1;
  Int posLast;

  const UInt * scanCG;
#if !REMOVE_NSQT
  if (uiWidth == uiHeight)
#endif
  {
    scanCG = g_auiSigLastScan[ uiScanIdx ][ uiLog2BlockSize > 3 ? uiLog2BlockSize-2-1 : 0 ];
    if( uiLog2BlockSize == 3 )
    {
      scanCG = g_sigLastScan8x8[ uiScanIdx ];
    }
    else if( uiLog2BlockSize == 5 )
    {
      scanCG = g_sigLastScanCG32x32;
    }
  }
#if !REMOVE_NSQT
  else
  {
    scanCG = g_sigCGScanNSQT[ uiLog2BlockSize - 2 ];
  }
#endif
  UInt uiSigCoeffGroupFlag[ MLS_GRP_NUM ];
  static const UInt uiShift = MLS_CG_SIZE >> 1;
  const UInt uiNumBlkSide = uiWidth >> uiShift;

    ::memset( uiSigCoeffGroupFlag, 0, sizeof(UInt) * MLS_GRP_NUM );

    do
    {
      posLast = scan[ ++scanPosLast ];

      // get L1 sig map
      UInt uiPosY    = posLast >> uiLog2BlockSize;
      UInt uiPosX    = posLast - ( uiPosY << uiLog2BlockSize );
      UInt uiBlkIdx  = uiNumBlkSide * (uiPosY >> uiShift) + (uiPosX >> uiShift);
#if !REMOVAL_8x2_2x8_CG
      if( uiWidth == 8 && uiHeight == 8 && (uiScanIdx == SCAN_HOR || uiScanIdx == SCAN_VER) )
      {
        if( uiScanIdx == SCAN_HOR )
        {
          uiBlkIdx = uiPosY >> 1;
        }
        else if( uiScanIdx == SCAN_VER )
        {
          uiBlkIdx = uiPosX >> 1;
        }
      }
#endif
      if( pcCoef[ posLast ] )
      {
        uiSigCoeffGroupFlag[ uiBlkIdx ] = 1;
      }

      uiNumSig -= ( pcCoef[ posLast ] != 0 );
    }
    while ( uiNumSig > 0 );

  // Code position of last coefficient
  Int posLastY = posLast >> uiLog2BlockSize;
  Int posLastX = posLast - ( posLastY << uiLog2BlockSize );
  codeLastSignificantXY(posLastX, posLastY, uiWidth, uiHeight, eTType, uiScanIdx);
  UInt uiBlkPosLast = posLastX + (posLastY<<uiLog2BlockSize);

#if ENC_DEC_TRACE
  switch (uiScanIdx)
  {
    case 0: DTRACE_CABAC_T("\tscanIdx=SCAN_ZIGZAG\n") break;
    case 1: DTRACE_CABAC_T("\tscanIdx=SCAN_HOR\n") break;
    case 2: DTRACE_CABAC_T("\tscanIdx=SCAN_VER\n") break;
    case 3: DTRACE_CABAC_T("\tscanIdx=SCAN_DIAG\n") break;
  }
#endif  
  
  //===== code significance flag =====
  ContextModel * const baseCoeffGroupCtx = m_cCUSigCoeffGroupSCModel.get( 0, eTType );
  ContextModel * const baseCtx = (eTType==TEXT_LUMA) ? m_cCUSigSCModel.get( 0, 0 ) : m_cCUSigSCModel.get( 0, 0 ) + NUM_SIG_FLAG_CTX_LUMA;
  
  const Int  iLastScanSet      = scanPosLast >> LOG2_SCAN_SET_SIZE;
#if REMOVE_NUM_GREATER1
  UInt c1 = 1;
#else
  UInt uiNumOne                = 0;
#endif
  UInt uiGoRiceParam           = 0;
  Int  iScanPosSig             = scanPosLast;

  for( Int iSubSet = iLastScanSet; iSubSet >= 0; iSubSet-- )
  {
    Int numNonZero = 0;
    Int  iSubPos     = iSubSet << LOG2_SCAN_SET_SIZE;
    uiGoRiceParam    = 0;
    Int absCoeff[16];
    UInt coeffSigns = 0;
    Int pos[SCAN_SET_SIZE];
    Int lastNZPosInCG = -1, firstNZPosInCG = SCAN_SET_SIZE;

    if( iScanPosSig == scanPosLast )
    {
      absCoeff[ 0 ] = abs( pcCoef[ posLast ] );
      coeffSigns    = ( pcCoef[ posLast ] < 0 );
      pos[ numNonZero ] = uiBlkPosLast;
      numNonZero    = 1;
      lastNZPosInCG  = iScanPosSig;
      firstNZPosInCG = iScanPosSig;
      iScanPosSig--;
    }

      // encode significant_coeffgroup_flag
      Int iCGBlkPos = scanCG[ iSubSet ];
      Int iCGPosY   = iCGBlkPos / uiNumBlkSide;
      Int iCGPosX   = iCGBlkPos - (iCGPosY * uiNumBlkSide);
#if !REMOVAL_8x2_2x8_CG
      if( uiWidth == 8 && uiHeight == 8 && (uiScanIdx == SCAN_HOR || uiScanIdx == SCAN_VER) )
      {
        iCGPosY = (uiScanIdx == SCAN_HOR ? iCGBlkPos : 0);
        iCGPosX = (uiScanIdx == SCAN_VER ? iCGBlkPos : 0);
      }
#endif

      DTRACE_CABAC_T( "\tCode SubSet " )
      DTRACE_CABAC_V( iSubSet )
      DTRACE_CABAC_T( " - iCGBlkPos:" )
      DTRACE_CABAC_V( iCGBlkPos )  
      DTRACE_CABAC_T( " (" )
      DTRACE_CABAC_V( iCGPosX )
      DTRACE_CABAC_T( "," )
      DTRACE_CABAC_V( iCGPosY )
      DTRACE_CABAC_T( ")\n" )

      if( iSubSet == iLastScanSet || iSubSet == 0)
      {
        uiSigCoeffGroupFlag[ iCGBlkPos ] = 1;
      }
      else
      {
          UInt uiSigCoeffGroup   = (uiSigCoeffGroupFlag[ iCGBlkPos ] != 0);
          UInt uiCtxSig  = TComTrQuant::getSigCoeffGroupCtxInc( uiSigCoeffGroupFlag, iCGPosX, iCGPosY, uiScanIdx, uiWidth, uiHeight );
          m_pcBinIf->encodeBin( uiSigCoeffGroup, baseCoeffGroupCtx[ uiCtxSig ] );
          DTRACE_CABAC_T( "\tsignificant_coeff_group_flag\tctx:" )
          DTRACE_CABAC_V( uiCtxSig )
          DTRACE_CABAC_N

      }
      
      // encode significant_coeff_flag
      if( uiSigCoeffGroupFlag[ iCGBlkPos ] )
      {
        Int patternSigCtx = TComTrQuant::calcPatternSigCtx( uiSigCoeffGroupFlag, iCGPosX, iCGPosY, uiWidth, uiHeight );
        UInt uiBlkPos, uiPosY, uiPosX, uiSig, uiCtxSig;
        for( ; iScanPosSig >= iSubPos; iScanPosSig-- )
        {
          uiBlkPos  = scan[ iScanPosSig ]; 
          uiPosY    = uiBlkPos >> uiLog2BlockSize;
          uiPosX    = uiBlkPos - ( uiPosY << uiLog2BlockSize );
          uiSig     = (pcCoef[ uiBlkPos ] != 0);
          if( iScanPosSig > iSubPos || iSubSet == 0 || numNonZero )
          {
#if REMOVAL_8x2_2x8_CG
            uiCtxSig  = TComTrQuant::getSigCtxInc( patternSigCtx, uiScanIdx, uiPosX, uiPosY, blockType, uiWidth, uiHeight, eTType );
#else
            uiCtxSig  = TComTrQuant::getSigCtxInc( patternSigCtx, uiPosX, uiPosY, blockType, uiWidth, uiHeight, eTType );
#endif
            m_pcBinIf->encodeBin( uiSig, baseCtx[ uiCtxSig ] );
            DTRACE_CABAC_T( "\tsignificant_coeff_flag\tctx:" )
            DTRACE_CABAC_V( uiCtxSig )
            DTRACE_CABAC_T( "\tscanPos:" )
            DTRACE_CABAC_V( iScanPosSig )
            DTRACE_CABAC_T( "\tblockPos:" )
            DTRACE_CABAC_V( uiBlkPos )
            DTRACE_CABAC_T( " (" )
            DTRACE_CABAC_V( uiPosX )
            DTRACE_CABAC_T( "," )
            DTRACE_CABAC_V( uiPosY )
            DTRACE_CABAC_T( ")\n" )
          }
          if( uiSig )
          {
            absCoeff[ numNonZero ] = abs( pcCoef[ uiBlkPos ] );
            coeffSigns = 2 * coeffSigns + ( pcCoef[ uiBlkPos ] < 0 );
            pos[ numNonZero ] = uiBlkPos;
            numNonZero++;
            if( lastNZPosInCG == -1 )
            {
              lastNZPosInCG = iScanPosSig;
            }
            firstNZPosInCG = iScanPosSig;
          }
        }
      }
      else
      {
        iScanPosSig = iSubPos - 1;
      }

    if( numNonZero > 0 )
    {
#if ENC_DEC_TRACE
      UInt uiPosY, uiPosX;
#endif
      Bool signHidden = ( lastNZPosInCG - firstNZPosInCG >= SBH_THRESHOLD );
#if !REMOVE_NUM_GREATER1
      UInt c1 = 1;
#endif
      UInt uiCtxSet = (iSubSet > 0 && eTType==TEXT_LUMA) ? 2 : 0;
      
#if REMOVE_NUM_GREATER1
      if( c1 == 0 )
#else
      if( uiNumOne > 0 )
#endif
      {
        uiCtxSet++;
      }
#if REMOVE_NUM_GREATER1
      c1 = 1;
#else
      uiNumOne       >>= 1;
#endif
      ContextModel *baseCtxMod = ( eTType==TEXT_LUMA ) ? m_cCUOneSCModel.get( 0, 0 ) + 4 * uiCtxSet : m_cCUOneSCModel.get( 0, 0 ) + NUM_ONE_FLAG_CTX_LUMA + 4 * uiCtxSet;
      
      Int numC1Flag = min(numNonZero, C1FLAG_NUMBER);
      Int firstC2FlagIdx = -1;
      for( Int idx = 0; idx < numC1Flag; idx++ )
      {
        UInt uiSymbol = absCoeff[ idx ] > 1;
        m_pcBinIf->encodeBin( uiSymbol, baseCtxMod[c1] );
#if ENC_DEC_TRACE
        DTRACE_CABAC_T( "\tcoeff_abs_level_greater1_flag\tctx:" )
        DTRACE_CABAC_V( c1 )
        DTRACE_CABAC_T( "\tblockPos:" )
        DTRACE_CABAC_V( pos[idx] )
        uiPosY = pos[idx] >> uiLog2BlockSize;
        uiPosX = pos[idx] - ( uiPosY << uiLog2BlockSize );
        DTRACE_CABAC_T( " (" )
        DTRACE_CABAC_V( uiPosX )
        DTRACE_CABAC_T( "," )
        DTRACE_CABAC_V( uiPosY )
        DTRACE_CABAC_T( ")\n" )
#endif
        if( uiSymbol )
        {
          c1 = 0;

          if (firstC2FlagIdx == -1)
          {
            firstC2FlagIdx = idx;
          }
        }
        else if( (c1 < 3) && (c1 > 0) )
        {
          c1++;
        }
      }
      
      if (c1 == 0)
      {

        baseCtxMod = ( eTType==TEXT_LUMA ) ? m_cCUAbsSCModel.get( 0, 0 ) + uiCtxSet : m_cCUAbsSCModel.get( 0, 0 ) + NUM_ABS_FLAG_CTX_LUMA + uiCtxSet;
        if ( firstC2FlagIdx != -1)
        {
          UInt symbol = absCoeff[ firstC2FlagIdx ] > 2;
          m_pcBinIf->encodeBin( symbol, baseCtxMod[0] );
#if ENC_DEC_TRACE          
          DTRACE_CABAC_T( "\tcoeff_abs_level_greater2_flag\tctx:" )
          DTRACE_CABAC_V( 0 )
          DTRACE_CABAC_T( "\tblockPos:" )
          DTRACE_CABAC_V( pos[firstC2FlagIdx] )
          uiPosY = pos[firstC2FlagIdx] >> uiLog2BlockSize;
          uiPosX = pos[firstC2FlagIdx] - ( uiPosY << uiLog2BlockSize );
          DTRACE_CABAC_T( " (" )
          DTRACE_CABAC_V( uiPosX )
          DTRACE_CABAC_T( "," )
          DTRACE_CABAC_V( uiPosY )
          DTRACE_CABAC_T( ")\n" )
#endif
        }
      }
      
      DTRACE_CABAC_T( "\tStart Coding signs\n" )
      if( beValid && signHidden )
      {
        m_pcBinIf->encodeBinsEP( (coeffSigns >> 1), numNonZero-1 );
      }
      else
      {
        m_pcBinIf->encodeBinsEP( coeffSigns, numNonZero );
      }
      DTRACE_CABAC_T( "\tEnd Coding signs - val:" )
      DTRACE_CABAC_V( (beValid && signHidden) ? coeffSigns>>1 : coeffSigns )
      DTRACE_CABAC_N
      
      Int iFirstCoeff2 = 1;    
      if (c1 == 0 || numNonZero > C1FLAG_NUMBER)
      {
        for ( Int idx = 0; idx < numNonZero; idx++ )
        {
          UInt baseLevel  = (idx < C1FLAG_NUMBER)? (2 + iFirstCoeff2 ) : 1;

          if( absCoeff[ idx ] >= baseLevel)
          {
            DTRACE_CABAC_T( "\tStart coding coeff_abs_level_remaining\n" )
            xWriteCoefRemainExGolomb( absCoeff[ idx ] - baseLevel, uiGoRiceParam );
#if ENC_DEC_TRACE
            DTRACE_CABAC_T( "\tEnd coding coeff_abs_level_remaining\tcoeff_before:" )
            DTRACE_CABAC_V( baseLevel )
            DTRACE_CABAC_T( "\tcoeff_after:" )
            DTRACE_CABAC_V( absCoeff[ idx ] )
            DTRACE_CABAC_T( "\tblockPos:" )
            DTRACE_CABAC_V( pos[idx] )
            uiPosY = pos[idx] >> uiLog2BlockSize;
            uiPosX = pos[idx] - ( uiPosY << uiLog2BlockSize );
            DTRACE_CABAC_T( " (" )
            DTRACE_CABAC_V( uiPosX )
            DTRACE_CABAC_T( "," )
            DTRACE_CABAC_V( uiPosY )
            DTRACE_CABAC_T( ")\n" )
#endif

            
            if(absCoeff[idx] > 3*(1<<uiGoRiceParam))
            {
               uiGoRiceParam = min<UInt>(uiGoRiceParam+ 1, 4);
            }
          }
          if(absCoeff[ idx ] >= 2)  
          {
            iFirstCoeff2 = 0;
#if !REMOVE_NUM_GREATER1
            uiNumOne++;
#endif
          }
        }        
      }
      #if ENC_DEC_TRACE
      UInt absSum = 0;
      if (beValid && signHidden)
      {
        for( Int idx = 0; idx < numNonZero; idx++ )
        {
          Int blkPos = pos[ idx ];  
          absSum += absCoeff[ idx ];
          if ( idx == numNonZero-1 && signHidden && beValid )
          {
            DTRACE_CABAC_T( "\tInfer hidden sign\tval:" )
            DTRACE_CABAC_V( absSum&0x1 )
            DTRACE_CABAC_T( "\tblockPos:" )
            DTRACE_CABAC_V( blkPos )
            uiPosY = blkPos >> uiLog2BlockSize;
            uiPosX = blkPos - ( uiPosY << uiLog2BlockSize );
            DTRACE_CABAC_T( " (" )
            DTRACE_CABAC_V( uiPosX )
            DTRACE_CABAC_T( "," )
            DTRACE_CABAC_V( uiPosY )
            DTRACE_CABAC_T( ")\n" )
          }
        }
      }
#endif
    }
#if !REMOVE_NUM_GREATER1
    else
    {
      uiNumOne >>= 1;
    }
#endif
  }

  DTRACE_CABAC_T( "\tEnd Code CoeffNxN" )
#if ENC_DEC_TRACE_COEFF
  DTRACE_CABAC_T( "\tCoeff: " )
  for (UInt i = 0; i < uiWidth * uiHeight; i++)
  {
    DTRACE_CABAC_V( pcCoef[i] )
    DTRACE_CABAC_T( " " )
  }
#endif
  DTRACE_CABAC_N

  return;
}

#if RWTH_RESI_REFINEMENT
Void TEncSbac::codeCoeffRefinement(TComDataCU* pcCU, TCoeff* pcCoef, TCoeff* pCoefRefLayer, UInt uiAbsPartIdx, UInt uiWidth, UInt uiHeight, UInt uiDepth, TextType eTType, TComDataCU* pcColCU )
{
  DTRACE_CABAC_T( "\tStart Code Refinement" )
  DTRACE_CABAC_T( "\teType=" )
  DTRACE_CABAC_V( eTType )
  DTRACE_CABAC_T( "\twidth=" )
  DTRACE_CABAC_V( uiWidth )
  DTRACE_CABAC_T( "\theight=" )
  DTRACE_CABAC_V( uiHeight )
  DTRACE_CABAC_T( "\tdepth=" )
  DTRACE_CABAC_V( uiDepth )
  DTRACE_CABAC_T( "\tabspartidx=" )
  DTRACE_CABAC_V( uiAbsPartIdx )
  DTRACE_CABAC_T( "\ttoCU-X=" )
  DTRACE_CABAC_V( pcCU->getCUPelX() )
  DTRACE_CABAC_T( "\ttoCU-Y=" )
  DTRACE_CABAC_V( pcCU->getCUPelY() )
  DTRACE_CABAC_T( "\tCU-addr=" )
  DTRACE_CABAC_V(  pcCU->getAddr() )
  DTRACE_CABAC_T( "\tinCU-X=" )
  DTRACE_CABAC_V( g_auiRasterToPelX[ g_auiZscanToRaster[uiAbsPartIdx] ] )
  DTRACE_CABAC_T( "\tinCU-Y=" )
  DTRACE_CABAC_V( g_auiRasterToPelY[ g_auiZscanToRaster[uiAbsPartIdx] ] )
  DTRACE_CABAC_T( "\tpredmode=" )
  DTRACE_CABAC_V(  pcCU->getPredictionMode( uiAbsPartIdx ) )
  DTRACE_CABAC_N

#if 0
  // DEBUG
  printf("encoding pcCoeff\n");
  for (int i = 0; i<uiHeight; i++){
    for (int j = 0; j<uiWidth; j++){
      printf("%d\t", pcCoef[j + i*uiWidth]);
    }
    printf("\n");
  }

  printf("pcCoeffRefLayer\n");
  for (int i = 0; i<uiHeight; i++){
    for (int j = 0; j<uiWidth; j++){
      printf("%d\t", pCoefRefLayer[j + i*uiWidth]);
    }
    printf("\n");
  }
#endif

  UInt uiNumSig = 0;

  // compute number of significant coefficients
  uiNumSig = TEncEntropy::countNonZeroCoeffs(pcCoef, uiWidth * uiHeight);

  if ( uiNumSig == 0 )
    return;

  eTType = eTType == TEXT_LUMA ? TEXT_LUMA : ( eTType == TEXT_NONE ? TEXT_NONE : TEXT_CHROMA );

  //----- encode significance map -----
  UInt uiScanIdx;
  if ( getEncCfg()->getMyCodingArgument()[USE_BL_SCAN_IDX] )
  {
    uiScanIdx = pcColCU->getCoefScanIdx(uiAbsPartIdx, uiWidth, eTType==TEXT_LUMA, pcColCU->isIntra(uiAbsPartIdx));
  }
  else
  {
    uiScanIdx = pcCU->getCoefScanIdx(uiAbsPartIdx, uiWidth, eTType==TEXT_LUMA, pcCU->isIntra(uiAbsPartIdx));
  }

  const UInt   uiLog2BlockSize   = g_aucConvertToBit[ uiWidth ] + 2;
  
  if (uiScanIdx == SCAN_ZIGZAG)
  {
    // Map zigzag to diagonal scan
    uiScanIdx = SCAN_DIAG;
  }
  Int blockType = uiLog2BlockSize;
  if (uiWidth != uiHeight)
  {
    uiScanIdx = SCAN_DIAG;
    blockType = 4;
  }

  const UInt *scan = g_auiSigLastScan[ uiScanIdx ][ uiLog2BlockSize - 1 ];
#if !REMOVE_NSQT
  if (uiWidth != uiHeight)
  {
    scan = g_sigScanNSQT[ uiLog2BlockSize - 2 ];
  }
#endif

  // Find position of last coefficient and first non-significant coefficient of reference layer
  Int scanPosLast = -1;
  Int posLast;
  Int iFirstNonSigRefLayer = -1;

  const UInt * scanCG;
#if !REMOVE_NSQT
  if (uiWidth == uiHeight)
#endif
  {
    scanCG = g_auiSigLastScan[ uiScanIdx ][ uiLog2BlockSize > 3 ? uiLog2BlockSize-2-1 : 0 ];
    if( uiLog2BlockSize == 3 )
    {
      scanCG = g_sigLastScan8x8[ uiScanIdx ];
    }
    else if( uiLog2BlockSize == 5 )
    {
      scanCG = g_sigLastScanCG32x32;
    }
  }
#if !REMOVE_NSQT
  else
  {
    scanCG = g_sigCGScanNSQT[ uiLog2BlockSize - 2 ];
  }
#endif

  UInt uiSigCoeffGroupFlag[ MLS_GRP_NUM ];
  UInt uiSigCoeffGroupFlagRefLayer[ MLS_GRP_NUM ];
  static const UInt uiShift = MLS_CG_SIZE >> 1;
  const UInt uiNumBlkSide = uiWidth >> uiShift;

    // Scan coeffs
  ::memset( uiSigCoeffGroupFlag,          0, sizeof(UInt) * MLS_GRP_NUM );
  ::memset( uiSigCoeffGroupFlagRefLayer,  0, sizeof(UInt) * MLS_GRP_NUM );
  do
  {
    posLast = scan[ ++scanPosLast ];

    // get L1 sig map
    UInt uiPosY    = posLast >> uiLog2BlockSize;
    UInt uiPosX    = posLast - ( uiPosY << uiLog2BlockSize );
    UInt uiBlkIdx  = uiNumBlkSide * (uiPosY >> uiShift) + (uiPosX >> uiShift);
#if !REMOVAL_8x2_2x8_CG
    if( uiWidth == 8 && uiHeight == 8 && (uiScanIdx == SCAN_HOR || uiScanIdx == SCAN_VER) )
    {
      if( uiScanIdx == SCAN_HOR )
      {
        uiBlkIdx = uiPosY >> 1;
      }
      else if( uiScanIdx == SCAN_VER )
      {
        uiBlkIdx = uiPosX >> 1;
      }
    }
#endif
    if( pcCoef[ posLast ] )
    {
      uiSigCoeffGroupFlag[ uiBlkIdx ] = 1;
    }
    if( pCoefRefLayer[ posLast ] )
    {
      uiSigCoeffGroupFlagRefLayer[ uiBlkIdx ] = 1;
    }
    else
    {
      if (iFirstNonSigRefLayer<0)
      {
        iFirstNonSigRefLayer = scanPosLast;
      }
    }
    uiNumSig -= ( pcCoef[ posLast ] != 0 );
  }
  while ( uiNumSig > 0 );

  // Code position of last coefficient
  UInt posLastY = posLast >> uiLog2BlockSize;
  UInt posLastX = posLast - ( posLastY << uiLog2BlockSize );
#if RWTH_CODE_LASTXY_DIFF
  if ( getEncCfg()->getMyCodingArgument()[CODE_LAST_SIGNIFICANT_DIFF] )
  {
    // Get the last significant XY from the lower layer
    Int iScanPostLastLowerLayer = -1;
    Int posLastRef;
    UInt uiNumSigLowerLayer = TEncEntropy::countNonZeroCoeffs(pCoefRefLayer, uiWidth * uiHeight);
    do
    {
      posLastRef = scan[ ++iScanPostLastLowerLayer ];
      uiNumSigLowerLayer -= ( pCoefRefLayer[ posLastRef ] != 0 );
    }
    while ( uiNumSigLowerLayer > 0 );
    Int posLastYRef = posLastRef >> uiLog2BlockSize;
    Int posLastXRef = posLastRef - ( posLastYRef << uiLog2BlockSize );
    
    codeLastSignificantDiff( posLastX, posLastXRef, posLastY, posLastYRef, uiWidth, uiHeight, eTType, uiScanIdx, scan );
  }
  else
#endif
  {
    codeLastSignificantXY(posLastX, posLastY, uiWidth, uiHeight, eTType, uiScanIdx);
  }

#if ENC_DEC_TRACE
  switch (uiScanIdx)
  {
    case 0: DTRACE_CABAC_T("\tscanIdx=SCAN_ZIGZAG\n") break;
    case 1: DTRACE_CABAC_T("\tscanIdx=SCAN_HOR\n") break;
    case 2: DTRACE_CABAC_T("\tscanIdx=SCAN_VER\n") break;
    case 3: DTRACE_CABAC_T("\tscanIdx=SCAN_DIAG\n") break;
  }
#endif

  //===== code significance flag refinement =====
  ContextModel * const baseCoeffGroupCtx = m_cCUSigCoeffGroupSCModel.get( 0, eTType );
  ContextModel * const baseCtx = (eTType==TEXT_LUMA) ? m_cCUSigSCModel.get( 0, 0 ) : m_cCUSigSCModel.get( 0, 0 ) + NUM_SIG_FLAG_CTX_LUMA;

  const Int  iLastScanSet      = scanPosLast >> LOG2_SCAN_SET_SIZE;
  Int  iScanPosSig             = scanPosLast;
  Int pos[SCAN_SET_SIZE];

  for( Int iSubSet = iLastScanSet; iSubSet >= 0; iSubSet-- )
  {
    Int numSigCoeffRefLayer = 0;
    Int iNumNewSig   = 0;
    Int  iSubPos     = iSubSet << LOG2_SCAN_SET_SIZE;
    Int refinement[16];
    UInt coeffSigns = 0;

    if( iScanPosSig == scanPosLast )
    {
      if (pCoefRefLayer[ posLast ] != 0)
      {
#if RWTH_QUANT_USE_REF_COEFF
        refinement[0] = abs( pcCoef[ posLast ] ) - abs( pCoefRefLayer[ posLast] );
#else
        refinement[0] = abs( pcCoef[ posLast ] ) - 2* abs( pCoefRefLayer[ posLast] );
#endif
        pos [ 0 ] = posLast;
        numSigCoeffRefLayer    = 1;
      }
      else
      {
        coeffSigns    = ( pcCoef[ posLast ] < 0 );
        iNumNewSig++;
      }
      iScanPosSig--;
    }

    // encode significant_coeffgroup_flag if reference layer group is not significant
    Int iCGBlkPos = scanCG[ iSubSet ];
    Int iCGPosY   = iCGBlkPos / uiNumBlkSide;
    Int iCGPosX   = iCGBlkPos - (iCGPosY * uiNumBlkSide);

#if !REMOVAL_8x2_2x8_CG
    if( uiWidth == 8 && uiHeight == 8 && (uiScanIdx == SCAN_HOR || uiScanIdx == SCAN_VER) )
    {
      iCGPosY = (uiScanIdx == SCAN_HOR ? iCGBlkPos : 0);
      iCGPosX = (uiScanIdx == SCAN_VER ? iCGBlkPos : 0);
    }
#endif

    if( iSubSet == iLastScanSet || iSubSet == 0 || uiSigCoeffGroupFlagRefLayer[ iCGBlkPos ] )
    {
      uiSigCoeffGroupFlag[ iCGBlkPos ] = 1;
    }
    else
    {
      UInt uiSigCoeffGroup   = (uiSigCoeffGroupFlag[ iCGBlkPos ] != 0);
      UInt uiCtxSig  = TComTrQuant::getSigCoeffGroupCtxInc( uiSigCoeffGroupFlag, iCGPosX, iCGPosY, uiScanIdx, uiWidth, uiHeight );

      m_pcBinIf->encodeBin( uiSigCoeffGroup, baseCoeffGroupCtx[ uiCtxSig ] );
      DTRACE_CABAC_T( "\tGroup Significance\t" )
      DTRACE_CABAC_V( uiCtxSig )
      DTRACE_CABAC_N
    }

      // encode significant_coeff_flag refinement
    if( uiSigCoeffGroupFlag[ iCGBlkPos ] )
    {
      Int patternSigCtx = TComTrQuant::calcPatternSigCtx( uiSigCoeffGroupFlag, iCGPosX, iCGPosY, uiWidth, uiHeight );
      UInt uiBlkPos, uiPosY, uiPosX, uiSig, uiSigRefLayer, uiCtxSig;
      for( ; iScanPosSig >= iSubPos; iScanPosSig-- )
      {
        uiBlkPos  = scan[ iScanPosSig ];
        uiPosY    = uiBlkPos >> uiLog2BlockSize;
        uiPosX    = uiBlkPos - ( uiPosY << uiLog2BlockSize );
        uiSig     = (pcCoef[ uiBlkPos ] != 0);
        uiSigRefLayer = (pCoefRefLayer[ uiBlkPos ] != 0);

        if(iScanPosSig > iSubPos || iSubSet == 0 || iNumNewSig || numSigCoeffRefLayer)
        {
          if(uiSigRefLayer==0 && iScanPosSig>= iFirstNonSigRefLayer)
          {
#if REMOVAL_8x2_2x8_CG
            uiCtxSig  = TComTrQuant::getSigCtxInc( patternSigCtx, uiScanIdx, uiPosX, uiPosY, blockType, uiWidth, uiHeight, eTType );
#else
            uiCtxSig  = TComTrQuant::getSigCtxInc( patternSigCtx, uiPosX, uiPosY, blockType, uiWidth, uiHeight, eTType );
#endif
            //uiCtxSig  = TComTrQuant::getSigCtxInc( pcCoef, uiPosX, uiPosY, blockType, uiWidth, uiHeight, eTType );
            m_pcBinIf->encodeBin( uiSig, baseCtx[ uiCtxSig ] );
            DTRACE_CABAC_T( "\tsignificant_coeff_flag\tctx:" )
            DTRACE_CABAC_V( uiCtxSig )
            DTRACE_CABAC_T( "\tscanPos:" )
            DTRACE_CABAC_V( iScanPosSig )
            DTRACE_CABAC_T( "\tblockPos:" )
            DTRACE_CABAC_V( uiBlkPos )
            DTRACE_CABAC_T( " (" )
            DTRACE_CABAC_V( uiPosX )
            DTRACE_CABAC_T( "," )
            DTRACE_CABAC_V( uiPosY )
            DTRACE_CABAC_T( ")\n" )
          }
        }
        if (uiSig)
        {
          if (uiSigRefLayer)
          {
#if RWTH_QUANT_USE_REF_COEFF
            refinement[ numSigCoeffRefLayer ] = abs( pcCoef[ uiBlkPos ] ) - abs( pCoefRefLayer[ uiBlkPos ] );
#else
            refinement[ numSigCoeffRefLayer ] = abs( pcCoef[ uiBlkPos ] ) - 2* abs( pCoefRefLayer[ uiBlkPos ] );
#endif
            pos [ numSigCoeffRefLayer ] = uiBlkPos;
            assert (refinement[ numSigCoeffRefLayer ]==0 || refinement[ numSigCoeffRefLayer ]==1);
            numSigCoeffRefLayer++;
          }
          else
          {
            coeffSigns = 2 * coeffSigns + ( pcCoef[ uiBlkPos ] < 0 );
            iNumNewSig++;
          }
        }
      }
    }
    else
    {
      iScanPosSig = iSubPos - 1;
    }

    // encode sign of new significant coefficients
    m_pcBinIf->encodeBinsEP( coeffSigns, iNumNewSig );

#if RWTH_CTX_FOR_REFINEMENT
    ContextModel * const baseCoeffRefinement = m_cCURefinementModel.get( 0, eTType? TEXT_CHROMA: TEXT_LUMA );
#endif
    // encode coefficient level refinement
    for ( Int idx = 0; idx < numSigCoeffRefLayer; idx++ )
    {
#if RWTH_CTX_FOR_REFINEMENT
      m_pcBinIf->encodeBin( refinement[idx], baseCoeffRefinement[0]);
#else
      m_pcBinIf->encodeBinEP( refinement[idx] );
#endif
#if ENC_DEC_TRACE
      DTRACE_CABAC_T( "\tcoeff_refinement\tctx:" )
      DTRACE_CABAC_V( 0 )
      DTRACE_CABAC_T( "\tscanPos:" )
      DTRACE_CABAC_V( idx )
      UInt uiBlockPos = pos[idx];
      UInt uiPosY = uiBlockPos >> uiLog2BlockSize;
      UInt uiPosX = uiBlockPos - ( uiPosY << uiLog2BlockSize );
      DTRACE_CABAC_T( "\tblockPos:" )
      DTRACE_CABAC_V( uiBlockPos )
      DTRACE_CABAC_T( " (" )
      DTRACE_CABAC_V( uiPosX )
      DTRACE_CABAC_T( "," )
      DTRACE_CABAC_V( uiPosY )
      DTRACE_CABAC_T( ")\n" )
#endif
    }
  }
  DTRACE_CABAC_T( "\tEnd Code Refinement\n" )

#if ENC_DEC_TRACE_COEFF
  // Trace coeffs and reference block coeffs
  DTRACE_CABAC_T( "\t     Coeff: " )
  for (int i = 0; i<uiHeight; i++){
    for (int j = 0; j<uiWidth; j++){
      DTRACE_CABAC_V( pcCoef[j + i*uiWidth] )
      DTRACE_CABAC_T( " " )
    }
  }
  DTRACE_CABAC_N

  UInt uiCoeff;
  DTRACE_CABAC_T( "\t Ref Coeff: " )
  for (int i = 0; i<uiHeight; i++){
    for (int j = 0; j<uiWidth; j++){
      uiCoeff = pCoefRefLayer[j + i*uiWidth] * (pcCoef[ j + i*uiWidth ] < 0 ? -1 : 1);
      DTRACE_CABAC_V( uiCoeff )
      DTRACE_CABAC_T( " " )
    }
  }
  DTRACE_CABAC_N
#endif

}

#endif


#if !REMOVE_ALF
Void TEncSbac::codeAlfCtrlFlag( Int compIdx, UInt code )
{
  m_pcBinIf->encodeBin( code, m_cCUAlfCtrlFlagSCModel.get( 0, 0, 0) );

  DTRACE_CABAC_VL( g_nSymbolCounter++ )
  DTRACE_CABAC_T( "codeAlfCtrlFlag()" )
  DTRACE_CABAC_T( "\tsymbol=" )
  DTRACE_CABAC_V( code )
  DTRACE_CABAC_T( "\tcompIdx=" )
  DTRACE_CABAC_V( compIdx )
  DTRACE_CABAC_T( "\n" )
}
#endif

/** code SAO offset sign
 * \param code sign value
 */
Void TEncSbac::codeSAOSign( UInt code )
{
  m_pcBinIf->encodeBinEP( code );
  DTRACE_CABAC_T( "\tSAOSign\n" )
}

Void TEncSbac::codeSaoMaxUvlc    ( UInt code, UInt maxSymbol )
{
  if (maxSymbol == 0)
  {
    return;
  }

  Int i;
  Bool bCodeLast = ( maxSymbol > code );

  if ( code == 0 )
  {
#if SAO_ABS_BY_PASS
    m_pcBinIf->encodeBinEP( 0 );
#else
    m_pcBinIf->encodeBin( 0, m_cSaoUvlcSCModel.get( 0, 0, 0 ) );
#endif
    DTRACE_CABAC_T( "\tSaoUvlc\n" )
  }
  else
  {
#if SAO_ABS_BY_PASS
    m_pcBinIf->encodeBinEP( 1 );
#else
    m_pcBinIf->encodeBin( 1, m_cSaoUvlcSCModel.get( 0, 0, 0 ) );
#endif
    DTRACE_CABAC_T( "\tSaoUvlc\n" )

    for ( i=0; i<code-1; i++ )
    {
#if SAO_ABS_BY_PASS
      m_pcBinIf->encodeBinEP( 1 );
#else
      m_pcBinIf->encodeBin( 1, m_cSaoUvlcSCModel.get( 0, 0, 1 ) );
#endif
      DTRACE_CABAC_T( "\tSaoUvlc\n" )
    }
    if( bCodeLast )
    {
#if SAO_ABS_BY_PASS
      m_pcBinIf->encodeBinEP( 0 );
#else
      m_pcBinIf->encodeBin( 0, m_cSaoUvlcSCModel.get( 0, 0, 1 ) );
#endif
      DTRACE_CABAC_T( "\tSaoUvlc\n" )
    }
  }
}


#if SAO_TYPE_CODING
/** Code SAO EO class or BO band position 
 * \param uiLength
 * \param uiCode
 */
Void TEncSbac::codeSaoUflc       ( UInt uiLength, UInt uiCode )
{
   m_pcBinIf->encodeBinsEP ( uiCode, uiLength );
}
#else
/** Code SAO band position 
 * \param uiCode
 */
Void TEncSbac::codeSaoUflc       ( UInt uiCode )
{
    m_pcBinIf->encodeBinsEP ( uiCode, 5 );
}
#endif
#if SAO_MERGE_ONE_CTX
/** Code SAO merge flags
 * \param uiCode
 * \param uiCompIdx
 */
Void TEncSbac::codeSaoMerge       ( UInt uiCode )
#else
/** Code SAO merge left flag 
 * \param uiCode
 * \param uiCompIdx
 */
Void TEncSbac::codeSaoMergeLeft       ( UInt uiCode, UInt uiCompIdx )
#endif
{
  if (uiCode == 0)
  {
#if SAO_MERGE_ONE_CTX
    m_pcBinIf->encodeBin(0,  m_cSaoMergeSCModel.get( 0, 0, 0 ));
#else
#if SAO_SINGLE_MERGE
    m_pcBinIf->encodeBin(0,  m_cSaoMergeLeftSCModel.get( 0, 0, 0 ));
#else
    m_pcBinIf->encodeBin(0,  m_cSaoMergeLeftSCModel.get( 0, 0, uiCompIdx ));
#endif
#endif
    DTRACE_CABAC_T( "\tSaoMergeLeft\n" )
  }
  else
  {
#if SAO_MERGE_ONE_CTX
    m_pcBinIf->encodeBin(1,  m_cSaoMergeSCModel.get( 0, 0, 0 ));
#else
#if SAO_SINGLE_MERGE
    m_pcBinIf->encodeBin(1,  m_cSaoMergeLeftSCModel.get( 0, 0, 0 ));
#else
    m_pcBinIf->encodeBin(1,  m_cSaoMergeLeftSCModel.get( 0, 0, uiCompIdx ));
#endif
#endif
   DTRACE_CABAC_T( "\tSaoMergeLeft\n" )
  }
}
#if !SAO_MERGE_ONE_CTX
/** Code SAO merge up flag 
 * \param uiCode
 */
Void TEncSbac::codeSaoMergeUp       ( UInt uiCode)
{
  if (uiCode == 0)
  {
    m_pcBinIf->encodeBin(0,  m_cSaoMergeUpSCModel.get( 0, 0, 0 ));
  }
  else
  {
    m_pcBinIf->encodeBin(1,  m_cSaoMergeUpSCModel.get( 0, 0, 0 ));
  }
  DTRACE_CABAC_T( "\tSaoMergeUp\n" )
}
#endif
/** Code SAO type index 
 * \param uiCode
 */
Void TEncSbac::codeSaoTypeIdx       ( UInt uiCode)
{
#if SAO_TYPE_CODING
  if (uiCode == 0)
  {
    m_pcBinIf->encodeBin( 0, m_cSaoTypeIdxSCModel.get( 0, 0, 0 ) );
    DTRACE_CABAC_T( "\tSaoTypeIdx\n" );
  }
  else
  {
    m_pcBinIf->encodeBin( 1, m_cSaoTypeIdxSCModel.get( 0, 0, 0 ) );
    DTRACE_CABAC_T( "\tSaoTypeIdx\n" );
    m_pcBinIf->encodeBinEP( uiCode <= 4 ? 1 : 0 );
    DTRACE_CABAC_T( "\tSaoTypeIdx\n" );
  }
#else
  Int i;
  if ( uiCode == 0 )
  {
    m_pcBinIf->encodeBin( 0, m_cSaoTypeIdxSCModel.get( 0, 0, 0 ) );
    DTRACE_CABAC_T( "\tSaoTypeIdx\n" )
  }
  else
  {
    m_pcBinIf->encodeBin( 1, m_cSaoTypeIdxSCModel.get( 0, 0, 0 ) );
    DTRACE_CABAC_T( "\tSaoTypeIdx\n" )
    for ( i=0; i<uiCode-1; i++ )
    {
      m_pcBinIf->encodeBin( 1, m_cSaoTypeIdxSCModel.get( 0, 0, 1 ) );
      DTRACE_CABAC_T( "\tSaoTypeIdx\n" )
    }
    m_pcBinIf->encodeBin( 0, m_cSaoTypeIdxSCModel.get( 0, 0, 1 ) );
    DTRACE_CABAC_T( "\tSaoTypeIdx\n" )
  }
#endif
}
/*!
 ****************************************************************************
 * \brief
 *   estimate bit cost for CBP, significant map and significant coefficients
 ****************************************************************************
 */
Void TEncSbac::estBit( estBitsSbacStruct* pcEstBitsSbac, Int width, Int height, TextType eTType )
{
  estCBFBit( pcEstBitsSbac, 0, eTType );

  estSignificantCoeffGroupMapBit( pcEstBitsSbac, 0, eTType );
  
  // encode significance map
  estSignificantMapBit( pcEstBitsSbac, width, height, eTType );
  
  // encode significant coefficients
  estSignificantCoefficientsBit( pcEstBitsSbac, 0, eTType );
}

/*!
 ****************************************************************************
 * \brief
 *    estimate bit cost for each CBP bit
 ****************************************************************************
 */
Void TEncSbac::estCBFBit( estBitsSbacStruct* pcEstBitsSbac, UInt uiCTXIdx, TextType eTType )
{
  ContextModel *pCtx = m_cCUQtCbfSCModel.get( 0 );

  for( UInt uiCtxInc = 0; uiCtxInc < 3*NUM_QT_CBF_CTX; uiCtxInc++ )
  {
    pcEstBitsSbac->blockCbpBits[ uiCtxInc ][ 0 ] = pCtx[ uiCtxInc ].getEntropyBits( 0 );
    pcEstBitsSbac->blockCbpBits[ uiCtxInc ][ 1 ] = pCtx[ uiCtxInc ].getEntropyBits( 1 );
  }

  pCtx = m_cCUQtRootCbfSCModel.get( 0 );
  
  for( UInt uiCtxInc = 0; uiCtxInc < 4; uiCtxInc++ )
  {
    pcEstBitsSbac->blockRootCbpBits[ uiCtxInc ][ 0 ] = pCtx[ uiCtxInc ].getEntropyBits( 0 );
    pcEstBitsSbac->blockRootCbpBits[ uiCtxInc ][ 1 ] = pCtx[ uiCtxInc ].getEntropyBits( 1 );
  }
}


/*!
 ****************************************************************************
 * \brief
 *    estimate SAMBAC bit cost for significant coefficient group map
 ****************************************************************************
 */
Void TEncSbac::estSignificantCoeffGroupMapBit( estBitsSbacStruct* pcEstBitsSbac, UInt uiCTXIdx, TextType eTType )
{
  Int firstCtx = 0, numCtx = NUM_SIG_CG_FLAG_CTX;

  for ( Int ctxIdx = firstCtx; ctxIdx < firstCtx + numCtx; ctxIdx++ )
  {
    for( UInt uiBin = 0; uiBin < 2; uiBin++ )
    {
      pcEstBitsSbac->significantCoeffGroupBits[ ctxIdx ][ uiBin ] = m_cCUSigCoeffGroupSCModel.get(  0, eTType, ctxIdx ).getEntropyBits( uiBin );
    }
  }
}


/*!
 ****************************************************************************
 * \brief
 *    estimate SAMBAC bit cost for significant coefficient map
 ****************************************************************************
 */
Void TEncSbac::estSignificantMapBit( estBitsSbacStruct* pcEstBitsSbac, Int width, Int height, TextType eTType )
{
  Int firstCtx = 1, numCtx = 8;
  if (max(width, height) >= 16)
  {
#if REMOVAL_8x2_2x8_CG
    firstCtx = (eTType == TEXT_LUMA) ? 21 : 12;
#else
    firstCtx = 18;
#endif
    numCtx = (eTType == TEXT_LUMA) ? 6 : 3;
  }
  else if (width == 8)
  {
    firstCtx = 9;
#if REMOVAL_8x2_2x8_CG
    numCtx = (eTType == TEXT_LUMA) ? 12 : 3;
#else
    numCtx = 9;
#endif
  }
  
  if (eTType == TEXT_LUMA )
  {
    for( UInt bin = 0; bin < 2; bin++ )
    {
      pcEstBitsSbac->significantBits[ 0 ][ bin ] = m_cCUSigSCModel.get(  0, 0, 0 ).getEntropyBits( bin );
    }

    for ( Int ctxIdx = firstCtx; ctxIdx < firstCtx + numCtx; ctxIdx++ )
    {
      for( UInt uiBin = 0; uiBin < 2; uiBin++ )
      {
        pcEstBitsSbac->significantBits[ ctxIdx ][ uiBin ] = m_cCUSigSCModel.get(  0, 0, ctxIdx ).getEntropyBits( uiBin );
      }
    }
  }
  else
  {
    for( UInt bin = 0; bin < 2; bin++ )
    {
      pcEstBitsSbac->significantBits[ 0 ][ bin ] = m_cCUSigSCModel.get(  0, 0, NUM_SIG_FLAG_CTX_LUMA + 0 ).getEntropyBits( bin );
    }
    for ( Int ctxIdx = firstCtx; ctxIdx < firstCtx + numCtx; ctxIdx++ )
    {
      for( UInt uiBin = 0; uiBin < 2; uiBin++ )
      {
        pcEstBitsSbac->significantBits[ ctxIdx ][ uiBin ] = m_cCUSigSCModel.get(  0, 0, NUM_SIG_FLAG_CTX_LUMA + ctxIdx ).getEntropyBits( uiBin );
      }
    }
  }
  Int iBitsX = 0, iBitsY = 0;
  Int blkSizeOffsetX, blkSizeOffsetY, shiftX, shiftY;

  blkSizeOffsetX = eTType ? 0: (g_aucConvertToBit[ width ] *3 + ((g_aucConvertToBit[ width ] +1)>>2));
  blkSizeOffsetY = eTType ? 0: (g_aucConvertToBit[ height ]*3 + ((g_aucConvertToBit[ height ]+1)>>2));
  shiftX = eTType ? g_aucConvertToBit[ width  ] :((g_aucConvertToBit[ width  ]+3)>>2);
  shiftY = eTType ? g_aucConvertToBit[ height ] :((g_aucConvertToBit[ height ]+3)>>2);

  Int ctx;
  ContextModel *pCtxX      = m_cCuCtxLastX.get( 0, eTType );
  for (ctx = 0; ctx < g_uiGroupIdx[ width - 1 ]; ctx++)
  {
    Int ctxOffset = blkSizeOffsetX + (ctx >>shiftX);
    pcEstBitsSbac->lastXBits[ ctx ] = iBitsX + pCtxX[ ctxOffset ].getEntropyBits( 0 );
    iBitsX += pCtxX[ ctxOffset ].getEntropyBits( 1 );
  }
  pcEstBitsSbac->lastXBits[ctx] = iBitsX;
  ContextModel *pCtxY      = m_cCuCtxLastY.get( 0, eTType );
  for (ctx = 0; ctx < g_uiGroupIdx[ height - 1 ]; ctx++)
  {
    Int ctxOffset = blkSizeOffsetY + (ctx >>shiftY);
    pcEstBitsSbac->lastYBits[ ctx ] = iBitsY + pCtxY[ ctxOffset ].getEntropyBits( 0 );
    iBitsY += pCtxY[ ctxOffset ].getEntropyBits( 1 );
  }
  pcEstBitsSbac->lastYBits[ctx] = iBitsY;
}

/*!
 ****************************************************************************
 * \brief
 *    estimate bit cost of significant coefficient
 ****************************************************************************
 */
Void TEncSbac::estSignificantCoefficientsBit( estBitsSbacStruct* pcEstBitsSbac, UInt uiCTXIdx, TextType eTType )
{
  if (eTType==TEXT_LUMA)
  {
    ContextModel *ctxOne = m_cCUOneSCModel.get(0, 0);
    ContextModel *ctxAbs = m_cCUAbsSCModel.get(0, 0);

    for (Int ctxIdx = 0; ctxIdx < NUM_ONE_FLAG_CTX_LUMA; ctxIdx++)
    {
      pcEstBitsSbac->m_greaterOneBits[ ctxIdx ][ 0 ] = ctxOne[ ctxIdx ].getEntropyBits( 0 );
      pcEstBitsSbac->m_greaterOneBits[ ctxIdx ][ 1 ] = ctxOne[ ctxIdx ].getEntropyBits( 1 );    
    }

    for (Int ctxIdx = 0; ctxIdx < NUM_ABS_FLAG_CTX_LUMA; ctxIdx++)
    {
      pcEstBitsSbac->m_levelAbsBits[ ctxIdx ][ 0 ] = ctxAbs[ ctxIdx ].getEntropyBits( 0 );
      pcEstBitsSbac->m_levelAbsBits[ ctxIdx ][ 1 ] = ctxAbs[ ctxIdx ].getEntropyBits( 1 );    
    }
  }
  else
  {
    ContextModel *ctxOne = m_cCUOneSCModel.get(0, 0) + NUM_ONE_FLAG_CTX_LUMA;
    ContextModel *ctxAbs = m_cCUAbsSCModel.get(0, 0) + NUM_ABS_FLAG_CTX_LUMA;

    for (Int ctxIdx = 0; ctxIdx < NUM_ONE_FLAG_CTX_CHROMA; ctxIdx++)
    {
      pcEstBitsSbac->m_greaterOneBits[ ctxIdx ][ 0 ] = ctxOne[ ctxIdx ].getEntropyBits( 0 );
      pcEstBitsSbac->m_greaterOneBits[ ctxIdx ][ 1 ] = ctxOne[ ctxIdx ].getEntropyBits( 1 );    
    }

    for (Int ctxIdx = 0; ctxIdx < NUM_ABS_FLAG_CTX_CHROMA; ctxIdx++)
    {
      pcEstBitsSbac->m_levelAbsBits[ ctxIdx ][ 0 ] = ctxAbs[ ctxIdx ].getEntropyBits( 0 );
      pcEstBitsSbac->m_levelAbsBits[ ctxIdx ][ 1 ] = ctxAbs[ ctxIdx ].getEntropyBits( 1 );    
    }
  }
}

/**
 - Initialize our context information from the nominated source.
 .
 \param pSrc From where to copy context information.
 */
Void TEncSbac::xCopyContextsFrom( TEncSbac* pSrc )
{  
  memcpy(m_contextModels, pSrc->m_contextModels, m_numContextModels*sizeof(m_contextModels[0]));
}

Void  TEncSbac::loadContexts ( TEncSbac* pScr)
{
  this->xCopyContextsFrom(pScr);
}
//! \}
